Require Import RelationClasses.
Require Import Coqlib.
Require Import Kanglib.
Require Import Integers.
Require Import Maps.
Require Import AST.
Require Import Values.
Require Import Monad.

Ltac tac :=
  repeat
    (simpl in *; subst;
     (try done);
     (try basictac_aux);
     (try destac_aux);
     (try monadtac_aux);
     (try maptac_aux);
     (try Range.tac_aux)).

(**
 - Memory supports 6 operations:
   + empty
   + load
   + store
   + malloc
   + free
   + realization: change logical block into a physical one.
 - Memory consists of blocks. Above operations on memory are properly
 dispatched into operations on blocks.
 *)

(** COMMENT(jeehoon.kang): Note that in this memory model, a "byte"
contains a value, whereas in reality 4 bytes (in 32bit machine)
contain a value. This is a simplification that is orthogonal to the
integer-pointer casting problem, I believe.  *)
Module MemC.
  (** Memory consists of memory blocks. *)
  Module Block.
    (** Memory block consists of:
     - valid: flag for valid blocks; free makes a block invalid.
     - size: of the block.
     - address:
       + `None`      if the block is logical;
       + `Some p`    if the block is physically realized and its address is `p`.
     - content: map from Z to Value.
     *)
    Structure t: Type := mk {
      valid: bool;
      accessible: bool;
      size: int;
      address: option int;
      content: ZMap.t ValueC.t
    }.

    (** Memory block should satisfy the following properties:
     - address + size should fit in the memory space;
     - address is nonzero iff size is nonzero.
     *)
    Structure wf (blk:t): Prop := mkWf {
      _size:
        forall
          p (Hp: blk.(address) = Some p),
          Range.wf (Range.mk p blk.(size));
      _addr:
        blk.(address)
    }.

    Definition init (size:int): t :=
      mk
        true
        true
        size
        None
        (ZMap.init (ValueC.int Int.zero)).
    (* Lemma init_wf (size:int): *)
    (*   wf (init size). *)
    (* Proof. *)
    (*   by constructor; unfold init; simpl; intros. *)
    (* Qed. *)

    (* None <= None,
       None <= Some p,
       Some p <= Some p.
       This is auxiliary definition for proofs on memory relation.
     *)
    Definition le_address (s t:option int): Prop :=
      forall p (Hsrc: s = Some p), t = Some p.

    Global Program Instance le_address_PreOrder : PreOrder le_address.
    Next Obligation.
      by repeat intro.
    Qed.
    Next Obligation.
      repeat intro. apply H0. by apply H.
    Qed.

    Definition load (o:int) (blk:t): option ValueC.t :=
      if blk.(valid) && blk.(accessible) && Range.contained o (Range.mk Int.zero blk.(size))
      then Some (ZMap.get (Int.unsigned o) blk.(content))
      else None.
    Lemma load_inv
          (o:int) (blk:t) (result:ValueC.t)
          (H: load o blk = Some result):
      blk.(valid) /\
      blk.(accessible) /\
      Range.contained o (Range.mk Int.zero blk.(size)) /\
      ZMap.get (Int.unsigned o) blk.(content) = result.
    Proof.
      unfold load in H. tac.
    Qed.

    Definition store (o:int) (v:ValueC.t) (blk:t): option t :=
      if blk.(valid) && blk.(accessible) && Range.contained o (Range.mk Int.zero blk.(size))
      then
        Some
          (mk
             true
             true
             blk.(size)
             blk.(address)
             (ZMap.set (Int.unsigned o) v blk.(content)))
      else None.
    Lemma store_wf
          (o:int) (v:ValueC.t)
          (blk1:t) (Hwf: wf blk1)
          blk2 (Hblk2: store o v blk1 = Some blk2) :
      wf blk2.
    Proof.
      unfold store in *. tac. inv Hwf.
      constructor; simpl; intros; auto.
    Qed.
    Lemma store_inv
          (o:int) (v:ValueC.t) (blk1 blk2:t)
          (H: store o v blk1 = Some blk2):
      blk1.(valid) /\ blk2.(valid) /\
      blk1.(accessible) /\ blk2.(accessible) /\
      blk1.(size) = blk2.(size) /\
      blk1.(address) = blk2.(address) /\
      Range.contained o (Range.mk Int.zero blk1.(size)) /\
      ZMap.set (Int.unsigned o) v blk1.(content) = blk2.(content).
    Proof.
      unfold store in H. tac.
      repeat (split; auto).
    Qed.
    Lemma store_le_address
          (o:int) (v:ValueC.t) (blk1 blk2:t)
          (H: store o v blk1 = Some blk2):
      le_address blk1.(address) blk2.(address).
    Proof.
      apply store_inv in H. tac.
      repeat intro. by rewrite <- H4.
    Qed.

    (** - `realize` assigns a physical address to a logical block.
        - Choice of physical adderss is nondeterministic.
        - Realization occurs if a pointer is casted into integer. See
          lang/Semantics.v for more details.
     *)
    Inductive realize (blk1:t) (addr:int) (blk2:t): Prop :=
    | realize_
        (Haddress: blk1.(address) = None)
        (Hblk2:
           blk2 =
           mk
             blk1.(valid)
             blk1.(accessible)
             blk1.(size)
             (Some addr)
             blk1.(content))
        (Hwf2: wf blk2):
        realize blk1 addr blk2
    .
    Lemma realize_le_address
          blk1 addr blk2
          (H: realize blk1 addr blk2):
      le_address blk1.(address) blk2.(address).
    Proof.
      inv H. repeat intro. by rewrite Haddress in Hsrc.
    Qed.
  End Block.

  (** Memory consists of:
   - nextb: the next block id to be allocated;
   - blocks: a map from positive to memory block.
   *)
  Structure t: Type := mk {
    nextb: positive;
    blocks:> PTree.t Block.t
  }.

  Definition eq (m1 m2:t): Prop :=
    m1.(nextb) = m2.(nextb) /\
    PTree_Properties.Equal eq_equivalence m1 m2.

  Lemma eq_refl: forall m, eq m m.
  Proof.
    intros. red. split; auto.
    apply PTree_Properties.Equal_refl.
  Qed.

  Lemma eq_sym: forall m1 m2, eq m1 m2 -> eq m2 m1.
  Proof.
    intros. inv H. red. split; auto.
    by apply PTree_Properties.Equal_sym.
  Qed.

  Lemma eq_trans: forall m1 m2 m3, eq m1 m2 -> eq m2 m3 -> eq m1 m3.
  Proof.
    intros. inv H. inv H0. red. split.
    - by rewrite H1.
    - eapply PTree_Properties.Equal_trans; eauto.
  Qed.

  Global Instance eq_Equivalence : Equivalence eq := {
    Equivalence_Reflexive := eq_refl;
    Equivalence_Symmetric := eq_sym;
    Equivalence_Transitive := eq_trans
  }.

  Lemma blocks_unsimpl blks nextb:
    blks = blocks (mk nextb blks).
  Proof. done. Qed.

  (** Memory should satisfy the following properties:
      - if a block id `b` is allocated then `b` < `nextb`.
      - all blocks are well-formed.
      - physically realized [and valid] memory blocks should be disjoint.
   *)  
  Structure wf (m:t): Prop := mkWf {
    _nextb:
      forall
        b blk (Hblk: PTree.get b m = Some blk),
        (b < m.(nextb))%positive;
    _blocks_wf:
      forall 
        b blk (Hblk: PTree.get b m = Some blk),
        Block.wf blk;
    _blocks_disjoint:
      forall
        b1 blk1 (Hblk1: PTree.get b1 m = Some blk1)
        (Hvalid1: blk1.(Block.valid))
        p1 (Hp1: blk1.(Block.address) = Some p1)
        b2 blk2 (Hblk2: PTree.get b2 m = Some blk2)
        (Hvalid2: blk2.(Block.valid))
        p2 (Hp2: blk2.(Block.address) = Some p2)
        (Hb12: b1 <> b2),
        Range.disjoint (Range.mk p1 blk1.(Block.size)) (Range.mk p2 blk2.(Block.size));
    _address_one:
      exists blk,
        PTree.get 1%positive m = Some blk /\
        blk.(Block.address) = Some Int.zero
  }.

  Definition le (m1 m2:PTree.t Block.t): Prop :=
    forall b blk1 (Hblk1: PTree.get b m1 = Some blk1),
      exists blk2,
        PTree.get b m2 = Some blk2 /\
        (blk2.(Block.valid) -> blk1.(Block.valid)) /\
        blk1.(Block.size) = blk2.(Block.size) /\
        Block.le_address blk1.(Block.address) blk2.(Block.address).

  Global Program Instance le_PreOrder : PreOrder le.
  Next Obligation.
    repeat intro. exists blk1. repeat split; try done.
  Qed.
  Next Obligation.
    repeat intro.
    exploit H; eauto. intro. tac.
    exploit H0; eauto. intro. tac.
    eexists. split; [by eauto|].
    repeat split; auto.
    - by rewrite H3.
    - by rewrite H4.
  Qed.

  Definition is_valid_ptr (b:ident) (o:int) (m:PTree.t Block.t): bool :=
    match PTree.get b m with
      | None => false
      | Some blk =>
        blk.(Block.valid) && Range.contained o (Range.mk 0 blk.(Block.size))
    end.

  Definition is_nonzero_ptr (b:ident) (o:int) (m:PTree.t Block.t): bool :=
    match PTree.get b m with
      | None => false
      | Some blk => Range.contained o (Range.mk 0 blk.(Block.size))
    end.

  Lemma get_nextb
        m (Hwf: wf m)
        b blk (Hget: PTree.get b m = Some blk):
    (b < m.(nextb))%positive.
  Proof.
    inv Hwf. eapply _nextb0. eauto.
  Qed.

  (** coerce a value to integer:
   - i => i;
   - (b,o) => |b|+o if b is physically realized as address |b|;
   - (b,o) => none
   *)
  Definition val2int (val:ValueC.t) (m:PTree.t Block.t): option int :=
    match val with
      | ValueC.int i => Some i
    end.

  Definition val2bool (val:ValueC.t) (m:PTree.t Block.t): option bool :=
    match val2int val m with
      | None => None
      | Some i => Some (Int.cmp Cne i Int.zero)
    end.

  Definition val2ptr (val:ValueC.t) (m:PTree.t Block.t): option (ident * int) :=
    match val with
      | ValueC.int i =>
        PTree.select
          (fun b blk =>
             doT base <- blk.(Block.address);
           assertB (blk.(Block.valid) && Range.contained i (Range.mk base blk.(Block.size)));
           Some (b, Int.sub i base))
          m
    end.

  Lemma val2ptr_is_valid_ptr ptr m b o
        (H: val2ptr ptr m = Some (b, o)):
    is_valid_ptr b o m.
  Proof.
    destruct ptr. inv H. tac.
    unfold is_valid_ptr. rewrite H, H1. simpl.
    apply andb_true_iff in H2. unfold Int.sub. tac.
    destruct intval as [x Hx], i as [begin Hbegin]. tac.
    apply andb_true_iff. tac. rewrite Int.Z_mod_modulus_eq.
    rewrite Zmod_small; [|omega].
    split.
    + destruct (zle 0 (x - begin)); [done|omega].
    + destruct (zlt (x - begin) (Int.unsigned (Block.size x0))); [done|omega].
  Qed.

  Lemma val2ptr_inv
        val m b o (H: val2ptr val m = Some (b, o)):
    exists j, val = ValueC.int j /\
    exists blk, PTree.get b m = Some blk /\
    blk.(Block.valid) /\
    exists i, blk.(Block.address) = Some i /\
    Range.contained j (Range.mk i blk.(Block.size)) /\
    o = Int.sub j i.
  Proof.
    destruct val; inv H.
    match goal with
      | [H: context [PTree.select ?pred ?m] |- _] =>
        generalize (PTree.gselectspec pred m);
          destruct (PTree.select pred m); [inv H|done]
    end.
    intros [b' [blk [Hblk ?]]].
    destruct (Block.address blk) eqn:X; inv H.
    destruct (Block.valid blk) eqn:Y; inv H1.
    destruct (Range.contained intval {| Range.begin := i; Range.size := Block.size blk |}) eqn:Z; inv H0.
    eexists. split; eauto. eexists. split; eauto.
  Qed.

  Definition init: t :=
    mk
      2%positive
      (PTree.set
         1%positive
         (Block.mk
            true
            false
            1
            (Some Int.zero)
            (ZMap.init (ValueC.int Int.zero)))
         (PTree.empty _)).
  Lemma init_wf: wf init.
  Proof.
    unfold init. constructor; intros.
    - rewrite <- blocks_unsimpl in Hblk. rewrite PTree.gsspec in Hblk.
      destruct (peq b 1); inv Hblk.
      + simpl. xomega.
      + rewrite PTree.gempty in H0. inv H0.
    - rewrite <- blocks_unsimpl in Hblk. rewrite PTree.gsspec in Hblk.
      destruct (peq b 1); inv Hblk.
      + constructor; simpl; intros; auto.
        inv Hp. auto.
      + rewrite PTree.gempty in H0. inv H0.
    - rewrite <- blocks_unsimpl in *. rewrite PTree.gsspec in *.
      destruct (peq b1 1); inv Hblk1.
      { destruct (peq b2 1); inv Hblk2.
        { congruence. }
        { rewrite PTree.gempty in *. congruence. }
      }
      { rewrite PTree.gempty in *. congruence. }
    - unfold blocks. rewrite PTree.gss. eexists. split; eauto.
  Qed.

  Definition load_aux (b:ident) (o:int) (m:PTree.t Block.t): option ValueC.t :=
    match PTree.get b m with
      | Some blk =>
        Block.load o blk
      | None => None
    end.
  Definition load (ptr:ValueC.t) (m:PTree.t Block.t): option ValueC.t :=
    match val2ptr ptr m with
      | None => None
      | Some (b, o) => load_aux b o m
    end.
  Lemma load_inv (ptr:ValueC.t) (m:PTree.t Block.t) (val:ValueC.t)
    (H: load ptr m = Some val):
    exists b o, val2ptr ptr m = Some (b, o) /\
    exists blk, PTree.get b m = Some blk /\
    Block.load o blk = Some val.
  Proof.
    unfold load, load_aux in H.
    destruct (val2ptr ptr m) as [[b o]|]; [|done].
    exists b. exists o. split; [done|].
    destruct (m ! b) as [blk|]; [|done].
    by exists blk.
  Qed.

  Definition store_aux (b:ident) (o:int) (val:ValueC.t) (m:PTree.t Block.t): option (PTree.t Block.t) :=
    match PTree.get b m with
      | None => None
      | Some blk =>
        option_map (fun newblk => PTree.set b newblk m) (Block.store o val blk)
    end.
  Definition store (ptr:ValueC.t) (val:ValueC.t) (m:t): option t :=
    match val2ptr ptr m with
      | None => None
      | Some (b, o) =>
        option_map (fun blks => mk m.(nextb) blks) (store_aux b o val m)
    end.
  Lemma store_wf
        (ptr:ValueC.t) (val:ValueC.t) (m1:t) (Hwf1: wf m1)
        m2 (Hm2: store ptr val m1 = Some m2):
    wf m2.
  Proof.
    unfold store, store_aux in *. tac. inv Hwf1.
    constructor; intros.
    - eapply _nextb0; eauto. tac.
      instantiate (1 := if peq b i then _ else _).
      destruct (peq b i); [subst b|]; tac; eauto.
    - (destruct (peq b i); [subst b|]; tac).
      + eapply Block.store_wf; eauto.
      + eapply _blocks_wf0. eauto.
    - (destruct (peq b1 i); [subst b1|]; tac);
      (destruct (peq b2 i); [subst b2|]; tac).
      + apply Block.store_inv in Hcond0. tac.
        exploit _blocks_disjoint0;
        [apply Hcond1|idtac|idtac|apply Hblk2|idtac|idtac|idtac|idtac];
        eauto.
        * rewrite H6. eauto.
        * rewrite H5. auto.
      + apply Block.store_inv in Hcond0. tac.
        exploit _blocks_disjoint0;
        [apply Hblk1|idtac|idtac|apply Hcond1|idtac|idtac|idtac|idtac];
        eauto.
        * rewrite H6. eauto.
        * rewrite H5. auto.
      + eapply _blocks_disjoint0; eauto.
    - unfold blocks. rewrite PTree.gsspec.
      destruct (peq 1 i).
      + destruct _address_one0 as [blk [Hblk Haddr]].
        eexists. split; [eauto|].
        apply Block.store_inv in Hcond0. des. congruence.
      + apply _address_one0; auto.
  Qed.
  Lemma store_inv (ptr:ValueC.t) (val:ValueC.t) (m1 m2:t)
             (H: store ptr val m1 = Some m2):
    exists b o, val2ptr ptr m1 = Some (b, o) /\
    exists blk, PTree.get b m1 = Some blk /\
    exists newblk, Block.store o val blk = Some newblk /\
    m2 = mk m1.(nextb) (PTree.set b newblk m1).
  Proof.
    unfold store, store_aux in H.
    destruct (val2ptr ptr m1) as [[b o]|]; [|done].
    exists b. exists o. split; [done|].
    destruct (m1 ! b) as [blk|]; [|done].
    exists blk. split; [done|].
    destruct (Block.store o val blk) as [newblk|]; [|done].
    exists newblk. split; [done|].
    by inv H.
  Qed.
  Lemma store_le
        ptr val m1 m2 (H: store ptr val m1 = Some m2):
    le m1 m2.
  Proof.
    apply store_inv in H. tac.
    apply Block.store_inv in H1. tac.
    repeat intro. tac.
    destruct (peq b x); tac.
    - eexists. repeat (split; [by eauto|]).
      rewrite H6. tac.
    - eexists. repeat (split; [by eauto|]).
      tac.
  Qed.

  (** allocation is *deterministic* using `nextb`. We believe this
      simplifies metatheory.
   *)
  Definition alloc (size:int) (m:t): ident * t :=
    let newblk := Block.init size in
    let newblocks := PTree.set m.(nextb) newblk m in
    (m.(nextb), mk (Pos.succ m.(nextb)) newblocks).
  (* Lemma alloc_wf *)
  (*       (size:int) (m1:t) (Hwf1: wf m1) *)
  (*       (b:ident) (m2:t) *)
  (*       (H: alloc size m1 = (b, m2)): *)
  (*   wf m2. *)
  (* Proof. *)
  (*   unfold alloc in H. tac. inv Hwf1. *)
  (*   constructor; intros. *)
  (*   - tac. destruct (peq b (nextb m1)); [subst b|]. *)
  (*     + apply Plt_succ. *)
  (*     + apply Plt_trans_succ. *)
  (*       eapply _nextb0. eauto. *)
  (*   - tac. match goal with *)
  (*       | [H: context[if ?cond then _ else _] |- _] => *)
  (*         let Hcond := fresh "Hcond" in *)
  (*         destruct cond eqn:Hcond *)
  (*     end; tac. *)
  (*     eapply _blocks_wf0. eauto. *)
  (*   - tac. *)
  (*     (destruct (peq b1 (nextb m1)); [subst b1|]; tac); *)
  (*     (destruct (peq b2 (nextb m1)); [subst b2|]; tac). *)
  (*     eapply _blocks_disjoint0; eauto. *)
  (*   - unfold blocks. rewrite PTree.gsspec. *)
  (*     destruct (peq 1 (nextb m1)). *)
  (*     + destruct _address_one0 as [blk [Hblk Haddr]]. *)
  (*       eexists. split; [eauto|]. *)
  (*       exploit _nextb0; eauto. by rewrite <- e. *)
  (*     + apply _address_one0; auto. *)
  (* Qed. *)
  Lemma alloc_inv size m1 b m2
        (H: alloc size m1 = (b, m2)):
    b = m1.(nextb) /\
    m2.(nextb) = Pos.succ m1.(nextb) /\
    m2.(blocks) = PTree.set m1.(nextb) (Block.init size) m1.(blocks).
  Proof. by inv H. Qed.
  Lemma alloc_le
        size m1 b m2
        (Hwf: wf m1)
        (H: alloc size m1 = (b, m2)):
    le m1 m2.
  Proof.
    apply alloc_inv in H. tac. rewrite H0.
    repeat intro. tac.
    destruct (peq b (nextb m1)); tac.
    - inv Hwf. exploit _nextb0; eauto.
      intro X. by apply Plt_ne in X.
    - eexists. repeat (split; [by eauto|]).
      tac.
  Qed.

  Definition free_aux (b:ident) (m:PTree.t Block.t): option (int * PTree.t Block.t) :=
    doT blk <- PTree.get b m;
    assertB (blk.(Block.valid));
    Some (blk.(Block.size),
          PTree.set
            b
            (Block.mk
               (negb blk.(Block.accessible))
               blk.(Block.accessible)
               blk.(Block.size)
               blk.(Block.address)
               blk.(Block.content))
            m).
  Definition free (ptr:ValueC.t) (m:t): option (int * t) :=
    match val2ptr ptr m with
      | None => None
      | Some (b, o) =>
        if Int.cmp Ceq o Int.zero
        then option_map (fun r => (fst r, mk m.(nextb) (snd r))) (free_aux b m)
        else None
    end.
  Lemma free_wf
        (val:ValueC.t) (m1:t) (Hwf1: wf m1)
        result m2 (H: free val m1 = Some (result, m2)):
    wf m2.
  Proof.
    unfold free, free_aux in H. tac.
    inv Hwf1.
    constructor; intros.
    - tac. destruct (peq b i); [subst|].
      + eapply _nextb0. eauto.
      + eapply _nextb0. eauto.
    - tac. destruct (peq b i); [subst|].
      + inv Hblk.
        exploit _blocks_wf0; eauto.
        intro X. inv X.
        constructor; auto.
      + eapply _blocks_wf0. eauto.
    - tac.
      destruct (peq b1 i); inv Hblk1;
      destruct (peq b2 i); inv Hblk2;
      subst; simpl in *;
      (try by exploit _blocks_disjoint0; eauto).
      + exploit _blocks_disjoint0; [apply Hcond2 | | | | | | |]; eauto.
      + exploit _blocks_disjoint0; [apply Hcond2 | | | | | | |]; eauto.
      + exploit _blocks_disjoint0; [apply H2 | | | | | | |]; eauto.
    - unfold blocks. rewrite PTree.gsspec.
      destruct (peq 1 i).
      + destruct _address_one0 as [blk [Hblk Haddr]].
        eexists. split; [eauto|].
        subst. simpl. congruence.
      + apply _address_one0; auto.
  Qed.
  Definition free_inv (ptr:ValueC.t) (m1:t) size m2
             (H: free ptr m1 = Some (size, m2)):
    exists b, val2ptr ptr m1 = Some (b, Int.zero) /\
    exists blk, PTree.get b m1 = Some blk /\
    blk.(Block.valid) /\
    size = blk.(Block.size) /\
    m2 = mk m1.(nextb) (PTree.set b
                                  (Block.mk
                                     (negb blk.(Block.accessible))
                                     blk.(Block.accessible)
                                     blk.(Block.size)
                                     blk.(Block.address)
                                     blk.(Block.content))
                                  m1).
  Proof.
    unfold free, free_aux in H.
    destruct (val2ptr ptr m1) as [[b o]|]; [|done].
    destruct (Int.cmp Ceq o Int.zero) eqn:Hcmp; [|done].
    simpl in Hcmp.
    generalize (Int.eq_spec o Int.zero). destruct (Int.eq o Int.zero); [|done]. intro. subst.
    exists b. split; [done|].
    destruct (m1 ! b) as [blk|]; [|done]. inv H. tac.
    exists blk. by split.
  Qed.
  Lemma free_le
        ptr m1 size m2 (H: free ptr m1 = Some (size, m2)):
    le m1 m2.
  Proof.
    apply free_inv in H. destruct H; tac.
    repeat intro. tac.
    destruct (peq b x); tac.
    - eexists. repeat (split; [by eauto|]).
      tac.
    - eexists. repeat (split; [by eauto|]).
      tac.
  Qed.

  Inductive realize (m1:t) (b:ident) (addr:int) (m2:t): Prop :=
  | realize_
      blk1 (Hblk1: PTree.get b m1 = Some blk1)
      blk2 (Hblk2: Block.realize blk1 addr blk2)
      (Hm2: mk m1.(nextb) (PTree.set b blk2 m1) = m2)
      (Hwf2: wf m2):
      realize m1 b addr m2
  .
  Lemma realize_le
        m1 b addr m2 (H: realize m1 b addr m2):
    le m1 m2.
  Proof.
    inv H. inv Hblk2. repeat intro. tac.
    destruct (peq b0 b); tac.
    - eexists. repeat (split; [by eauto|]).
      rewrite Haddress. tac.
    - eexists. repeat (split; [by eauto|]).
      tac.
  Qed.

  Inductive allocC (size:int) (m1:t): forall (result:ident * int * t), Prop :=
  | allocC_intro
      m2 b addr m3
      (Halloc: alloc size m1 = (b, m2))
      (Hrealize: realize m2 b addr m3)
      (Hwf: wf m3):
      allocC size m1 (b, addr, m3)
  .
  Lemma allocC_le size m1 b addr m2
        (Hwf1: wf m1)
        (H: allocC size m1 (b, addr, m2)):
    le m1 m2.
  Proof.
    inv H. inv Hrealize. inv Hblk2. simpl in *. repeat intro. tac.
    apply alloc_inv in Halloc. des. subst.
    destruct (peq b0 (nextb m1)); tac.
    - exploit _nextb; try apply Hblk0; eauto. xomega.
    - eexists. rewrite H1. rewrite PTree.gso; auto.
      repeat (split; eauto). repeat intro. auto.
  Qed.
End MemC.
Coercion MemC.blocks: MemC.t >-> PTree.t.

Ltac memctac_aux :=
  match goal with
    | [H1: MemC.store ?p ?v ?m = ?r1,
       H2: MemC.store ?p ?v ?m = ?r2 |- _] =>
      rewrite H1 in H2; inv H2
  end.
