Require Import RelationClasses.
Require Import Axioms.
Require Import Coqlib.
Require Import Kanglib.
Require Import paco.
Require Import Integers.
Require Import Maps.
Require Import Monad.
Require Import AST.
Require Import Values.
Require Import Memory.
Require Import MemoryC.

Ltac tac :=
  repeat
    (simpl in *; subst;
     (try done);
     (try basictac_aux);
     (try destac_aux);
     (try monadtac_aux);
     (try valuetac_aux);
     (try maptac_aux)).

Module MemEquivC.
  (** Memory equivalence relation.
      This is basically equivalent to CompCert's memory injection.
      Well-formed `t` is the partial bijection between identifiers.
   *)
  Structure t := mk {
    src_tgt: ident -> option ident;
    tgt_src: ident -> option ident
  }.

  Definition empty: t :=
    mk (fun _ => None) (fun _ => None).

  Definition wf (rel:t): Prop :=
    forall s t, rel.(src_tgt) s = Some t <-> rel.(tgt_src) t = Some s.

  Lemma empty_wf: wf empty.
  Proof.
    repeat intro. constructor; intro; inv H.
  Qed.

  Definition to_relation (rel:t) (s t:ident):Prop :=
    rel.(src_tgt) s = Some t.
  Coercion to_relation: t >-> Funclass.

  Lemma src_unique s t1 t2 (rel:t): rel s t1 -> rel s t2 -> t1 = t2.
  Proof.
    unfold to_relation. intros.
    rewrite H in H0. by inv H0.
  Qed.    

  Lemma tgt_unique s1 s2 tgt (rel:t) (Hwf:wf rel): rel s1 tgt -> rel s2 tgt -> s1 = s2.
  Proof.
    unfold to_relation. intros.
    apply Hwf in H. apply Hwf in H0.
    rewrite H in H0. by inv H0.
  Qed.    

  Definition add (src tgt:ident) (rel:t): t :=
    mk
      (fun s => if peq s src then Some tgt else rel.(src_tgt) s)
      (fun t => if peq t tgt then Some src else rel.(tgt_src) t).

  Lemma add_wf src tgt (rel:t)
        (Hsrc: ~ exists t, rel src t)
        (Htgt: ~ exists s, rel s tgt)
        (Hrel: wf rel):
    wf (add src tgt rel).
  Proof.
    unfold add, wf in *. simpl. intros.
    destruct (peq s src), (peq t0 tgt); tac.
    - constructor; intro; [by inv H|].
      apply Hrel in H. contradict Hsrc. by eexists t0.
    - constructor; intro; [|by inv H].
      contradict Htgt. by exists s.
  Qed.

  (*** Equivalence predicates. It would be helpful to read top-down
       (i.e., bottom-up in Coq file.) *)

  (** equivalence between normalized values. 

      Two normalized values are euquivalent if 
      (i)  they are the same integers, or
      (ii) they are basically the normalized values of two corresponding blocks.
           However, for simplicity, we require that the source address should be logical
           because otherwise both the source and target addresses are integers
           in the quasi-conrete model.
  *)
  Inductive rel_value (e:t) (sm:PTree.t Mem.Block.t) (tm:PTree.t MemC.Block.t): forall (s:Value.t) (t:ValueC.t), Prop :=
  | rel_value_i i:
      rel_value e sm tm (Value.int i) (ValueC.int i)
  | rel_value_p
      si ti (Hi: e si ti) o
      tblk (Htblk: tm ! ti = Some tblk)
      taddr (Htaddr: tblk.(MemC.Block.address) = Some taddr):
      rel_value e sm tm (Value.ptr si o) (Int.add taddr o)
  .

  (** equivalence between blocks.
      + validity and size are equal.
      + content values are related.
   *)
  Inductive rel_block (e:t) (sm:PTree.t Mem.Block.t) (tm:PTree.t MemC.Block.t): forall (so:option Mem.Block.t) (to:option MemC.Block.t), Prop :=
  | rel_block_
      s t
      (Hvalid: s.(Mem.Block.valid) = t.(MemC.Block.valid))
      (Haccessible: s.(Mem.Block.accessible) = t.(MemC.Block.accessible))
      (Hsize: s.(Mem.Block.size) = t.(MemC.Block.size))
      (Hcontent:
         forall
           (Hvalid: s.(Mem.Block.valid))
           i (Hi: Range.contained i (Range.mk Int.zero s.(Mem.Block.size))),
           rel_value
             e sm tm
             (ZMap.get (Int.unsigned i) s.(Mem.Block.content))
             (ZMap.get (Int.unsigned i) t.(MemC.Block.content))):
      rel_block e sm tm (Some s) (Some t)
  .

  (** Source block's address <= Target block's address.  Particularly,
      if a source block is realized at `p`, than the corresponding
      target block should also be realized at `p`.  *)
  Definition le_address (rel:t) (m_src:PTree.t Mem.Block.t) (m_tgt:PTree.t MemC.Block.t): Prop :=
    forall b_src blk_src (Hblk: PTree.get b_src m_src = Some blk_src)
           p (Hp: blk_src.(Mem.Block.address) = Some p),
    exists b_tgt,
      rel b_src b_tgt /\
      exists blk_tgt,
        PTree.get b_tgt m_tgt = Some blk_tgt /\
        blk_tgt.(MemC.Block.address) = Some p.

  (** equivalence between memory: all blocks are related.  *)
  Inductive sem (e:t) (s:Mem.t) (t:MemC.t): Prop :=
  | sem_ 
      (Hone: e 1%positive 1%positive)
      (Haddress: le_address e s t)
      (Hpublic: forall si ti (Hi: e si ti),
         rel_block e s t (PTree.get si s) (PTree.get ti t))
  .

  (* If block identifiers related, it should be less then the
     `nextb`. *)
  Lemma nextb
        rel (m_src:Mem.t) (m_tgt:MemC.t) m_src_public m_tgt_public
        (Hpublic: sem rel m_src_public m_tgt_public)
        (Hsrc_public: PTree.subset m_src_public m_src)
        (Htgt_public: PTree.subset m_tgt_public m_tgt)
        (Hsrc_wf: Mem.wf m_src)
        (Htgt_wf: MemC.wf m_tgt)
        b_src b_tgt (Hrel: rel b_src b_tgt):
    (b_src < m_src.(Mem.nextb) /\ b_tgt < m_tgt.(MemC.nextb))%positive.
  Proof.
    destruct Hpublic. exploit Hpublic; eauto. intro X. inv X.
    symmetry in H0, H.
    apply Hsrc_public in H0.
    apply Htgt_public in H.
    inv Hsrc_wf. exploit _nextb; eauto.
    inv Htgt_wf. exploit _nextb0; eauto.
  Qed.
End MemEquivC.
Coercion MemEquivC.to_relation: MemEquivC.t >-> Funclass.

Module MemRelC.
  (** - Memory binary relation consists of:
        + `equiv`: memory injection
        + `src_private`: private memory for source
        + `tgt_private`: private memory for target.

      - We assume that memory is split into public, private, and
        garbage part. Public parts are related by `equiv`, and private
        parts are tracked by `*_private`. We believe this is enough
        for proving interesting examples.

      - Here, `[*]` are tracked in the memory relation.
        memory_src = src_public + [src_private] + src_garbage
                        [~]
        memory_tgt = tgt_public + [tgt_private] + tgt_garbage
   *)
  Structure t := mk {
    equiv: MemEquivC.t;
    src_private: PTree.t Mem.Block.t;
    tgt_private: PTree.t MemC.Block.t;
    src: Mem.t;
    tgt: MemC.t
  }.

  (** Semantics of memory relation.
    - `*_garbage` is not modeled. We are only interested in public parts
       and private parts.
    - `Haddress`: If a source block has an address `p`:
      + it should have a corresponding target block via `equiv` and
      + the target block's address is also `p`.
    - `Hpublic`: the public parts are related.
   *)
  Inductive sem (rel:t) (m_src:Mem.t) (m_tgt:MemC.t): Prop :=
  | sem_
      (Hequiv_wf: MemEquivC.wf rel.(equiv))
      (Hsrc_wf: Mem.wf m_src)
      (Htgt_wf: MemC.wf m_tgt)
      (Hsrc_private: forall b_src, rel.(equiv).(MemEquivC.src_tgt) b_src <> None -> PTree.get b_src rel.(src_private) = None)
      (Htgt_private: forall b_tgt, rel.(equiv).(MemEquivC.tgt_src) b_tgt <> None -> PTree.get b_tgt rel.(tgt_private) = None)
      (Hsrc_private_subset: PTree.subset rel.(src_private) m_src)
      (Htgt_private_subset: PTree.subset rel.(tgt_private) m_tgt)
      (Hpublic: MemEquivC.sem rel.(equiv) m_src m_tgt)
      (Hsrc: rel.(src) = m_src)
      (Htgt: rel.(tgt) = m_tgt)
  .
  Coercion sem: t >-> Funclass.

  (** memory relation order for simulation proof. We require only
      increment of memory injection.  *)
  Inductive le (mrel1 mrel2:t): Prop :=
  | le_
      (Hmbinrel: mrel1.(equiv) <2= mrel2.(equiv))
      (Hsrc: Mem.le mrel1.(src) mrel2.(src))
      (Htgt: MemC.le mrel1.(tgt) mrel2.(tgt))
  .

  Global Program Instance le_PreOrder : PreOrder le.
  Next Obligation.
    constructor; eauto; reflexivity.
  Qed.
  Next Obligation.
    repeat intro. inv H. inv H0.
    constructor; eauto; etransitivity; eauto.
  Qed.

  (** public memory relation order for simulation proof. We require:
     - increment of memory injection;
     - privates are preserved, i.e., private memories before a
     function call should be preserved after the function's return.
   *)
  Inductive le_public (rel1 rel2:t):Prop :=
  | le_public_
      (Hequiv: le rel1 rel2)
      (Hsrc_private: PTree.subset rel1.(src_private) rel2.(src_private))
      (Htgt_private: PTree.subset rel1.(tgt_private) rel2.(tgt_private))
  .

  Global Program Instance le_public_PreOrder : PreOrder le_public.
  Next Obligation.
    constructor; tac; reflexivity.
  Qed.
  Next Obligation.
    repeat intro. inv H. inv H0.
    constructor; repeat intro.
    - etransitivity; eauto.
    - apply Hsrc_private0. apply Hsrc_private. tac.
    - apply Htgt_private0. apply Htgt_private. tac.
  Qed.

  Lemma le_public_le: le_public <2= le.
  Proof.
    intros a b H. by inv H.
  Qed.

  Ltac tac2 :=
    (repeat
       match goal with
         | [H: Some _ = PTree.get _ _ |- _] => symmetry in H
       end);
    repeat
      (intros; tac;
       try match goal with
             | [Hptr: MemEquivC.rel_value _ _ _ (ValueC.int _) _|- _] =>
               inv Hptr
             | [H: ?a.(Mem.Block.address) = _ |- context[?a.(Mem.Block.address)]] =>
               rewrite H
             | [H: ?a.(MemC.Block.address) = _ |- context[?a.(MemC.Block.address)]] =>
               rewrite H
           end).

  Lemma address
        (rel:t) m_src m_tgt (Hrel: rel m_src m_tgt):
    MemEquivC.le_address rel.(equiv) m_src m_tgt.
  Proof.
    inv Hrel. destruct Hpublic. apply Haddress.
  Qed.

  Ltac tac3 :=
    repeat
      (tac2;
       try match goal with
             | [H: MemC.val2ptr _ _ = Some (_, _) |- _] =>
               apply MemC.val2ptr_inv in H
             | [Hwf: sem ?rel ?m_src ?m_tgt,
               Hblk: PTree.get ?b ?m_src.(MemC.blocks) = Some ?blk,
               Haddr: ?blk.(MemC.Block.address) = Some _ |- _] =>
               exploit (address (rel:t) m_src m_tgt Hwf b); eauto; symmetry in Haddr
           end).

  Lemma mem_double_contained
        b1 b2
        (m:MemC.t) (Hm: MemC.wf m)
        blk1 (Hblk1: PTree.get b1 m = Some blk1)
        blk2 (Hblk2: PTree.get b2 m = Some blk2)
        (Hvalid1: blk1.(MemC.Block.valid))
        (Hvalid2: blk2.(MemC.Block.valid))
        addr1 (Haddr1: blk1.(MemC.Block.address) = Some addr1)
        addr2 (Haddr2: blk2.(MemC.Block.address) = Some addr2)
        i
        (Hi1: Range.contained i (Range.mk addr1 blk1.(MemC.Block.size)))
        (Hi2: Range.contained i (Range.mk addr2 blk2.(MemC.Block.size))):
    b1 = b2.
  Proof.
    destruct (peq b1 b2); [done|].
    inv Hm.
    exploit _blocks_disjoint.
    - apply Hblk1. - auto. - eauto.
    - apply Hblk2. - auto. - eauto.
    - done.
    - intro Hdisj.
      by exploit Range.gdspec; [| |apply Hdisj|]; eauto.
  Qed.

  Lemma val2int_public
        v_src v_tgt mem_src mem_tgt equiv
        (Hv: MemEquivC.rel_value equiv mem_src mem_tgt v_src v_tgt)
        i (Hv_src: Mem.val2int v_src mem_src = Some i):
    MemC.val2int v_tgt mem_tgt = Some i.
  Proof.
    unfold Mem.val2int, MemC.val2int in *. inv Hv; inv Hv_src. auto.
  Qed.

  Lemma val2ptr_public_aux
        ptr_src ptr_tgt
        (rel:t) m_src m_tgt
        (Hrel: rel m_src m_tgt)
        (Hptr: MemEquivC.rel_value rel.(equiv) m_src m_tgt ptr_src ptr_tgt)
        b_src o (Hptr_src: Mem.val2ptr ptr_src m_src = Some (b_src, o))
        blk_src (Hblk_src: PTree.get b_src m_src = Some blk_src)
        (Hvalid: blk_src.(Mem.Block.valid))
        (Ho: Range.contained o (Range.mk Int.zero blk_src.(Mem.Block.size))):
    exists b_tgt,
      MemC.val2ptr ptr_tgt m_tgt = Some (b_tgt, o) /\
      rel.(equiv) b_src b_tgt.
  Proof.
    tac3.
    destruct ptr_tgt. inv Hptr; inv Hptr_src.
    unfold MemC.val2ptr. tac3.
    match goal with
      | [|- context [PTree.select ?pred ?m]] =>
        generalize (PTree.gselectspec pred m);
          destruct (PTree.select pred m)
    end; tac3.
    - inv Hrel. destruct Hpublic. exploit Hpublic; try apply Hi; eauto.
      intro X. inv X. tac3.
      rewrite Hvalid0, Hsize in *.
      exploit (mem_double_contained ti p); eauto.
      { rewrite Int.add_commut.
        apply Range.contained_trans; auto.
        inv Htgt_wf. exploit _blocks_wf; eauto.
        intro X. inv X.
        apply _size. eauto.
      }
      intro. subst. tac3.
      repeat eexists; eauto.
      rewrite Hcond in *. inv Htaddr.
      by rewrite Int.sub_add_l, Int.sub_idem, Int.add_zero_l.
    - contradict H0. eexists. eexists. split; eauto.
      rewrite Htaddr. simpl.
      inv Hrel. destruct Hpublic. exploit Hpublic; try apply Hi; eauto.
      intro X. inv X. tac3.
      rewrite <- Hvalid0, H. simpl.
      rewrite Int.add_commut. erewrite Range.contained_trans.
      + congruence.
      + by rewrite <- Hsize.
      + inv Htgt_wf. eapply _blocks_wf; eauto.
  Qed.

  Lemma val2ptr_public
        ptr_src ptr_tgt
        (rel:t) m_src m_tgt
        (Hrel: rel m_src m_tgt)
        (Hptr: MemEquivC.rel_value rel.(equiv) m_src m_tgt ptr_src ptr_tgt)
        b_src o (Hptr_src: Mem.val2ptr ptr_src m_src = Some (b_src, o)):
    exists b_tgt,
      MemC.val2ptr ptr_tgt m_tgt = Some (b_tgt, o) /\
      rel.(equiv) b_src b_tgt.
  Proof.
    inv Hptr; [inv Hptr_src|].
    unfold Mem.val2ptr in Hptr_src.
    destruct (m_src ! si) eqn:X; inv Hptr_src.
    destruct (Mem.Block.valid t0) eqn:Y; inv H0.
    destruct (Range.contained o0 {| Range.begin := 0; Range.size := Mem.Block.size t0 |}) eqn:Z; inv H1.
    eapply val2ptr_public_aux; eauto.
    - eapply MemEquivC.rel_value_p; eauto.
    - unfold Mem.val2ptr. rewrite X. simpl.
      rewrite Y, Z. auto.
  Qed.

  Lemma incr_rel_value_aux
        (equiv1 equiv2:MemEquivC.t) (src1 src2:Mem.t) (tgt1 tgt2:MemC.t) v_src v_tgt
        (Hequiv: equiv1 <2= equiv2)
        (Hsrc: Mem.le src1 src2)
        (Htgt: MemC.le tgt1 tgt2)
        (H2: MemEquivC.le_address equiv2 src2 tgt2)
        (Hrel: MemEquivC.rel_value equiv1 src1 tgt1 v_src v_tgt):
    MemEquivC.rel_value equiv2 src2 tgt2 v_src v_tgt.
  Proof.
    inv Hrel; try (constructor; auto; fail).
    unfold Mem.normalize.
    exploit Htgt; eauto. intros [blk2_tgt [Hblk2_tgt [Hvalid_tgt [Hsize_tgt Haddress]]]].
    exploit Haddress; eauto. intro Haddr2_tgt.
    econstructor; eauto.
  Qed.

  Lemma incr_rel_value
        (mrel1 mrel2:MemRelC.t) (src1 src2:Mem.t) (tgt1 tgt2:MemC.t) v_src v_tgt
        (Hle: MemRelC.le mrel1 mrel2)
        (Hmrel1: mrel1 src1 tgt1)
        (Hmrel2: mrel2 src2 tgt2)
        (Hrel_value1: MemEquivC.rel_value mrel1.(equiv) src1 tgt1 v_src v_tgt):
    MemEquivC.rel_value mrel2.(equiv) src2 tgt2 v_src v_tgt.
  Proof.
    inversion Hle. inversion Hmrel1. inversion Hmrel2. destruct Hpublic, Hpublic0.
    eapply incr_rel_value_aux; eauto; tac.
  Qed.

  (*** Lemmas on memory relation w.r.t. basic memory operations.

       For example:
       - `load_src`: if a value is loaded from src_private, then the
         value is also loaded memory_src.
       - `store_tgt`: if a value is successfully stored in
         tgt_private, then the value is also successfully stored in
         memory_tgt.
       - `free_public`: if a pointer is successfully freed in
         memory_src, then the corresponding (w.r.t. memory injection)
         pointer is also successfully freed in memory_tgt.

       Note that for `store`, `malloc`, and `free`, there is
       assymmetry between source and target. For src, we prove that
       [operation on memory_src succeeds] => [operation on src_private
       succeeds], while for tgt, we prove that [operation on
       tgt_private succeeds] => [operation on memory_tgt succeeds].
   *)

  Lemma load_src
        b_src
        (rel:t) m_src m_tgt
        (Hrel: rel m_src m_tgt)
        ptr_src offset_src (Hptr_src: Mem.val2ptr ptr_src m_src = Some (b_src, offset_src))
        blk_src (Hb_src: PTree.get b_src rel.(src_private) = Some blk_src):
    Mem.load ptr_src m_src = Mem.load_aux b_src offset_src rel.(src_private).
  Proof.
    unfold Mem.load. rewrite Hptr_src.
    unfold Mem.load_aux. rewrite Hb_src.
    inv Hrel. eapply Hsrc_private_subset in Hb_src; eauto.
    by rewrite Hb_src.
  Qed.

  Lemma load_tgt
        b_tgt
        (rel:t) m_src m_tgt
        (Hrel: rel m_src m_tgt)
        ptr_tgt offset_tgt (Hptr_tgt: MemC.val2ptr ptr_tgt m_tgt = Some (b_tgt, offset_tgt))
        blk_tgt (Hb_tgt: PTree.get b_tgt rel.(tgt_private) = Some blk_tgt):
    MemC.load ptr_tgt m_tgt = MemC.load_aux b_tgt offset_tgt rel.(tgt_private).
  Proof.
    unfold MemC.load. rewrite Hptr_tgt.
    unfold MemC.load_aux. rewrite Hb_tgt.
    inv Hrel. eapply Htgt_private_subset in Hb_tgt; eauto.
    by rewrite Hb_tgt.
  Qed.

  Lemma load_public
        (rel:t) m_src m_tgt
        (Hrel: rel m_src m_tgt)
        ptr_src ptr_tgt (Hptr: MemEquivC.rel_value rel.(equiv) m_src m_tgt ptr_src ptr_tgt)
        val_src (Hval_src: Mem.load ptr_src m_src = Some val_src):
    exists val_tgt,
      MemC.load ptr_tgt m_tgt = Some val_tgt /\
      MemEquivC.rel_value rel.(equiv) m_src m_tgt val_src val_tgt.
  Proof.
    apply Mem.load_inv in Hval_src. tac2.
    apply Mem.Block.load_inv in H1. tac2.
    exploit val2ptr_public; eauto.
    intros [b_tgt [Hptr_tgt Hb_tgt]].
    unfold MemC.load. rewrite Hptr_tgt.
    inversion Hrel. destruct Hpublic. exploit Hpublic; try apply Hb_tgt; eauto.
    intro X. inv X. tac2.
    unfold MemC.load_aux. rewrite H6.
    unfold MemC.Block.load.
    rewrite <- Hvalid, H1.
    rewrite <- Haccessible, H2.
    rewrite <- Hsize, H3.
    eexists. split; [done|].
    exploit Hcontent; eauto.
  Qed.

  Lemma store_src
        b_src
        (rel:t) m_src m_tgt
        (Hrel: rel m_src m_tgt)
        ptr_src offset_src (Hptr_src: Mem.val2ptr ptr_src m_src = Some (b_src, offset_src))
        val_src
        blk_src (Hb_src: PTree.get b_src rel.(src_private) = Some blk_src)
        m2_src (Hm2_src: Mem.store ptr_src val_src m_src = Some m2_src):
    exists src2_private,
      Mem.store_aux b_src offset_src val_src rel.(src_private) = Some src2_private /\
      le rel (mk rel.(equiv) src2_private rel.(tgt_private) m2_src m_tgt) /\
      (mk rel.(equiv) src2_private rel.(tgt_private) m2_src m_tgt) m2_src m_tgt.
  Proof.
    inv Hrel. destruct Hpublic.
    generalize Hm2_src. intro X.
    unfold Mem.store in X. rewrite Hptr_src in X.
    unfold Mem.store_aux in *. tac2.
    rewrite Hb_src, Hcond0. simpl.
    eexists. split; [done|].
    split.
    { constructor; eauto.
      + eapply Mem.store_le; eauto.
      + reflexivity. }
    constructor; auto; [| | |constructor].
    - by eapply Mem.store_wf; [|by eauto].
    - simpl. intros. exploit Hsrc_private; eauto.
      rewrite PTree.gsspec. destruct (peq b_src0 b_src); [subst|done].
      by rewrite Hb_src.
    - simpl. repeat intro. rewrite ? PTree.gsspec in *.
      destruct (peq i b_src); [done|].
      by apply Hsrc_private_subset.
    - auto.
    - repeat intro. tac.
      apply Mem.Block.store_inv in Hcond0. tac2.
      destruct (peq b_src0 b_src); [subst|].
      + inv Hblk. eapply Haddress; eauto.
        rewrite H4. eauto.
      + eapply Haddress; eauto.
    - repeat intro. tac.
      exploit Hpublic; eauto. intro X. inv X. tac2.
      destruct (peq si b_src); [subst|].
      { exploit Hsrc_private; eauto.
        - instantiate (1 := b_src). inv Hi. by rewrite H2.
        - intro X. rewrite Hb_src in X. inv X.
      }
      constructor; auto. intros.
      match goal with
        | [|- MemEquivC.rel_value _ ?blocks _ _ _] =>
          erewrite (Mem.blocks_unsimpl blocks)
      end.
      eapply incr_rel_value_aux; eauto.
      { eapply Mem.store_le; eauto. }
      { reflexivity. }
      { repeat intro. tac.
        destruct (peq b_src0 b_src); [subst|].
        - inv Hblk. apply Mem.Block.store_inv in Hcond0. tac.
          exploit Haddress.
          + apply Hsrc_private_subset. eauto.
          + by rewrite H6, Hp.
          + tac.
        - exploit Haddress; eauto.
      }
  Qed.

  Lemma store_tgt
        b_tgt
        (rel:t) m_src m_tgt
        (Hrel: rel m_src m_tgt)
        ptr_tgt offset_tgt (Hptr_tgt: MemC.val2ptr ptr_tgt m_tgt = Some (b_tgt, offset_tgt))
        val_tgt
        blk_tgt (Hb_tgt: PTree.get b_tgt rel.(tgt_private) = Some blk_tgt)
        tgt2_private (Hm2_tgt_private: MemC.store_aux b_tgt offset_tgt val_tgt rel.(tgt_private) = Some tgt2_private):
    exists m2_tgt,
      MemC.store ptr_tgt val_tgt m_tgt = Some m2_tgt /\
      le rel (mk rel.(equiv) rel.(src_private) tgt2_private m_src m2_tgt) /\
      (mk rel.(equiv) rel.(src_private) tgt2_private m_src m2_tgt) m_src m2_tgt.
  Proof.
    inv Hrel. destruct Hpublic.
    unfold MemC.store_aux in *. tac2.
    assert (Hm2_tgt: MemC.store ptr_tgt val_tgt (tgt rel) = Some (MemC.mk (tgt rel).(MemC.nextb) (PTree.set b_tgt t0 (tgt rel)))).
    { unfold MemC.store. rewrite Hptr_tgt.
      unfold MemC.store_aux. tac. by rewrite Hcond0.
    }
    eexists. split; [apply Hm2_tgt|].
    split.
    { constructor; eauto.
      + reflexivity.
      + eapply MemC.store_le; eauto. }
    apply MemC.Block.store_inv in Hcond0.
    destruct Hcond0 as [Hv1 [Hv2 [Ha1 [Ha2 [Hsz [Haddr2 [Hcontain Hcontent]]]]]]].
    constructor; auto; [| | |constructor].
    - eapply MemC.store_wf. apply Htgt_wf.
      unfold MemC.store. instantiate (1 := val_tgt).
      instantiate (1 := ptr_tgt). rewrite Hptr_tgt.
      unfold MemC.store_aux. tac2.
      unfold MemC.Block.store. rewrite Hv1, Hcontain. simpl.
      destruct t0; simpl in *.
      by rewrite Hv2, Ha1, Ha2, Hsz, Haddr2, Hcontent.
    - simpl. intros. exploit Htgt_private; eauto.
      rewrite PTree.gsspec. destruct (peq b_tgt0 b_tgt); [subst|done].
      by rewrite Hcond.
    - simpl. repeat intro. rewrite ? PTree.gsspec in *.
      destruct (peq i b_tgt); tac.
    - auto.
    - repeat intro. exploit Haddress; eauto. intro. tac.
      eexists. split; [by eauto|]. tac.
      destruct (peq x b_tgt); [subst|].
      + eexists. split; tac. by rewrite <- Haddr2.
      + rewrite H0. by eexists.
    - repeat intro. tac.
      exploit Hpublic; eauto. intro X. inv X. tac2.
      destruct (peq ti b_tgt); [subst|].
      { exploit Htgt_private; eauto.
        - instantiate (1 := b_tgt). inv Hi.
          apply Hequiv_wf in H2. by rewrite H2.
        - intro X. rewrite Hcond in X. inv X.
      }
      constructor; auto. intros.
      match goal with
        | [|- MemEquivC.rel_value _ _ ?blocks _ _] =>
          erewrite (MemC.blocks_unsimpl blocks)
      end.
      eapply incr_rel_value_aux; eauto.
      { reflexivity. }
      { eapply MemC.store_le; eauto. }
      { repeat intro. tac.
        exploit Haddress; eauto. intro. tac.
        eexists. split; [by eauto|]. tac.
        destruct (peq x b_tgt); [subst|].
        - eexists. split; tac. by rewrite <- Haddr2.
        - eexists. split; eauto.
      }
  Qed.

  Lemma store_public
        (rel:t) m_src m_tgt
        (Hrel: rel m_src m_tgt)
        ptr_src ptr_tgt (Hptr: MemEquivC.rel_value rel.(equiv) m_src m_tgt ptr_src ptr_tgt)
        val_src val_tgt (Hval: MemEquivC.rel_value rel.(equiv) m_src m_tgt val_src val_tgt)
        m2_src (Hm2_src: Mem.store ptr_src val_src m_src = Some m2_src):
    exists m2_tgt,
      MemC.store ptr_tgt val_tgt m_tgt = Some m2_tgt /\
      le rel (mk rel.(equiv) rel.(src_private) rel.(tgt_private) m2_src m2_tgt) /\
      (mk rel.(equiv) rel.(src_private) rel.(tgt_private) m2_src m2_tgt) m2_src m2_tgt.
  Proof.
    exploit Mem.store_inv; eauto. tac2.
    exploit Mem.Block.store_inv; eauto. tac2.
    exploit (val2ptr_public ptr_src ptr_tgt); eauto.
    intros [b_tgt [Hb_tgt Hb]].
    inv Hrel. destruct Hpublic. exploit Hpublic; try apply Hb; eauto.
    intro X. inv X. tac2.
    assert (Hm2_tgt_block: MemC.Block.store x0 val_tgt t0 = Some (MemC.Block.mk t0.(MemC.Block.valid) t0.(MemC.Block.accessible) t0.(MemC.Block.size) t0.(MemC.Block.address) (ZMap.set (Int.unsigned x0) val_tgt (MemC.Block.content t0)))).
    { unfold MemC.Block.store.
      by rewrite <- Hvalid, H2, <- Haccessible, H4, <- Hsize, H8.
    }
    assert (Hm2_tgt: MemC.store ptr_tgt val_tgt (tgt rel) = Some (MemC.mk (tgt rel).(MemC.nextb) (PTree.set b_tgt (MemC.Block.mk t0.(MemC.Block.valid) t0.(MemC.Block.accessible) t0.(MemC.Block.size) t0.(MemC.Block.address) (ZMap.set (Int.unsigned x0) val_tgt (MemC.Block.content t0))) (tgt rel)))).
    { unfold MemC.store. rewrite Hb_tgt.
      unfold MemC.store_aux. rewrite H12.
      by rewrite Hm2_tgt_block.
    }
    eexists. split; [by apply Hm2_tgt|].
    split.
    { constructor; eauto.
      + eapply Mem.store_le; eauto.
      + eapply MemC.store_le; eauto. }
    constructor; auto; [| | | |constructor].
    - eapply (Mem.store_wf ptr_src val_src (src rel)); eauto.
    - eapply (MemC.store_wf ptr_tgt val_tgt (tgt rel)); eauto.
    - simpl. repeat intro. tac2.
      destruct (peq i x); tac.
      exploit Hsrc_private; eauto.
      * instantiate (1 := x). by rewrite Hb.
      * by rewrite H10.
    - simpl. repeat intro. rewrite PTree.gsspec.
      destruct (peq i b_tgt); [subst|].
      + exploit Htgt_private; eauto.
        * apply Hequiv_wf in Hb.
          instantiate (1 := b_tgt). by rewrite Hb.
        * by rewrite H10.
      + exploit Hpublic; eauto.
    - auto.
    - repeat intro. tac.
      destruct (peq b_src x); tac.
      + exploit Haddress; eauto.
        * rewrite H7. eauto.
        * intro. tac. eexists. split; [by eauto|]. tac. rewrite H11, if_Some.
          eexists. split; tac.
          destruct (peq x1 b_tgt); tac.
      + exploit Haddress; eauto.
        intro. tac. eexists. split; [by eauto|]. tac. rewrite H11, if_Some.
        eexists. split; tac.
        destruct (peq x1 b_tgt); tac.
    - repeat intro. tac.
      exploit Hpublic; eauto. intro X. inv X. rewrite ? if_Some.
      assert (J: (si = x /\ ti = b_tgt) \/ (si <> x /\ ti <> b_tgt)).
      { destruct (peq si x); tac.
        - exploit (MemEquivC.src_unique x b_tgt ti); eauto.
        - destruct (peq ti b_tgt); tac.
          + exploit (MemEquivC.tgt_unique si x b_tgt); eauto.
          + by right.
      }
      constructor.
      + destruct J as [[? ?]|[? ?]]; tac.
        rewrite ? peq_true. by rewrite <- Hvalid, H2, H3.
      + destruct J as [[? ?]|[? ?]]; tac.
        rewrite ? peq_true. by rewrite <- Haccessible, H4.
      + destruct J as [[? ?]|[? ?]]; tac.
        rewrite ? peq_true. by rewrite <- Hsize, H6.
      + intros.
        match goal with
          | [|- MemEquivC.rel_value _ ?blocks_src ?blocks_tgt _ _] =>
            erewrite (Mem.blocks_unsimpl blocks_src);
            erewrite (MemC.blocks_unsimpl blocks_tgt)
        end.
        eapply incr_rel_value_aux; eauto.
        { eapply Mem.store_le; eauto. }
        { eapply MemC.store_le; eauto. }
        { repeat intro. tac.
          destruct (peq b_src x); [subst|].
          - inv Hblk.
            eexists. split; [by eauto|]. tac.
            eexists. split; [by eauto|]. tac.
            exploit Haddress; eauto.
            { by rewrite H7, Hp. }
            intro. tac.
            exploit (MemEquivC.src_unique x b_tgt x1); eauto.
            intro. tac.
          - exploit Haddress; eauto. intro. tac.
            eexists. split; [by eauto|]. tac. rewrite H14, if_Some.
            eexists. split; tac.
            destruct (peq x1 b_tgt); tac.
        }
        { destruct J as [[? ?]|[? ?]]; tac.
          - rewrite ? peq_true. simpl. rewrite <- H9. rewrite ? ZMap.gsspec.
            destruct (ZIndexed.eq (Int.unsigned i) (Int.unsigned x0)); eauto.
            apply Hcontent; eauto.
            by rewrite H6.
          - apply Hcontent0; eauto.
        }
  Qed.

  Lemma alloc_src
        (rel:t) m_src m_tgt
        (Hrel: rel m_src m_tgt)
        size
        b_src m2_src (Hm2_src: Mem.alloc size m_src = (b_src, m2_src)):
    le rel (mk rel.(equiv) (PTree.set b_src (Mem.Block.init size) rel.(src_private)) rel.(tgt_private) m2_src m_tgt) /\
    (mk rel.(equiv) (PTree.set b_src (Mem.Block.init size) rel.(src_private)) rel.(tgt_private) m2_src m_tgt)
      m2_src m_tgt.
  Proof.
    destruct (Mem.alloc_inv size m_src b_src m2_src Hm2_src) as [? [Hnextb_src Hblocks_src]]. subst.
    inv Hrel. destruct Hpublic.
    split.
    { constructor; eauto.
      + eapply Mem.alloc_le; eauto.
      + reflexivity. }
    econstructor; auto; [| | |constructor].
    - eapply Mem.alloc_wf; [|eauto]; eauto.
    - simpl. intros. rewrite PTree.gsspec.
      destruct (peq b_src (Mem.nextb (src rel))); [subst|].
      + destruct (MemEquivC.src_tgt rel.(equiv) (Mem.nextb (src rel))) eqn:Hnb; [|done].
        exploit Hpublic; eauto.
        intro X. inv X. tac2.
        inv Hsrc_wf. exploit _nextb; eauto.
        intro X. by apply Plt_ne in X.
      + by apply Hsrc_private.
    - simpl. repeat intro.
      rewrite Hblocks_src. rewrite PTree.gsspec in *.
      destruct (peq i (Mem.nextb (src rel))); [by subst|].
      tac2.
    - auto.
    - rewrite Hblocks_src. tac.
      repeat intro. tac.
      destruct (peq b_src (Mem.nextb (src rel))); [subst|].
      + inv Hblk. inv Hp.
      + exploit Haddress; eauto.
    - repeat intro. tac.
      exploit Hpublic; eauto. intro X. inv X. tac2.
      rewrite Hblocks_src. tac2. rewrite H0, if_Some.
      destruct (peq si (Mem.nextb (src rel))); [subst|].
      + exploit Hsrc_private; eauto.
        * inv Hi. instantiate (1 := (Mem.nextb (src rel))). by rewrite H2.
        * intro X. inv Hsrc_wf. exploit _nextb; eauto. intro Y. apply Plt_ne in Y.
          by contradict Y.
      + constructor; auto. intros.
        match goal with
          | [|- MemEquivC.rel_value _ ?blocks_src ?blocks_tgt _ _] =>
            erewrite (Mem.blocks_unsimpl blocks_src);
            erewrite (MemC.blocks_unsimpl blocks_tgt)
        end.
        eapply incr_rel_value_aux; eauto.
        { eapply (Mem.alloc_le size (src rel)); eauto.
          destruct m2_src. tac.
        }
        { reflexivity. }
        { repeat intro. tac.
          destruct (peq b_src (src rel).(Mem.nextb)); [subst|].
          - inv Hblk.
            eexists. split; [by eauto|]. tac.
          - exploit Haddress; eauto.
        }
  Grab Existential Variables.
    { auto. }
    { auto. }
  Qed.

  Lemma alloc_tgt
        (rel:t) m_src m_tgt
        (Hrel: rel m_src m_tgt)
        size
        b_tgt addr_tgt m2_tgt (Hm2_tgt: MemC.allocC size m_tgt (b_tgt, addr_tgt, m2_tgt)):
    let tgt'_private :=
        PTree.set
          b_tgt
          (MemC.Block.mk
             true
             true
             size
             (Some addr_tgt)
             (ZMap.init (ValueC.int Int.zero)))
          rel.(tgt_private)
    in
    le rel (mk rel.(equiv) rel.(src_private) tgt'_private m_src m2_tgt) /\
    (mk rel.(equiv) rel.(src_private) tgt'_private m_src m2_tgt)
      m_src m2_tgt.
  Proof.
    inv Hm2_tgt. inv Hrealize. inv Hblk2.
    destruct (MemC.alloc_inv size m_tgt b_tgt m2 Halloc) as [? [Hnextb_tgt Hblocks_tgt]]. subst.
    rewrite Hblocks_tgt in Hblk1. rewrite PTree.gss in Hblk1. inv Hblk1.
    inv Hrel. destruct Hpublic.
    split.
    { constructor; eauto.
      + reflexivity. 
      + simpl. repeat intro. exploit MemC.alloc_le; eauto.
        intros [blk2 [Hblk2 [X [Y Z]]]].
        rewrite PTree.gsspec. destruct (peq b (MemC.nextb (tgt rel))).
        * subst. eexists. split; eauto.
          rewrite Hblocks_tgt in Hblk2. rewrite PTree.gss in Hblk2. inv Hblk2.
          split; [done|]. split; [done|].
          repeat intro. apply Z in Hsrc. congruence.
        * eexists. eauto.
    }
    econstructor; simpl; auto; [| |constructor]; auto.
    - intros. rewrite PTree.gsspec. rewrite peq_false; auto.
      intro. subst.
      destruct (MemEquivC.tgt_src (equiv rel) (MemC.nextb (tgt rel))) as [b_src|] eqn:X; [|done].
      exploit (Hpublic b_src). apply Hequiv_wf. eauto.
      intro Y. inv Y. exploit MemC._nextb; eauto. xomega.
    - intros b blk. rewrite PTree.gsspec. destruct (peq b (MemC.nextb (tgt rel))).
      + intro X. inv X. by rewrite PTree.gss.
      + intro. rewrite PTree.gso; auto.
        rewrite Hblocks_tgt. rewrite PTree.gso; auto.
    - repeat intro. exploit Haddress0; eauto. intros [b_tgt [Hb_tgt [blk_tgt [Hblk_tgt Haddr]]]].
      exists b_tgt. split; auto. exists blk_tgt. split; auto.
      cut (b_tgt <> MemC.nextb (tgt rel)).
      { intro X. simpl. rewrite PTree.gso; auto.
        rewrite Hblocks_tgt. rewrite PTree.gso; auto.
      }
      intro. subst. exploit MemC._nextb; eauto. xomega.
    - repeat intro. tac.
      exploit Hpublic; eauto. intro X. inv X. tac2.
      rewrite Hblocks_tgt. tac2. rewrite H, if_Some.
      destruct (peq ti (MemC.nextb (tgt rel))); [subst|].
      + exploit Htgt_private; eauto.
        * inv Hi. instantiate (1 := (MemC.nextb (tgt rel))).
          apply Hequiv_wf in H2. by rewrite H2.
        * intro X. inv Htgt_wf. exploit _nextb; eauto. intro Y. apply Plt_ne in Y.
          by contradict Y.
      + constructor; auto. intros.
        match goal with
          | [|- MemEquivC.rel_value _ ?blocks_src ?blocks_tgt _ _] =>
            erewrite (Mem.blocks_unsimpl blocks_src);
            erewrite (MemC.blocks_unsimpl blocks_tgt)
        end.
        eapply incr_rel_value_aux; eauto.
        { reflexivity. }
        { etransitivity. eapply (MemC.alloc_le size (tgt rel)); eauto.
          rewrite Hblocks_tgt. repeat intro.
          rewrite PTree.gsspec in Hblk1. destruct (peq b (MemC.nextb (tgt rel))); tac2.
          - eexists. split; eauto.
            simpl. split; auto. split; auto.
            repeat intro. congruence.
          - eexists. split; eauto.
            simpl. split; auto. split; auto.
            repeat intro. congruence.
        }
        { repeat intro. tac.
          exploit Haddress0; eauto. intro. tac.
          exists x. split; tac.
          destruct (peq x (tgt rel).(MemC.nextb)); tac.
          - inv Htgt_wf. exploit _nextb; eauto. intro Y. apply Plt_ne in Y.
            by contradict Y.
          - rewrite H2. eexists. split; eauto.
        }
  Grab Existential Variables.
    { auto. }
    { auto. }
  Qed.

  Lemma free_src
        b_src
        (rel:t) m_src m_tgt
        (Hrel: rel m_src m_tgt)
        ptr_src (Hptr_src: Mem.val2ptr ptr_src m_src = Some (b_src, Int.zero))
        blk_src (Hb_src: PTree.get b_src rel.(src_private) = Some blk_src)
        sz m2_src (Hm2_src: Mem.free ptr_src m_src = Some (sz, m2_src)):
    exists src2_private, Mem.free_aux b_src rel.(src_private) = Some (sz, src2_private) /\
    le rel (mk rel.(equiv) src2_private rel.(tgt_private) m2_src m_tgt) /\
    (mk rel.(equiv) src2_private rel.(tgt_private) m2_src m_tgt) m2_src m_tgt.
  Proof.
    exploit Mem.free_inv; eauto. intro Hm2'_src.
    inv Hrel. destruct Hpublic. tac2.
    rewrite Hptr_src in H. inv H.
    unfold Mem.free_aux. tac2. rewrite Hb_src. simpl.
    rewrite H1. eexists. split; [done|].
    split.
    { constructor; eauto.
      + eapply Mem.free_le; eauto.
      + reflexivity. }
    constructor; auto; [| | |constructor].
    - eapply Mem.free_wf. apply Hsrc_wf.
      unfold Mem.free. instantiate (2 := ptr_src). rewrite Hptr_src. tac2.
      unfold Mem.free_aux. simpl. rewrite Int.eq_true. tac2.
      rewrite H1. simpl. eauto.
    - simpl. intros. tac2.
      destruct (peq b_src x); tac.
      + exploit Hsrc_private; eauto. by rewrite Hb_src.
      + by apply Hsrc_private.
    - repeat intro. tac.
      destruct (peq i x); tac.
    - auto.
    - repeat intro. tac.
      destruct (peq b_src x); tac.
      + eapply Haddress; eauto.
      + eapply Haddress; eauto.
    - repeat intro. tac.
      exploit Hpublic; eauto. intro X. inv X. tac2.
      destruct (peq si x); tac.
      { exploit Hsrc_private; eauto.
        - instantiate (1 := x). inv Hi. by rewrite H2.
        - intro X. rewrite Hb_src in X. inv X.
      }
      constructor; auto. intros.
      match goal with
        | [|- MemEquivC.rel_value _ ?blocks _ _ _] =>
          erewrite (Mem.blocks_unsimpl blocks)
      end.
      eapply incr_rel_value_aux; eauto.
      { eapply Mem.free_le; eauto. }
      { reflexivity. }
      { repeat intro. tac.
        destruct (peq b_src x); tac.
        - exploit Haddress.
          + apply Hsrc_private_subset. eauto.
          + eauto.
          + tac.
        - exploit Haddress; eauto.
      }
  Qed.

  Lemma free_tgt
        b_tgt
        (rel:t) m_src m_tgt
        (Hrel: rel m_src m_tgt)
        ptr_tgt (Hptr_tgt: MemC.val2ptr ptr_tgt m_tgt = Some (b_tgt, Int.zero))
        blk_tgt (Hb_tgt: PTree.get b_tgt rel.(tgt_private) = Some blk_tgt)
        sz tgt2_private (Hm2_tgt_private: MemC.free_aux b_tgt rel.(tgt_private) = Some (sz, tgt2_private)):
    exists m2_tgt, MemC.free ptr_tgt m_tgt = Some (sz, m2_tgt) /\
    le rel (mk rel.(equiv) rel.(src_private) tgt2_private m_src m2_tgt) /\
    (mk rel.(equiv) rel.(src_private) tgt2_private m_src m2_tgt) m_src m2_tgt.
  Proof.
    inv Hrel. destruct Hpublic.
    assert (Hm2_tgt: MemC.free ptr_tgt (tgt rel) = Some (sz, MemC.mk (tgt rel).(MemC.nextb) (PTree.set b_tgt (MemC.Block.mk (negb blk_tgt.(MemC.Block.accessible)) blk_tgt.(MemC.Block.accessible) blk_tgt.(MemC.Block.size) blk_tgt.(MemC.Block.address) blk_tgt.(MemC.Block.content)) (tgt rel)))).
    { unfold MemC.free. rewrite Hptr_tgt. simpl. rewrite Int.eq_true.
      unfold MemC.free_aux in *. tac2.
    }
    eexists. split; [apply Hm2_tgt|].
    split.
    { constructor; eauto.
      + reflexivity.
      + eapply MemC.free_le; eauto. }
    constructor; auto; [| | |constructor].
    - eapply MemC.free_wf, Hm2_tgt; eauto.
    - intros. tac.
      exploit Htgt_private; eauto. intro X.
      unfold MemC.free_aux in Hm2_tgt_private. tac.
      destruct (peq b_tgt0 b_tgt); tac.
      by rewrite X in Hcond.
    - repeat intro. tac.
      unfold MemC.free_aux in Hm2_tgt_private. tac.
      destruct (peq i b_tgt); tac.
    - auto.
    - repeat intro. tac.
      exploit Haddress; eauto. intro. tac.
      eexists. split; [by eauto|]; tac.
      destruct (peq x b_tgt); [subst|].
      + eexists. split; [done|]. simpl.
        exploit Htgt_private_subset; eauto. tac.
      + rewrite H0. eexists. split; eauto.
    - repeat intro. tac.
      exploit Hpublic; eauto. intro X. inv X. tac2.
      destruct (peq ti b_tgt); tac.
      { exploit Htgt_private; eauto.
        - instantiate (1 := b_tgt).
          apply Hequiv_wf in Hi. by rewrite Hi.
        - intro X. rewrite Hb_tgt in X. inv X.
      }
      constructor; auto. intros.
      match goal with
        | [|- MemEquivC.rel_value _ _ ?blocks _ _] =>
          erewrite (MemC.blocks_unsimpl blocks)
      end.
      eapply incr_rel_value_aux; eauto.
      { reflexivity. }
      { eapply MemC.free_le; eauto. }
      { repeat intro. tac.
        exploit Haddress; eauto. intro. tac.
        exists x. split; [done|]. tac. rewrite H2, if_Some.
        eexists. split; [done|].
        destruct (peq x b_tgt); tac.
      }
  Qed.

  Lemma free_public
        (rel:t) m_src m_tgt
        (Hrel: rel m_src m_tgt)
        ptr_src ptr_tgt (Hptr_src: MemEquivC.rel_value rel.(equiv) m_src m_tgt ptr_src ptr_tgt)
        sz m2_src (Hm2_src: Mem.free ptr_src m_src = Some (sz, m2_src)):
    exists m2_tgt,
      MemC.free ptr_tgt m_tgt = Some (sz, m2_tgt) /\
      le rel (mk rel.(equiv) rel.(src_private) rel.(tgt_private) m2_src m2_tgt) /\
      (mk rel.(equiv) rel.(src_private) rel.(tgt_private) m2_src m2_tgt) m2_src m2_tgt.
  Proof.
    cut (exists m2_tgt,
      MemC.free ptr_tgt m_tgt = Some (sz, m2_tgt) /\
      (mk rel.(equiv) rel.(src_private) rel.(tgt_private) m2_src m2_tgt) m2_src m2_tgt).
    { intros [m2_tgt [Hfree_tgt Hmem_rel]].
      eexists; split; eauto.
      split; eauto.
      inv Hrel.
      constructor; eauto.
      - eapply Mem.free_le; eauto.
      - eapply MemC.free_le; eauto. }

    exploit Mem.free_inv; eauto. intro Hm2'_src.
    tac2.
    exploit (val2ptr_public ptr_src ptr_tgt); eauto.
    intros [b_tgt [Hb_tgt Hb]].
    unfold MemC.free. rewrite Hb_tgt, Int.eq_true.
    unfold MemC.free_aux.
    inv Hrel. destruct Hpublic. exploit Hpublic; try apply Hb; eauto.
    intro X. inv X. tac2.
    rewrite <- Hvalid, H1, Haccessible, Hsize. simpl.  eexists. tac2. split; [done|].
    constructor; auto; [| | | |constructor].
    - eapply Mem.free_wf. apply Hsrc_wf.
      unfold Mem.free. instantiate (2 := ptr_src). tac2. rewrite H.
      unfold Mem.free_aux. rewrite H0. simpl.
      rewrite Int.eq_true, H1, Haccessible, Hsize. simpl. eauto.
    - eapply MemC.free_wf. apply Htgt_wf.
      unfold MemC.free. instantiate (2 := ptr_tgt). rewrite Hb_tgt.
      unfold MemC.free_aux. tac2.
      rewrite Int.eq_true, H4. simpl.
      rewrite <- Hvalid, H1. simpl. eauto.
    - simpl. repeat intro. tac2.
      destruct (peq i x); tac.
      exploit Hsrc_private; eauto.
      * instantiate (1 := x). by rewrite Hb.
      * by rewrite H2.
    - repeat intro. tac.
      destruct (peq i b_tgt); tac.
      exploit Htgt_private; eauto.
      + instantiate (1 := b_tgt).
        apply Hequiv_wf in Hb. by rewrite Hb.
      + intro X. rewrite H2 in X. inv X.
    - auto.
    - repeat intro. tac.
      destruct (peq b_src x); tac.
      + exploit Haddress; eauto. intro. tac.
        exploit (MemEquivC.src_unique x b_tgt x0); eauto. intro. tac.
        exists x0. split; tac.
        eexists. split; tac.
      + exploit Haddress; eauto. intro. tac.
        eexists. split; eauto; tac. rewrite H3, if_Some.
        eexists. split; tac.
        destruct (peq x0 b_tgt); tac.
    - repeat intro. tac.
      exploit Hpublic; eauto. intro X. inv X. rewrite ? if_Some.
      assert (J: (si = x /\ ti = b_tgt) \/ (si <> x /\ ti <> b_tgt)).
      { destruct (peq si x); tac.
        - exploit (MemEquivC.src_unique x b_tgt ti); eauto.
        - destruct (peq ti b_tgt); tac.
          + exploit (MemEquivC.tgt_unique si x b_tgt); eauto.
          + by right.
      }
      constructor.
      + destruct J as [[? ?]|[? ?]]; tac.
        by rewrite ? peq_true.
      + destruct J as [[? ?]|[? ?]]; tac.
        by rewrite ? peq_true.
      + destruct J as [[? ?]|[? ?]]; tac.
        by rewrite ? peq_true.
      + intros.
        match goal with
          | [|- MemEquivC.rel_value _ ?blocks_src ?blocks_tgt _ _] =>
            erewrite (Mem.blocks_unsimpl blocks_src);
            erewrite (MemC.blocks_unsimpl blocks_tgt)
        end.
        eapply incr_rel_value_aux; eauto.
        { rewrite <- Haccessible, <- Hsize. eapply Mem.free_le; eauto. }
        { instantiate (1 := (tgt rel).(MemC.nextb)).
          instantiate (1 := (tgt rel)).
          eapply MemC.free_le; eauto.
          instantiate (2 := ptr_tgt).
          unfold MemC.free.
          rewrite Hb_tgt, Int.eq_true.
          unfold MemC.free_aux. rewrite H4. simpl.
          rewrite <- Hvalid, H1. simpl. eauto.
        }
        { repeat intro. tac.
          destruct (peq b_src x); [subst|].
          - inv Hblk.
            eexists. split; [by eauto|]. tac.
            eexists. split; [by eauto|]. tac.
            exploit Haddress; eauto. intro. tac.
            exploit (MemEquivC.src_unique x b_tgt x0); eauto.
            intro. tac.
          - exploit Haddress; eauto. intro. tac.
            eexists. split; [by eauto|]. tac. rewrite H6, if_Some.
            eexists. split; tac.
            destruct (peq x0 b_tgt); tac.
        }
        { destruct J as [[? ?]|[? ?]]; tac.
          rewrite ? peq_true. simpl.
          - apply Hcontent; auto. by rewrite Hsize.
          - apply Hcontent0; auto.
        }
  Qed.

  Lemma realize_tgt
        b_tgt
        (rel:t) m_src m_tgt
        (Hrel: rel m_src m_tgt)
        blk_tgt (Hb_tgt: PTree.get b_tgt m_tgt = Some blk_tgt)
        addr_tgt blk2_tgt (Hrealize_blk: MemC.Block.realize blk_tgt addr_tgt blk2_tgt)
        m2_tgt (Hrealize: MemC.realize m_tgt b_tgt addr_tgt m2_tgt):
    m2_tgt = MemC.mk m_tgt.(MemC.nextb) (PTree.set b_tgt blk2_tgt m_tgt) /\
    le rel (mk rel.(equiv) rel.(src_private) (if PTree.get b_tgt rel.(tgt_private)
                                              then PTree.set b_tgt blk2_tgt rel.(tgt_private)
                                              else rel.(tgt_private))
                                             m_src m2_tgt) /\
    (mk rel.(equiv) rel.(src_private) (if PTree.get b_tgt rel.(tgt_private)
                                       then PTree.set b_tgt blk2_tgt rel.(tgt_private)
                                       else rel.(tgt_private))
                                      m_src m2_tgt) m_src m2_tgt.
  Proof.
    cut (m2_tgt = MemC.mk m_tgt.(MemC.nextb) (PTree.set b_tgt blk2_tgt m_tgt) /\
      (mk rel.(equiv) rel.(src_private) (if PTree.get b_tgt rel.(tgt_private)
                                         then PTree.set b_tgt blk2_tgt rel.(tgt_private)
                                         else rel.(tgt_private)) m_src m2_tgt)
        m_src m2_tgt).
    { intros [X Y].
      split; eauto.
      split; eauto.
      inv Hrel.
      constructor; eauto.
      + reflexivity.
      + exploit MemC.realize_le; eauto. intro. tac.
        econstructor; tac. rewrite Hblk1, if_Some.
        split; eauto.
        destruct (peq b b_tgt); tac.
        by repeat split.
    }

    assert (Hm2_tgt: m2_tgt = MemC.mk m_tgt.(MemC.nextb) (PTree.set b_tgt blk2_tgt m_tgt)).
    { inv Hrealize_blk. inv Hrealize. inv Hblk2.
      inv Hrel. tac.
    }
    split; tac. inv Hrel. destruct Hpublic. constructor; tac; [| | |constructor].
    - by inv Hrealize.
    - repeat intro. tac.
      exploit Htgt_private; eauto. intro X.
      destruct (peq b_tgt0 b_tgt); tac.
      + by rewrite X.
      + destruct ((tgt_private rel) ! b_tgt); tac.
    - repeat intro. tac.
      destruct (peq i b_tgt); tac.
      + destruct ((tgt_private rel) ! b_tgt) eqn:X; tac.
        by rewrite X in H1.
      + destruct ((tgt_private rel) ! b_tgt); tac.
    - auto.
    - repeat intro.
      exploit Haddress; eauto. intro. tac.
      eexists. split; [by eauto|]. tac. rewrite H0, if_Some.
      eexists. split; [done|].
      destruct (peq x b_tgt); tac.
      inv Hrealize_blk. rewrite H1 in Haddress0. inv Haddress0.
    - repeat intro. exploit Hpublic; eauto. tac.
      intro X. inv X. rewrite if_Some. constructor; tac.
      + destruct (peq ti b_tgt); tac.
        inv Hrealize_blk. tac2.
      + destruct (peq ti b_tgt); tac.
        inv Hrealize_blk. tac2.
      + destruct (peq ti b_tgt); tac.
        inv Hrealize_blk. tac2.
      + assert (J: MemC.Block.content (if peq ti b_tgt then blk2_tgt else t0) = MemC.Block.content t0); [|rewrite J; clear J].
        { destruct (peq ti b_tgt); tac.
          inv Hrealize_blk. tac2.
        }
        intros. exploit Hcontent; eauto. intro X.
        match goal with
          | [|- MemEquivC.rel_value _ _ ?blocks _ _] =>
            erewrite (MemC.blocks_unsimpl blocks)
        end.
        eapply incr_rel_value_aux; eauto; tac.
        { reflexivity. }
        { repeat intro. tac. rewrite Hblk1, if_Some.
          eexists. split; tac.
          destruct (peq b b_tgt); tac.
          - inv Hrealize_blk. tac.
            by rewrite Haddress0.
          - repeat split; tac.
        }
        { repeat intro. exploit Haddress; eauto. intro. tac.
          eexists. split; eauto. tac. rewrite H2, if_Some.
          eexists. split; eauto.
          destruct (peq x b_tgt); tac.
          inv Hrealize_blk. tac.
          rewrite Haddress0 in H3. inv H3.
        }
  Grab Existential Variables.
    { apply ((tgt rel).(MemC.nextb)). }
  Qed.        

  Lemma realize_public
        b_src b_tgt
        (rel:t) m_src m_tgt
        (Hrel: rel m_src m_tgt)
        (Hequiv: rel.(equiv) b_src b_tgt)
        addr m2_tgt (Hrealize: MemC.realize m_tgt b_tgt addr m2_tgt):
    exists m2_src,
      Mem.realize m_src b_src addr m2_src /\
      le_public rel (mk rel.(equiv) rel.(src_private) rel.(tgt_private) m2_src m2_tgt) /\
      (mk rel.(equiv) rel.(src_private) rel.(tgt_private) m2_src m2_tgt) m2_src m2_tgt.
  Proof.
    cut (exists m2_src,
      Mem.realize m_src b_src addr m2_src /\
      (mk rel.(equiv) rel.(src_private) rel.(tgt_private) m2_src m2_tgt) m2_src m2_tgt).
    { intros [m2_src [X Y]].
      eexists; split; eauto.
      split; eauto.
      inv Hrel.
      constructor; simpl; eauto.
      + constructor; simpl; eauto.
        * eapply Mem.realize_le; eauto.
        * eapply MemC.realize_le; eauto.
      + repeat intro. auto.
      + repeat intro. auto.
    }

    inv Hrel. destruct Hpublic. inv Hrealize.
    exploit Hpublic; try apply Hequiv; eauto. rewrite Hblk1. intro X. inv X.
    assert (s.(Mem.Block.address) = None).
    { destruct (s.(Mem.Block.address)) eqn:Haddr_src; tac.
      exploit Haddress; eauto. intro. tac.
      exploit (MemEquivC.src_unique b_src b_tgt x); eauto. intro. tac.
      inv Hblk2. rewrite Haddress0 in H2. inv H2.
    }
    assert (Hsrc2_wf: Mem.wf (Mem.mk (src rel).(Mem.nextb)
                             (PTree.set
                                b_src
                                (Mem.Block.mk s.(Mem.Block.valid) s.(Mem.Block.accessible) s.(Mem.Block.size) (Some addr) s.(Mem.Block.content)) (src rel)))).
    { constructor; intros.
      - tac. inv Hsrc_wf. destruct (peq b b_src); tac.
        + exploit _nextb; eauto.
        + exploit _nextb; eauto.
      - tac. destruct (peq b b_src); tac.
        + inv Hwf2. tac. rewrite Hsize. inv Hblk2. inv Hwf2.
          constructor; intros; tac.
          exploit _size; eauto.
        + inv Hsrc_wf. exploit _blocks_wf; eauto.
      - tac. destruct (peq b1 b_src), (peq b2 b_src); tac.
        + inv Hblk2. exploit Haddress; eauto. intro. tac2.
          assert (b_tgt <> x).
          { intro. tac.
            exploit (MemEquivC.tgt_unique b2 b_src x); eauto.
          }
          inv Hwf2.
          exploit Hpublic; eauto. rewrite Hblk3, H2. intro X. inv X.
          exploit _blocks_disjoint; tac.
          { instantiate (3 := b_tgt). rewrite peq_true. tac. }
          { tac. by rewrite <- Hvalid. }
          { tac. }
          { instantiate (3 := x). rewrite peq_false; eauto. }
          { by rewrite <- Hvalid0, Hvalid2. }
          { eauto. }
          { eauto. }
          { tac. by rewrite Hsize, Hsize0. }
        + inv Hblk2. exploit Haddress; eauto. intro. tac2.
          assert (b_tgt <> x).
          { intro. tac.
            exploit (MemEquivC.tgt_unique b1 b_src x); eauto.
          }
          inv Hwf2.
          exploit Hpublic; eauto. intro X. inv X.
          exploit _blocks_disjoint; tac.
          { instantiate (3 := x). rewrite peq_false; eauto. }
          { tac2. by rewrite <- Hvalid0, Hvalid1. }
          { eauto. }
          { instantiate (3 := b_tgt). rewrite peq_true. tac. }
          { tac. by rewrite <- Hvalid. }
          { tac. }
          { eauto. }
          { tac2. by rewrite Hsize, Hsize0. }
        + inv Hsrc_wf. eapply _blocks_disjoint; eauto.
      - unfold Mem.blocks. rewrite PTree.gsspec.
        destruct (peq 1 b_src).
        + subst. eexists. split; [eauto|]. simpl.
          exploit Mem._address_one; eauto. intros [blk [Hblk Haddr]].
          rewrite <- H0 in Hblk. inv Hblk.
          exploit (MemEquivC.src_unique 1%positive 1%positive b_tgt); eauto. intro. subst.
          congruence.
        + by apply Mem._address_one.
    }
    eexists. split.
    - econstructor; eauto.
      constructor; tac.
      rewrite Hsize. inv Hblk2. inv Hwf0; tac.
    - constructor; tac; [| |constructor].
      { repeat intro. tac.
        destruct (peq i b_src); tac.
        exploit Hsrc_private; eauto.
        - instantiate (1 := b_src). by rewrite Hequiv.
        - intro X. rewrite H1 in X. inv X.
      }
      { repeat intro. tac.
        destruct (peq i b_tgt); tac.
        exploit Htgt_private; eauto.
        - instantiate (1 := b_tgt). apply Hequiv_wf in Hequiv. by rewrite Hequiv.
        - intro X. rewrite H1 in X. inv X.
      }
      { auto. }
      { repeat intro. tac.
        destruct (peq b_src0 b_src); tac.
        - eexists. split; eauto. tac.
          eexists. split; eauto.
          by inv Hblk2.
        - exploit Haddress; eauto. intro. tac.
          eexists. split; eauto. tac. rewrite H2, if_Some.
          eexists. split; eauto.
          destruct (peq x b_tgt); tac.
          inv Hblk2. tac.
          rewrite Haddress0 in H3. inv H3.
      }
      { repeat intro. tac.
        destruct (peq si b_src), (peq ti b_tgt); tac.
        - constructor; tac.
          { by inv Hblk2. }
          { inv Hblk2. tac. }
          { inv Hblk2. tac. }
          { inv Hblk2. tac.
            intros. exploit Hcontent; eauto.
            intro X.
            match goal with
              | [|- MemEquivC.rel_value _ ?blocks_src ?blocks_tgt _ _] =>
                erewrite (Mem.blocks_unsimpl blocks_src);
                erewrite (MemC.blocks_unsimpl blocks_tgt)
            end.
            eapply incr_rel_value_aux; eauto; tac.
            - repeat intro. tac. rewrite Hblk0, if_Some.
              eexists. split; eauto.
              destruct (peq b b_src); tac2; repeat split; tac.
              
            - repeat intro. tac. rewrite Hblk0, if_Some.
              eexists. split; eauto.
              destruct (peq b b_tgt); tac2; repeat split; tac.
            - repeat intro. tac.
              destruct (peq b_src0 b_src); tac.
              + eexists. split; eauto. tac.
                eexists. split; eauto.
              + exploit Haddress; eauto. intro. tac.
                eexists. split; eauto. tac. rewrite H2, if_Some.
                eexists. split; eauto.
                destruct (peq x b_tgt); tac.
                rewrite Haddress0 in H3. inv H3.
          }
        - exploit (MemEquivC.src_unique b_src b_tgt ti); eauto.
          intro. tac.
        - exploit (MemEquivC.tgt_unique si b_src b_tgt); eauto.
          intro. tac.
        - exploit Hpublic; eauto. intro X. inv X.
          constructor; tac. intros.
          exploit Hcontent0; eauto. intro X.
          match goal with
            | [|- MemEquivC.rel_value _ ?blocks_src ?blocks_tgt _ _] =>
              erewrite (Mem.blocks_unsimpl blocks_src);
              erewrite (MemC.blocks_unsimpl blocks_tgt)
          end.
          eapply incr_rel_value_aux; eauto; tac.
          + repeat intro. tac. rewrite Hblk0, if_Some.
            eexists. split; eauto.
            destruct (peq b b_src); tac2; repeat split; tac.
          + repeat intro. tac. rewrite Hblk0, if_Some.
            eexists. split; eauto. inv Hblk2.
            destruct (peq b b_tgt); tac2; repeat split; tac.
          + repeat intro. tac.
            destruct (peq b_src0 b_src); tac.
            * eexists. split; eauto. tac.
              eexists. split; eauto.
              by inv Hblk2.
            * exploit Haddress; eauto. intro. tac.
              eexists. split; eauto. tac. rewrite H4, if_Some.
              eexists. split; eauto.
              destruct (peq x b_tgt); tac.
              inv Hblk2. tac.
              rewrite Haddress0 in H5. inv H5.
      }
  Grab Existential Variables.
    { apply (tgt rel).(MemC.nextb). }
    { apply (tgt rel).(MemC.nextb). }
    { apply (tgt rel).(MemC.nextb). }
    { apply (tgt rel).(MemC.nextb). }
  Qed.

  Lemma realize_src
        b_src b_tgt blk_src blk_tgt addr
        (rel:t) m_src m_tgt
        (Hrel: rel m_src m_tgt)
        (Hequiv: rel.(equiv) b_src b_tgt)
        (Hblk_src: m_src ! b_src = Some blk_src)
        (Hblk_tgt: m_tgt ! b_tgt = Some blk_tgt)
        (Haddr_src: blk_src.(Mem.Block.address) = None)
        (Haddr_tgt: blk_tgt.(MemC.Block.address) = Some addr):
    exists m2_src,
      Mem.realize m_src b_src addr m2_src /\
      le rel (mk rel.(equiv) rel.(src_private) rel.(tgt_private) m2_src m_tgt) /\
      (mk rel.(equiv) rel.(src_private) rel.(tgt_private) m2_src m_tgt) m2_src m_tgt.
  Proof.
    cut (exists m2_src,
      Mem.realize m_src b_src addr m2_src /\
      (mk rel.(equiv) rel.(src_private) rel.(tgt_private) m2_src m_tgt) m2_src m_tgt).
    { intros [m2_src [X Y]].
      eexists; split; eauto.
      split; eauto.
      inv Hrel.
      constructor; eauto.
      + eapply Mem.realize_le; eauto.
      + reflexivity.
    }

    inv Hrel. destruct Hpublic.
    exploit Hpublic; try apply Hequiv; eauto. rewrite Hblk_tgt. intro X. inv X. tac2.
    assert (Hsrc2_wf: Mem.wf (Mem.mk (src rel).(Mem.nextb)
                             (PTree.set
                                b_src
                                (Mem.Block.mk s.(Mem.Block.valid) s.(Mem.Block.accessible) s.(Mem.Block.size) (Some addr) s.(Mem.Block.content)) (src rel)))).
    { constructor; intros.
      - inv Hsrc_wf. destruct (peq b b_src); tac.
        + exploit _nextb; eauto.
        + exploit _nextb; eauto.
      - destruct (peq b b_src); tac.
        + inv Htgt_wf. exploit _blocks_wf; eauto. rewrite Hsize. intro X. inv X.
          constructor; intros; tac.
          exploit _size; eauto.
        + inv Hsrc_wf. exploit _blocks_wf; eauto.
      - destruct (peq b1 b_src), (peq b2 b_src); tac.
        + inv Hblk2. exploit Haddress; eauto. intro. tac2.
          assert (b_tgt <> x).
          { intro. tac.
            exploit (MemEquivC.tgt_unique b2 b_src x); eauto.
          }
          exploit Hpublic; eauto. rewrite H0, H1. intro X. inv X.
          inv Htgt_wf.
          exploit _blocks_disjoint; tac.
          { instantiate (2 := b_tgt). eauto. }
          { by rewrite <- Hvalid. }
          { eauto. }
          { instantiate (2 := x). eauto. }
          { by rewrite <- Hvalid0, Hvalid2. }
          { eauto. }
          { eauto. }
          { by rewrite Hsize, Hsize0. }
        + inv Hblk1. exploit Haddress; eauto. intro. tac2.
          assert (b_tgt <> x).
          { intro. tac.
            exploit (MemEquivC.tgt_unique b1 b_src x); eauto.
          }
          exploit Hpublic; eauto. rewrite H0, H1. intro X. inv X.
          inv Htgt_wf.
          exploit _blocks_disjoint; tac.
          { instantiate (2 := x). eauto. }
          { by rewrite <- Hvalid0, Hvalid1. }
          { eauto. }
          { instantiate (2 := b_tgt). eauto. }
          { by rewrite <- Hvalid. }
          { eauto. }
          { eauto. }
          { by rewrite Hsize, Hsize0. }
        + inv Hsrc_wf. eapply _blocks_disjoint; eauto.
      - unfold Mem.blocks. rewrite PTree.gsspec.
        destruct (peq 1 b_src).
        + subst. eexists. split; [eauto|]. simpl.
          exploit Mem._address_one; eauto. intros [blk [Hblk Haddr]].
          rewrite Hblk_src in Hblk. inv Hblk.
          exploit (MemEquivC.src_unique 1%positive 1%positive b_tgt); eauto. intro. subst.
          congruence.
        + by apply Mem._address_one.
    }
    eexists. split.
    - econstructor; eauto.
      constructor; tac. inv Hsrc2_wf. exploit _blocks_wf; eauto. tac.
      rewrite peq_true; tac.
    - constructor; tac; [|constructor].
      { repeat intro. tac.
        destruct (peq i b_src); tac.
        exploit Hsrc_private; eauto.
        - instantiate (1 := b_src). by rewrite Hequiv.
        - intro X. rewrite H1 in X. inv X.
      }
      { auto. }
      { repeat intro. tac.
        destruct (peq b_src0 b_src); tac.
        - eexists. split; eauto. 
        - exploit Haddress; eauto.
      }
      { repeat intro. tac.
        destruct (peq si b_src), (peq ti b_tgt); tac.
        - rewrite Hblk_tgt. constructor; tac.
          intros. exploit Hcontent; eauto.
          intro X.
          match goal with
            | [|- MemEquivC.rel_value _ ?blocks_src ?blocks_tgt _ _] =>
              erewrite (Mem.blocks_unsimpl blocks_src);
              erewrite (MemC.blocks_unsimpl blocks_tgt)
          end.
          eapply incr_rel_value_aux; eauto; tac.
          + repeat intro. tac. rewrite Hblk1, if_Some.
            eexists. split; eauto.
            destruct (peq b b_src); tac2; repeat split; tac.
          + repeat intro. tac. rewrite Hblk1.
            eexists. split; eauto.
            destruct (peq b b_tgt); tac2; repeat split; tac.
          + repeat intro. tac.
            destruct (peq b_src0 b_src); tac.
            * eexists. split; eauto.
            * exploit Haddress; eauto.
        - exploit (MemEquivC.src_unique b_src b_tgt ti); eauto.
          intro. tac.
        - exploit (MemEquivC.tgt_unique si b_src b_tgt); eauto.
          intro. tac.
        - exploit Hpublic; eauto. intro X. inv X.
          constructor; tac. intros.
          exploit Hcontent0; eauto. intro X.
          match goal with
            | [|- MemEquivC.rel_value _ ?blocks_src ?blocks_tgt _ _] =>
              erewrite (Mem.blocks_unsimpl blocks_src);
              erewrite (MemC.blocks_unsimpl blocks_tgt)
          end.
          eapply incr_rel_value_aux; eauto; tac.
          + repeat intro. tac. rewrite Hblk1, if_Some.
            eexists. split; eauto.
            destruct (peq b b_src); tac2; repeat split; tac.
          + repeat intro. tac. rewrite Hblk1.
            eexists. split; eauto.
            destruct (peq b b_tgt); tac2; repeat split; tac.
          + repeat intro. tac.
            destruct (peq b_src0 b_src); tac.
            * eexists. split; eauto.
            * exploit Haddress; eauto.
      }
  Grab Existential Variables.
    { apply (tgt rel).(MemC.nextb). }
    { apply (tgt rel).(MemC.nextb). }
    { apply (tgt rel).(MemC.nextb). }
    { apply (tgt rel).(MemC.nextb). }
  Qed.

  Lemma denormalize_public
        i b_src o
        (rel:t) m_src m_tgt
        (Hrel: rel m_src m_tgt)
        (Hb_src: Mem.denormalize i m_src = Some (b_src, o)):
    MemEquivC.rel_value rel.(equiv) m_src m_tgt (Value.ptr b_src o) i.
  Proof.
    unfold Mem.denormalize in *.
    match goal with
      | [H: context [PTree.select ?pred ?m] |- _] =>
        generalize (PTree.gselectspec pred m);
          destruct (PTree.select pred m); [inv H|done]
    end.
    intros [b_src' [blk_src [Hblk_src X]]].
    destruct (Mem.Block.address blk_src) as [addr_src|] eqn:Haddr_src; inv X.
    destruct (Mem.Block.valid blk_src) eqn:Hvalid_src; inv H0.
    destruct (Range.contained i {| Range.begin := addr_src; Range.size := Mem.Block.size blk_src |}) eqn:Hcontained_src; inv H1.
    inv Hrel. destruct Hpublic. exploit Haddress; eauto.
    intros [b_tgt [Hb_tgt [blk_tgt [Hblk_tgt Haddr_tgt]]]].
    exploit Hpublic; eauto. intro X. inv X.
    rewrite Hblk_src in H0. inv H0.
    rewrite Hblk_tgt in H. inv H.
    replace i with (Int.add addr_src (Int.sub i addr_src)) at 2; [econstructor; eauto|].
    by rewrite Int.sub_add_opp, Int.add_permut, Int.add_neg_zero, Int.add_zero.
  Qed.

  Lemma ptrintcast_public
        b_src b_tgt o
        (rel:t) m_src m_tgt
        (Hwf_src: Mem.wf m_src) (Hwf_tgt: MemC.wf m_tgt)
        (Hb: rel.(equiv) b_src b_tgt)
        (Hrel: rel m_src m_tgt)
        i (Hrel_value: MemEquivC.rel_value rel.(equiv) m_src m_tgt (Value.ptr b_src o) i):
    exists m2_src (rel2:MemRelC.t),
      (exists j, i = ValueC.int j) /\
      (match Mem.normalize b_src o m_src with
         | Some i => SomeP (ValueC.int i, m_src)
         | None =>
           doPP val <- liftP (fun _:int => True);
         doPP mem <- liftP (Mem.realize m_src b_src (Int.sub val o));
         SomeP (ValueC.int val, mem)
       end) (Some (i, m2_src)) /\
      rel2 m2_src m_tgt /\
      MemRelC.le_public rel rel2.
  Proof.
    inv Hrel_value.
    exploit (MemEquivC.src_unique b_src b_tgt ti); eauto. intro. subst ti.
    rename tblk into blk_tgt. rename Htblk into Hblk_tgt.
    rename taddr into addr_tgt. rename Htaddr into Haddr_tgt.
    unfold Mem.normalize in *.
    destruct (match m_src ! b_src with
                | Some blk =>
                  match Mem.Block.address blk with
                    | Some p => Some (Int.add p o)
                    | None => None
                  end
                | None => None
              end) eqn:Hdecide_src.
    { destruct (m_src ! b_src) as [blk_src|] eqn:Hblk_src; inv Hdecide_src.
      destruct (Mem.Block.address blk_src) as [addr_src|] eqn:Haddr_src; inv H0.
      inv Hrel. destruct Hpublic. exploit Haddress; eauto.
      intros [b_tgt' [Hb_tgt [blk_tgt' [Hblk_tgt' Haddr_tgt']]]].
      exploit (MemEquivC.src_unique b_src b_tgt b_tgt'); eauto. intro. subst b_tgt'.
      rewrite Hblk_tgt in Hblk_tgt'. symmetry in Hblk_tgt'. inv Hblk_tgt'.
      rewrite Haddr_tgt in Haddr_tgt'. symmetry in Haddr_tgt'. inv Haddr_tgt'.
      eexists. eexists. split; [eexists; eauto|]. split; [reflexivity|]. split; [|reflexivity].
      constructor; auto. constructor; auto.
    }
    { inv Hrel. destruct Hpublic. exploit Hpublic; try apply Hb; eauto. intro X. inv X.
      rewrite Hblk_tgt in H. inv H. rename s into blk_src. rename H0 into Hblk_src.
      rewrite <- Hblk_src in Hdecide_src.
      destruct (Mem.Block.address blk_src) eqn:Haddress_src; inv Hdecide_src.
      exists
        {|
          Mem.nextb := Mem.nextb (src rel);
          Mem.blocks := PTree.set b_src
                                  {|
                                    Mem.Block.valid := Mem.Block.valid blk_src;
                                    Mem.Block.accessible := Mem.Block.accessible blk_src;
                                    Mem.Block.size := Mem.Block.size blk_src;
                                    Mem.Block.address := Some addr_tgt;
                                    Mem.Block.content := Mem.Block.content blk_src |} (src rel) |}.
      eexists
        (MemRelC.mk
           rel.(equiv)
           rel.(src_private)
           rel.(tgt_private)
           _
           rel.(tgt)).
      split; [eexists; eauto|].
      assert (Hwf_src':
                Mem.wf
                  {|
                    Mem.nextb := Mem.nextb (src rel);
                    Mem.blocks := PTree.set b_src
                                            {|
                                              Mem.Block.valid := Mem.Block.valid blk_src;
                                              Mem.Block.accessible := Mem.Block.accessible blk_src;
                                              Mem.Block.size := Mem.Block.size blk_src;
                                              Mem.Block.address := Some addr_tgt;
                                              Mem.Block.content := Mem.Block.content blk_src |} (src rel) |}).
      { constructor; intros.
        - simpl in *. rewrite PTree.gsspec in Hblk.
          destruct (peq b b_src); [subst|].
          + inv Hblk. inv Hwf_src. eapply _nextb; eauto.
          + inv Hwf_src. eapply _nextb; eauto.
        - simpl in *. rewrite PTree.gsspec in Hblk.
          destruct (peq b b_src); [subst|].
          + inv Hblk. constructor. simpl. intros. inv Hp.
            rewrite Hsize. apply MemC.Block._size; auto.
            eapply MemC._blocks_wf; try apply Hblk_tgt; eauto.
          + inv Hwf_src. eapply _blocks_wf; eauto.
        - simpl in *. rewrite PTree.gsspec in *.
          destruct (peq b1 b_src); [subst; inv Hblk1; inv Hp1|]; simpl in *.
          + destruct (peq b2 b_src); [subst; inv Hblk2; inv Hp2|]; simpl in *.
            * congruence.
            * exploit Haddress; try apply Hb; eauto. intros [? [? [? [? ?]]]].
              exploit Hpublic; eauto. intro X. inv X.
              rewrite H0 in H4. inv H4.
              rewrite Hblk2 in H3. inv H3.
              rewrite Hsize, Hsize0.
              eapply MemC._blocks_disjoint; try apply Hwf_tgt; eauto; try congruence.
              intro. subst.
              rewrite Hblk_tgt in H0. symmetry in H0. inv H0.
              exploit (MemEquivC.tgt_unique b2 b_src); eauto.
          + destruct (peq b2 b_src); [subst; inv Hblk2; inv Hp2|]; simpl in *.
            * exploit Haddress; try apply Hb; eauto. intros [? [? [? [? ?]]]].
              exploit Hpublic; eauto. intro X. inv X.
              rewrite H0 in H4. inv H4.
              rewrite Hblk1 in H3. inv H3.
              rewrite Hsize, Hsize0.
              eapply MemC._blocks_disjoint; try apply Hwf_tgt; eauto; try congruence.
              intro. subst.
              rewrite Hblk_tgt in H0. symmetry in H0. inv H0.
              exploit (MemEquivC.tgt_unique b1 b_src); eauto.
            * eapply Mem._blocks_disjoint; try apply Hwf_src; eauto.
        - unfold Mem.blocks. rewrite PTree.gsspec.
          destruct (peq 1 b_src).
          + subst. eexists. split; [eauto|]. simpl.
            exploit Mem._address_one; eauto. intros [blk [Hblk Haddr]].
            rewrite <- Hblk_src in Hblk. symmetry in Hblk. inv Hblk.
            exploit (MemEquivC.src_unique 1%positive 1%positive b_tgt); eauto. intro. subst.
            congruence.
          + by apply Mem._address_one.
      }
      split.
      { eapply bindPP_right.
        { instantiate (1 := Int.add addr_tgt o). simpl. auto. }
        rewrite Int.sub_add_opp, Int.add_assoc, Int.add_neg_zero, Int.add_zero.
        eapply bindPP_right.
        { econstructor; eauto.
          { econstructor; eauto. econstructor; eauto. simpl. intros. inv Hp.
            exploit MemC._blocks_wf; try apply Hblk_tgt; eauto. intro X. inv X.
            rewrite Hsize. apply _size. auto.
          }
        }
        { constructor. }
      }
      { split.
        { constructor; simpl; eauto.
          { repeat intro. exploit Hsrc_private_subset; eauto. intro X.
            rewrite PTree.gsspec. destruct (peq i b_src); [subst|]; auto.
            rewrite Hsrc_private in H1; [congruence|].
            inv Hb. congruence.
          }
          constructor; simpl; repeat intro; auto.
          - rewrite PTree.gsspec in Hblk. destruct (peq b_src0 b_src); [subst; inv Hblk|]; simpl in *.
            + inv Hp. eauto.
            + eapply Haddress; eauto.
          - rewrite PTree.gsspec. destruct (peq si b_src); [subst|].
            + exploit Hpublic; eauto. intro X. inv X.
              rewrite <- Hblk_src in H0. inv H0.
              constructor; simpl; intros; try congruence.
              exploit Hcontent0; eauto; try congruence. intro X. inv X.
              * constructor.
              * econstructor; eauto.
            + exploit Hpublic; eauto. intro X. inv X.
              constructor; auto. intros.
              exploit Hcontent0; eauto. intro X. inv X.
              * constructor.
              * econstructor; eauto.
          }
          { constructor; simpl; auto.
            - constructor; simpl; auto; [|reflexivity].
              repeat intro. rewrite PTree.gsspec. destruct (peq b b_src); [subst|].
              + rewrite <- Hblk_src in Hblk1. inv Hblk1.
                eexists. repeat (simpl; split; eauto).
                rewrite Haddress_src. repeat intro. congruence.
              + eexists. repeat (split; eauto).
                repeat intro. auto.
            - repeat intro. auto.
            - repeat intro. auto.
          }
      }
    }
  Qed.

  Lemma public
        b_src b_tgt
        (rel:t) m_src m_tgt
        (Hrel: rel m_src m_tgt)
        blk_src (Hblk_src: PTree.get b_src rel.(src_private) = Some blk_src)
        blk_tgt (Hblk_tgt: PTree.get b_tgt rel.(tgt_private) = Some blk_tgt)
        (Hblk: MemEquivC.rel_block (MemEquivC.add b_src b_tgt rel.(equiv)) m_src m_tgt
                                  (Some blk_src)
                                  (Some blk_tgt)):
    le rel (mk (MemEquivC.add b_src b_tgt rel.(equiv))
        (PTree.remove b_src rel.(src_private))
        (PTree.remove b_tgt rel.(tgt_private)) m_src m_tgt) /\
    (mk (MemEquivC.add b_src b_tgt rel.(equiv))
        (PTree.remove b_src rel.(src_private))
        (PTree.remove b_tgt rel.(tgt_private)) m_src m_tgt)
      m_src m_tgt.
  Proof.
    cut ((mk (MemEquivC.add b_src b_tgt rel.(equiv))
        (PTree.remove b_src rel.(src_private))
        (PTree.remove b_tgt rel.(tgt_private)) m_src m_tgt)
      m_src m_tgt).
    {
      intros; split; eauto.
      inv Hrel.
      constructor.
      + repeat intro. tac.
        unfold MemEquivC.add, MemEquivC.to_relation. tac.
        destruct (peq x0 b_src); tac.
        exploit Hsrc_private.
        * instantiate (1 := b_src). by rewrite PR.
        * intro X. rewrite Hblk_src in X. inv X.
      + reflexivity.
      + reflexivity.
    }

    inv Hblk. inv Hrel. destruct Hpublic. econstructor; auto; [| | | | |constructor].
    - apply MemEquivC.add_wf; auto.
      + intros [b'_tgt Hb'_tgt].
        exploit Hsrc_private; eauto.
        * instantiate (1 := b_src). by rewrite Hb'_tgt.
        * intro X. rewrite Hblk_src in X. inv X.
      + intros [b'_src Hb'_src].
        exploit Htgt_private; eauto.
        * instantiate (1 := b_tgt).
          apply Hequiv_wf in Hb'_src. by rewrite Hb'_src.
        * intro X. rewrite Hblk_tgt in X. inv X.
    - simpl. intros. rewrite PTree.grspec. unfold PTree.elt_eq.
      unfold MemEquivC.add in H. simpl in H.
      destruct (peq b_src0 b_src); [done|].
      apply Hsrc_private; auto.
    - simpl. intros. rewrite PTree.grspec. unfold PTree.elt_eq.
      unfold MemEquivC.add in H. simpl in H.
      destruct (peq b_tgt0 b_tgt); [done|].
      apply Htgt_private; auto.
    - simpl. repeat intro. tac2.
    - simpl. repeat intro. tac2.
    - simpl. unfold MemEquivC.add, MemEquivC.to_relation. simpl.
      destruct (peq 1 b_src); subst.
      + exploit Hsrc_private; eauto.
        * instantiate (1 := 1%positive).
          rewrite Hone. congruence.
        * intro X. rewrite Hblk_src in X. congruence.
      + auto.
    - repeat intro. tac.
      exploit Haddress; eauto. intro. tac.
      eexists. split; eauto.
      unfold MemEquivC.add, MemEquivC.to_relation. tac.
      destruct (peq b_src0 b_src); tac.
      exploit Hsrc_private; eauto.
      + instantiate (1 := b_src). by rewrite H.
      + intro X. rewrite Hblk_src in X. inv X.
    - repeat intro. tac.
      assert (J: (si = b_src /\ ti = b_tgt) \/ (si <> b_src /\ rel.(equiv) si ti)).
      { unfold MemEquivC.add, MemEquivC.to_relation in Hi. simpl in Hi.
        destruct (peq si b_src); tac.
        - by left.
        - by right.
      }
      destruct J as [[? ?]|[? ?]]; tac.
      exploit Hpublic; eauto. intro X. inv X.
      constructor; tac.
      intros.
      eapply (incr_rel_value_aux rel.(equiv) (MemEquivC.add b_src b_tgt rel.(equiv))); eauto.
      { intros.
        unfold MemEquivC.add, MemEquivC.to_relation. simpl.
        destruct (peq x0 b_src); tac.
        exploit Hsrc_private.
        - instantiate (1 := b_src). by rewrite PR.
        - intro X. rewrite Hblk_src in X. inv X.
      }
      { reflexivity. }
      { reflexivity. }
      { repeat intro. exploit Haddress; eauto. intro. tac.
        exists x. split; eauto; tac.
        unfold MemEquivC.add, MemEquivC.to_relation. simpl.
        destruct (peq b_src0 b_src); tac.
        exploit Hsrc_private; eauto.
        - instantiate (1 := b_src). by rewrite H1.
        - intro X. rewrite Hblk_src in X. inv X.
      }
  Qed.

  Lemma alloc_public
        (rel:t) m_src m_tgt
        (Hrel: rel m_src m_tgt)
        (Hwf_src: Mem.wf m_src) (Hwf_tgt: MemC.wf m_tgt)
        size
        addr_tgt b_tgt m2_tgt (Hm2_tgt: MemC.allocC size m_tgt (b_tgt, addr_tgt, m2_tgt)):
    exists b_src m2_src,
      Mem.alloc size m_src = (b_src, m2_src) /\
      le rel (mk (MemEquivC.add b_src b_tgt rel.(equiv)) rel.(src_private) rel.(tgt_private) m2_src m2_tgt) /\
      (mk (MemEquivC.add b_src b_tgt rel.(equiv)) rel.(src_private) rel.(tgt_private) m2_src m2_tgt)
        m2_src m2_tgt.
  Proof.
    exploit alloc_tgt; eauto. intros [Hle2 Hrel2].
    destruct (Mem.alloc size m_src) as [b_src m2_src] eqn:Halloc_src.
    eexists. eexists. split; [eauto|].
    exploit alloc_src; eauto. intros [Hle3 Hrel3].
    exploit public; eauto; simpl.
    { rewrite PTree.gss. eauto. }
    { rewrite PTree.gss. eauto. }
    { constructor; auto. intros. simpl. rewrite ? ZMap.gi. constructor. }
    simpl. intros [Hle4 Hrel4].
    split.
    - etransitivity; eauto. etransitivity; eauto. etransitivity; eauto.
      constructor; simpl; auto.
      + reflexivity.
      + reflexivity.
    - assert ((src_private rel) ! b_src = None).
      { apply Mem.alloc_inv in Halloc_src. des. subst.
        destruct ((src_private rel) ! (Mem.nextb m_src)) eqn:X; [|done].
        inv Hrel. exploit Hsrc_private_subset; eauto. intro Y.
        exploit Mem._nextb; try apply Y; eauto. xomega.
      }
      assert ((tgt_private rel) ! b_tgt = None).
      { inv Hm2_tgt. apply MemC.alloc_inv in Halloc. des. subst.
        inv Hrealize. inv Hblk2.
        destruct ((tgt_private rel) ! (MemC.nextb m_tgt)) eqn:X; [|done].
        inv Hrel. exploit Htgt_private_subset; eauto. intro Y.
        exploit MemC._nextb; try apply Y; eauto. xomega.
      }
      inversion Hrel4. simpl in *. constructor; simpl; auto.
      + intros. exploit Hsrc_private; eauto.
        rewrite PTree.grspec, PTree.gsspec.
        unfold PTree.elt_eq. destruct (peq b_src0 b_src); subst; auto.
      + intros. exploit Htgt_private; eauto.
        rewrite PTree.grspec, PTree.gsspec.
        unfold PTree.elt_eq. destruct (peq b_tgt0 b_tgt); subst; auto.
      + eapply PTree.gsubset_trans; eauto. repeat intro.
        rewrite PTree.grspec, PTree.gsspec.
        unfold PTree.elt_eq. destruct (peq i b_src); subst; auto.
        congruence.
      + eapply PTree.gsubset_trans; eauto. repeat intro.
        rewrite PTree.grspec, PTree.gsspec.
        unfold PTree.elt_eq. destruct (peq i b_tgt); subst; auto.
        congruence.
  Qed.
End MemRelC.
Coercion MemRelC.sem: MemRelC.t >-> Funclass.
Coercion MemRelC.equiv: MemRelC.t >-> MemEquivC.t.

Lemma load_public'
      (rel:MemRelC.t) m_src m_tgt
      (Hrel: rel m_src m_tgt)
      ptr_src ptr_tgt (Hptr: MemEquivC.rel_value rel m_src m_tgt ptr_src ptr_tgt)
      val_tgt (Hval_tgt: MemC.load ptr_tgt m_tgt = Some val_tgt):
  Mem.load ptr_src m_src = None \/
  exists val_src,
    Mem.load ptr_src m_src = Some val_src /\
    MemEquivC.rel_value rel m_src m_tgt val_src val_tgt.
Proof.
  destruct (Mem.load ptr_src m_src) eqn: load_src; [right | left; auto].
  exploit MemRelC.load_public; eauto; intro; tac.
  rewrite H in Hval_tgt; tac; eauto.
Qed.

Lemma store_public'
      (rel:MemRelC.t) m_src m_tgt
      (Hrel: rel m_src m_tgt)
      ptr_src ptr_tgt (Hptr: MemEquivC.rel_value rel m_src m_tgt ptr_src ptr_tgt)
      val_src val_tgt (Hval: MemEquivC.rel_value rel m_src m_tgt val_src val_tgt)
      m2_tgt (Hm2_tgt: MemC.store ptr_tgt val_tgt m_tgt = Some m2_tgt):
  Mem.store ptr_src val_src m_src = None \/
  exists m2_src,
    Mem.store ptr_src val_src m_src = Some m2_src /\
    MemRelC.le rel (MemRelC.mk rel rel.(MemRelC.src_private) rel.(MemRelC.tgt_private) m2_src m2_tgt) /\
    (MemRelC.mk rel rel.(MemRelC.src_private) rel.(MemRelC.tgt_private) m2_src m2_tgt) m2_src m2_tgt.
Proof.
  destruct (Mem.store ptr_src val_src m_src) eqn: store_src; [right | left; auto].
  exploit MemRelC.store_public; [eauto | apply Hptr | apply Hval | eauto | intro; tac; eauto].
  rewrite Hm2_tgt in H. inv H.
  eexists. split; eauto.
Qed.

Lemma free_public'
      (rel:MemRelC.t) m_src m_tgt
      (Hrel: rel m_src m_tgt)
      ptr_src ptr_tgt (Hptr_src: MemEquivC.rel_value rel m_src m_tgt ptr_src ptr_tgt)
      sz m2_tgt (Hm2_tgt: MemC.free ptr_tgt m_tgt = Some (sz, m2_tgt)):
  Mem.free ptr_src m_src = None \/
  exists m2_src,
    Mem.free ptr_src m_src = Some (sz, m2_src) /\
    MemRelC.le rel (MemRelC.mk rel rel.(MemRelC.src_private) rel.(MemRelC.tgt_private) m2_src m2_tgt) /\
    (MemRelC.mk rel rel.(MemRelC.src_private) rel.(MemRelC.tgt_private) m2_src m2_tgt) m2_src m2_tgt.
Proof.
  destruct (Mem.free ptr_src m_src) eqn: free_src; [right | left; auto].
  destruct p.
  exploit MemRelC.free_public; eauto; intro X; tac.
  rewrite H in Hm2_tgt; tac; eauto.
Qed.

Lemma get_tgt
      b_tgt blk_tgt
      (mrel:MemRelC.t) m_src m_tgt
      (Hwf: mrel m_src m_tgt)
      (Hblk_tgt: PTree.get b_tgt mrel.(MemRelC.tgt_private) = Some blk_tgt):
  PTree.get b_tgt m_tgt = Some blk_tgt.
Proof. inv Hwf. tac. Qed.

Lemma get_src
      b_src blk_src
      (mrel:MemRelC.t) m_src m_tgt
      (Hwf: mrel m_src m_tgt)
      (Hblk_src: PTree.get b_src mrel.(MemRelC.src_private) = Some blk_src):
  PTree.get b_src m_src = Some blk_src.
Proof. inv Hwf. tac. Qed.
