Require Import Coqlib.
Require Import Kanglib.
Require Import Integers.
Require Import Maps.
Require Import AST.

(** Range [begin, begin+size),
 where Rbegin, Rsize are unsigned int32. *)
Module Range.
  Structure t: Type := mk {
    begin: int;
    size: int
  }.

  Definition wf (r:t): bool :=
    zlt ((Int.unsigned r.(begin)) + (Int.unsigned r.(size))) Int.modulus.

  Definition contained (x:int) (r:t): bool :=
    zle (Int.unsigned r.(begin)) (Int.unsigned x) &&
    zlt (Int.unsigned x) (Int.unsigned r.(begin) + Int.unsigned r.(size)).

  Definition disjoint (r1 r2:t): bool :=
    zle (Int.unsigned r1.(begin) + Int.unsigned r1.(size)) (Int.unsigned r2.(begin)) ||
    zle (Int.unsigned r2.(begin) + Int.unsigned r2.(size)) (Int.unsigned r1.(begin)).

  Lemma intval_repr x:
    Int.intval (Int.repr x) = Int.Z_mod_modulus x.
  Proof. done. Qed.

  Ltac tac_aux :=
    match goal with
      | [H: is_true (proj_sumbool (zlt ?a ?b)) |- _] =>
        destruct (zlt a b); [|done]
      | [H: proj_sumbool (zlt ?a ?b) = true |- _] =>
        destruct (zlt a b); [|done]
      | [H: is_true (proj_sumbool (zle ?a ?b)) |- _] =>
        destruct (zle a b); [|done]
      | [H: proj_sumbool (zle ?a ?b) = true |- _] =>
        destruct (zle a b); [|done]
      | [H: context[Int.unsigned (Int.repr ?i)] |- _] =>
        rewrite (intval_repr) in H
      | [|- context[Int.unsigned (Int.repr ?i)]] =>
        rewrite (intval_repr)
    end.

  Lemma contained_trans i sz trs
        (Hc: contained i (Range.mk Int.zero sz))
        (Hwf2: wf (Range.mk trs sz)):
    contained (Int.add i trs) (Range.mk trs sz).
  Proof.
    unfold wf, contained in *.
    destruct i, sz, trs. simpl in *.
    unfold Int.ltu, Int.add, Int.zero in *. simpl in *.
    rewrite intval_repr in *. simpl in *.
    rename intval into i.
    rename intval0 into sz.
    rename intval1 into trs.
    repeat
      (simpl in *; subst;
       (try done);
       (try basictac_aux);
       (try rewrite Int.Z_mod_modulus_eq in *);
       (try tac_aux)).
    rewrite (Zmod_small (i + trs)); [|omega].
    split.
    - destruct (zle trs (i + trs)); [done|omega].
    - destruct (zlt (i + trs) (trs + sz)); [done|omega].
  Qed.

  Lemma contained_trans' i sz trs
        (Hc: contained i (Range.mk trs sz))
        (Hwf2: wf (Range.mk trs sz)):
    contained (Int.sub i trs) (Range.mk Int.zero sz).
  Proof.
    unfold wf, contained in *.
    destruct i, sz, trs. simpl in *.
    unfold Int.ltu, Int.sub, Int.zero in *. simpl in *.
    rewrite intval_repr in *. simpl in *.    
    rename intval into i.
    rename intval0 into sz.
    rename intval1 into trs.
    repeat
      (simpl in *; subst;
       (try done);
       (try basictac_aux);
       (try rewrite Int.Z_mod_modulus_eq in *);
       (try tac_aux)).
    rewrite (Zmod_small (i - trs)) in *; [|omega].
    split.
    - destruct (zle 0 (i - trs)); [done|omega].
    - destruct (zlt (i - trs) sz); [done|omega].
  Qed.

  Lemma gdspec
        x (r1 r2:t)
        (H1: contained x r1) (H2: contained x r2)
        (H12: disjoint r1 r2):
    False.
  Proof.
    destruct r1, r2.
    destruct x, begin0, size0, begin1, size1.
    rename intval into x.
    rename intval0 into b1.
    rename intval1 into s1.
    rename intval2 into b2.
    rename intval3 into s2.
    unfold contained, disjoint, Int.add in *. simpl in *.
    repeat
      (simpl in *; subst;
       (try done);
       (try omega);
       (try basictac_aux);
       (try rewrite Int.Z_mod_modulus_eq in *);
       (try tac_aux)).
  Qed.
End Range.

Module Typ.
  Inductive t: Type :=
  | int
  | ptr
  .

  Definition eq (x y:t): {x = y} + {x <> y}.
  Proof. decide equality. Defined.
End Typ.

(** Value is either:
 - machine integer;
 - logical pointer (b, o), where b is block identifier and o is offset
*)

Coercion Int.repr: Z >-> int.

Module Value.
  Inductive t: Type :=
  | int (intval:int)
  | ptr (b:ident) (o:int)
  .

  Definition eq (x y:t): {x=y} + {x<>y}.
  Proof.
    decide equality. 
    - apply Int.eq_dec.
    - apply Int.eq_dec.
    - apply ident_eq.
  Defined.

  Definition has_type (v:t) (typ:Typ.t): bool :=
    match v, typ with
      | int _, Typ.int => true
      | ptr _ _, Typ.ptr => true
      | _, _ => false
    end.
End Value.
Coercion Value.int: int >-> Value.t.

Module ValueC.
  Inductive t: Type :=
  | int (intval:int)
  .

  Definition eq (x y:t): {x=y} + {x<>y}.
  Proof.
    decide equality. 
    apply Int.eq_dec.
  Defined.

  Definition has_type (v:t) (typ:Typ.t): bool :=
    match v, typ with
      | int _, Typ.int => true
      | int _, Typ.ptr => true
    end.
End ValueC.
Coercion ValueC.int: int >-> ValueC.t.

Ltac valuetac_aux :=
  match goal with
    | [H: Value.int _ = Value.int _ |- _] => inv H
    | [H: Value.int _ = Value.ptr _ _ |- _] => inv H
    | [H: Value.ptr _ _ = Value.int _ |- _] => inv H
    | [H: Value.ptr _ _ = Value.ptr _ _ |- _] => inv H

    | [H: ValueC.int _ _ = ValueC.int _ _ |- _] => inv H

    | [H: match ?i with
            | {| Int.intval := intval |} => intval
          end =
          match ?j with
            | {| Int.intval := intval |} => intval
          end |- _] =>
      apply Int.intval_injection in H; subst
  end.
