Require Import Coqlib.
Require Import Kanglib.
Require Import Integers.
Require Import AST.
Require Import Values.
Require Import Monad.
Require Import Maps.

Ltac tac :=
  repeat
    (simpl in *; subst;
     (try done);
     (try basictac_aux);
     (try destac_aux);
     (try monadtac_aux);
     (try maptac_aux)).

(** Binary Operations. For now, we support only addition and bitwise
and operations. *)

Module Bop.
  Inductive t: Type :=
  | plus
  | minus
  | times
  | band
  | equal
  .

  Definition eq (x y:t): {x = y} + {x <> y}.
  Proof. decide equality. Defined.

  Definition has_type (bop:t) (t1 t2:Typ.t): option Typ.t :=
    match bop, t1, t2 with
      | plus, Typ.int, Typ.int => Some Typ.int
      | minus, Typ.int, Typ.int => Some Typ.int
      | times, Typ.int, Typ.int => Some Typ.int
      | band, Typ.int, Typ.int => Some Typ.int
      | equal, Typ.int, Typ.int => Some Typ.int

      | plus, Typ.int, Typ.ptr => Some Typ.ptr
      | plus, Typ.ptr, Typ.int => Some Typ.ptr
      | minus, Typ.ptr, Typ.int => Some Typ.ptr
      | minus, Typ.ptr, Typ.ptr => Some Typ.int
      | equal, Typ.ptr, Typ.ptr => Some Typ.int

      | _, _, _ => None
    end.
End Bop.

Module Var.
  Definition t := (Typ.t * ident)%type.
  Definition eq (x y:t): {x = y} + {x <> y} := pair_eq Typ.eq ident_eq x y.
End Var.

Module Exp.
  Inductive bt: Type :=
  | const (const:int)
  | global (id:ident)
  | local (var:Var.t)
  .

  Inductive t: Type :=
  | base (bexp: bt) 
  | bop (bop:Bop.t) (t1:Typ.t) (e1:t) (t2:Typ.t) (e2:t)
  .

  Coercion Exp.const: int >-> Exp.bt.
  Coercion Exp.local: Var.t >-> Exp.bt.
  Coercion Exp.global: ident >-> Exp.bt.
  Coercion Exp.base: Exp.bt >-> Exp.t.

  Definition beq (x y:bt): {x = y} + {x <> y}.
  Proof.
    decide equality;
    (try by apply Int.eq_dec);
    (try by apply Var.eq);
    (try by apply ident_eq).
  Defined.

  Definition eq (x y:t): {x = y} + {x <> y}.
  Proof.
    decide equality;
    (try by apply beq);
    (try by apply Typ.eq);
    (try by apply Bop.eq).
  Defined.

  Fixpoint has_type (globals:list ident) (locals:list Var.t) (e:t) {struct e}: option Typ.t :=
    match e with
      | const _ => Some Typ.int
      | global gvar =>
        assertB (bin ident_eq gvar globals);
        Some Typ.ptr
      | local var =>
        assertB (bin Var.eq var locals);
        Some (fst var)
      | bop b typ1 e1 typ2 e2 =>
        match has_type globals locals e1, has_type globals locals e2 with
          | Some t1, Some t2 =>
            assertB (Typ.eq typ1 t1);
            assertB (Typ.eq typ2 t2);
            Bop.has_type b t1 t2
          | _, _ => None
        end
    end.

  Lemma has_type_mon
        e typ
        globals1 globals2 (Hg: list_subset globals1 globals2)
        locals1 locals2 (Hl: list_subset locals1 locals2)
        (H1: Exp.has_type globals1 locals1 e = Some typ):
    Exp.has_type globals2 locals2 e = Some typ.
  Proof.
    generalize dependent typ.
    generalize dependent e. induction e; intros.
    - destruct bexp; tac.
      + erewrite bin_subset; eauto; tac.
      + erewrite bin_subset; eauto; tac.
    - inv H1. tac.
      erewrite IHe1; auto.
      erewrite IHe2; auto.
      by rewrite Hcond1, Hcond2.
  Qed.
End Exp.
Coercion Exp.const: int >-> Exp.bt.
Coercion Exp.local: Var.t >-> Exp.bt.
Coercion Exp.global: ident >-> Exp.bt.
Coercion Exp.base: Exp.bt >-> Exp.t.

Ltac wf_montac :=
  repeat
    (simpl in *; subst;
     (try done);
     (try basictac_aux);
     try match goal with
           | [H: context[match Exp.has_type ?a ?b ?c with Some _ => _ | None => _ end] |- _] =>
             let Hcond := fresh "Hcond" in
             destruct (Exp.has_type a b c) eqn:Hcond
           | [H: context[match ?t with Typ.int => _ | Typ.ptr => _ end] |- _] =>
             let Hcond := fresh "Hcond" in
             destruct t eqn:Hcond
           | [Hg: list_subset ?g1 _,
              H: Exp.has_type ?g1 ?l1 ?e = Some ?t |- _] =>
             eapply Exp.has_type_mon in H; eauto
           | [H: Exp.has_type ?g1 ?l1 ?e = Some ?t |-
              context[Exp.has_type ?g1 ?l1 ?e]] =>
             rewrite H
         end).

(** (Right) expressions that may be effectful (event and memory change).
 - malloc:
   + align is the number of bits that is guaranteed to be zero from the
LSB. For example, if align=3, the returned pointer p & 0x07 = 0.
 - input system call: arbitrary integer is returned.
 - output system call: argument is sent outside.
*)
Module RExp.
  Inductive t: Type :=
  | exp (exp:Exp.t)
  | malloc (size:Exp.t) (align:int)
  | free (exp:Exp.t)
  | cast (typ:Typ.t) (exp:Exp.t)
  | input
  | output (exp:Exp.t)
  .

  Definition eq (x y:t): {x = y} + {x <> y}.
  Proof.
    decide equality;
    (try by apply Int.eq_dec);
    (try by apply ident_eq);
    (try by apply bool_dec);
    (try by apply Typ.eq);
    (try by apply Exp.eq).
  Defined.

  Fixpoint has_type (globals:list ident) (locals:list Var.t) (rexp:t) {struct rexp}: option Typ.t :=
    match rexp with
      | exp e => Exp.has_type globals locals e
      | malloc size _ =>
        match Exp.has_type globals locals size with
          | Some Typ.int => Some Typ.ptr
          | _ => None
        end
      | free ptr =>
        match Exp.has_type globals locals ptr with
          | Some Typ.ptr => Some Typ.int
          | _ => None
        end
      | cast typ e =>
        if Exp.has_type globals locals e
        then Some typ
        else None
      | input => Some Typ.int
      | output e =>
        match Exp.has_type globals locals e with
          | Some Typ.int => Some Typ.int
          | _ => None
        end
    end.
  Lemma has_type_mon
        re typ
        globals1 globals2 (Hg: list_subset globals1 globals2)
        locals1 locals2 (Hl: list_subset locals1 locals2)
        (H1: has_type globals1 locals1 re = Some typ):
    has_type globals2 locals2 re = Some typ.
  Proof.
    destruct re; intros; wf_montac.
  Qed.
End RExp.
Coercion RExp.exp: Exp.t >-> RExp.t.

(** Instructions.
 - `while` and `if` instructions may contain list of instructions.
 - `realized` instruction makes sure that `ptr` is physically realized
 *)
Module Instr.
  Inductive t: Type :=
  | initialize (locals: list Var.t)
  | call (fid:ident) (args:list Exp.bt)
  | assign (var:option Var.t) (rhs:RExp.t)
  | load (var:option Var.t) (rhs:Exp.t)
  | store (lhs:Exp.t) (rhs:Exp.t)
  | while (cond:Exp.t) (body:list t)
  | ifte (cond:Exp.t) (s1 s2:list t)
  .

  Definition is_call (i:t): bool :=
    match i with
      | call _ _ => true
      | _ => false
    end.

  (** Recursion/Induction principles *)
  Definition rect:
    forall P : t -> Type,
      (forall (locals: list Var.t), P (initialize locals)) ->
      (forall (fid : ident) (args : list Exp.bt), P (call fid args)) ->
      (forall (var : option Var.t) (rhs : RExp.t), P (assign var rhs)) ->
      (forall (var : option Var.t) (rhs : Exp.t), P (load var rhs)) ->
      (forall (lhs : Exp.t) (rhs : Exp.t), P (store lhs rhs)) ->
      (forall (cond : Exp.t) (body : list t) (Hbody: ForallT P body), P (while cond body)) ->
      (forall (cond : Exp.t) (s1 s2 : list t) (Hs1: ForallT P s1) (Hs2: ForallT P s2), P (ifte cond s1 s2)) ->
      forall i : t, P i.
  Proof.
    intros. revert i. fix IH 1.
    intro i. destruct i; try (by clear IH; eauto).
    - apply X4.
      induction body; [by clear IH; auto|].
      constructor; [|apply IHbody].
      apply IH.
    - apply X5.
      + induction s1; [by clear IH; auto|].
        constructor; [|apply IHs1].
        apply IH.
      + induction s2; [by clear IH; auto|].
        constructor; [|apply IHs2].
        apply IH.
  Defined.

  Lemma ind:
    forall P : t -> Prop,
      (forall (locals: list Var.t), P (initialize locals)) ->
      (forall (fid : ident) (args : list Exp.bt), P (call fid args)) ->
      (forall (var : option Var.t) (rhs : RExp.t), P (assign var rhs)) ->
      (forall (var : option Var.t) (rhs : Exp.t), P (load var rhs)) ->
      (forall (lhs : Exp.t) (rhs : Exp.t), P (store lhs rhs)) ->
      (forall (cond : Exp.t) (body : list t) (Hbody: Forall P body), P (while cond body)) ->
      (forall (cond : Exp.t) (s1 s2 : list t) (Hs1: Forall P s1) (Hs2: Forall P s2), P (ifte cond s1 s2)) ->
      forall i : t, P i.
  Proof.
    intros. revert i. fix IH 1.
    intro i. destruct i; try (by clear IH; eauto).
    - apply H4.
      induction body; [by clear IH; auto|].
      constructor; [|apply IHbody].
      apply IH.
    - apply H5.
      + induction s1; [by clear IH; auto|].
        constructor; [|apply IHs1].
        apply IH.
      + induction s2; [by clear IH; auto|].
        constructor; [|apply IHs2].
        apply IH.
  Qed.

  Definition rec
             (T: Type)
             (Hinitialize: forall (locals: list Var.t), T)
             (Hcall: forall (fid : ident) (args : list Exp.bt), T)
             (Hassign: forall (var : option Var.t) (rhs : RExp.t), T)
             (Hload: forall (var : option Var.t) (rhs : Exp.t), T)
             (Hstore: forall (lhs : Exp.t) (rhs : Exp.t), T)
             (Hwhile: forall (cond : Exp.t) (body : list t) (Hbody: list T), T)
             (Hifte: forall (cond : Exp.t) (s1 s2 : list t) (Hs1: list T) (Hs2: list T), T):
    forall (i: t), T :=
    fix rec (i : t) : T :=
    match i with
      | initialize locals => Hinitialize locals
      | call fid args => Hcall fid args
      | assign var rhs => Hassign var rhs
      | load var rhs => Hload var rhs
      | store lhs rhs => Hstore lhs rhs
      | while cond body => Hwhile cond body (map rec body)
      | ifte cond s1 s2 => Hifte cond s2 s2 (map rec s1) (map rec s2)
    end.

  Definition eq (x y:t): {x = y} + {x <> y}.
  Proof.
    generalize dependent y.
    induction x using rect;
      intro y; destruct y; (try by right); fold Var.t in *;
    repeat
      match goal with
        | [x1: Typ.t, x2: Typ.t |- _] =>
          destruct (Typ.eq x1 x2);
            [subst|by right; intro X; inv X]
        | [x1: ident, x2: ident |- _] =>
          destruct (ident_eq x1 x2);
            [subst|by right; intro X; inv X]
        | [x1: Var.t, x2: Var.t |- _] =>
          destruct (Var.eq x1 x2);
            [subst|by right; intro X; inv X]
        | [x1: option Var.t, x2: option Var.t |- _] =>
          destruct (option_eq Var.eq x1 x2);
            [subst|by right; intro X; inv X]
        | [x1: Typ.t * ident, x2: Typ.t * ident |- _] =>
          destruct (pair_dec Typ.eq ident_eq x1 x2);
            [subst|by right; intro X; inv X]
        | [x1: Exp.t, x2: Exp.t |- _] =>
          destruct (Exp.eq x1 x2);
            [subst|by right; intro X; inv X]
        | [x1: RExp.t, x2: RExp.t |- _] =>
          destruct (RExp.eq x1 x2);
            [subst|by right; intro X; inv X]
        | [x1: list Exp.bt, x2: list Exp.bt |- _] =>
          destruct (list_eq_dec Exp.beq x1 x2);
            [subst|by right; intro X; inv X]
        | [x1: list Var.t, x2: list Var.t |- _] =>
          destruct (list_eq_dec Var.eq x1 x2);
            [subst|by right; intro X; inv X]
      end;
    try by left.
    - destruct (ForallT_eq body0 Hbody);
      [subst|by right; intro X; inv X].
      by left.
    - destruct (ForallT_eq s0 Hs1);
      [subst|by right; intro X; inv X].
      destruct (ForallT_eq s3 Hs2);
      [subst|by right; intro X; inv X].
      by left.
  Defined.

  Definition wf (globals:list ident) (locals:list Var.t): forall (istr:t), bool :=
    rec
      bool
      (fun locals => (* initialize *)
         false)
      (fun fid args => (* call *)
         for_all
           (fun (e:Exp.bt) => Exp.has_type globals locals e)
           args)
      (fun var rhs => (* assign *)
         match var with
           | None =>
             RExp.has_type globals locals rhs
           | Some var =>
             bin Var.eq var locals &&
             option_eq Typ.eq (RExp.has_type globals locals rhs) (Some (fst var))
         end)
      (fun var rhs => (* load *)
         (match var with
            | None => true
            | Some var => bin Var.eq var locals
          end) &&
         (option_eq Typ.eq (Exp.has_type globals locals rhs) (Some Typ.ptr)))
      (fun lhs rhs => (* store *)
         (option_eq Typ.eq (Exp.has_type globals locals lhs) (Some Typ.ptr)) &&
         (Exp.has_type globals locals rhs))
      (fun cond body Hbody => (* while *)
         for_all id Hbody &&
         option_eq Typ.eq (Exp.has_type globals locals cond) (Some Typ.int))
      (fun cond s1 s2 Hs1 Hs2 => (* ifte *)
         for_all id Hs1 && for_all id Hs2 &&
         (option_eq Typ.eq (Exp.has_type globals locals cond) (Some Typ.int)))
  .
  Lemma wf_mon
        i
        globals1 globals2 (Hg: list_subset globals1 globals2)
        locals1 locals2 (Hl: list_subset locals1 locals2)
        (H1: wf globals1 locals1 i):
    wf globals2 locals2 i.
  Proof.
    generalize dependent H1.
    induction i using Instr.ind; intros; eauto.
    - apply for_all_spec. intros.
      eapply for_all_spec in H1; eauto.
      destruct (Exp.has_type globals1 locals1 a) eqn:Hwf; [|done].
      erewrite Exp.has_type_mon; eauto.
    - destruct var; wf_montac.
      + erewrite RExp.has_type_mon; eauto.
        split.
        * eapply bin_subset; eauto; tac.
        * destruct t0. tac.
      + destruct (RExp.has_type globals1 locals1 rhs) eqn:Hwf; [|done].
        erewrite RExp.has_type_mon; eauto.
    - destruct var; wf_montac. split.
      + erewrite bin_subset; eauto; tac.
      + wf_montac.
    - apply andb_true_iff in H1. destruct H1.
      apply andb_true_iff. split.
      + wf_montac.
      + destruct (Exp.has_type globals1 locals1 rhs) eqn:Hrhs1; tac.
        wf_montac.
    - simpl in *. apply andb_true_iff in H1. destruct H1.
      tac. split.
      + generalize dependent Hbody. induction body; intros; tac.
        inv Hbody. split.
        * by apply H3.
        * by apply IHbody.
      + wf_montac.
    - tac. split. tac. split.
      + generalize dependent H1.
        induction s1; intros; tac.
        inv Hs1. split; auto.
        by apply H4.
      + generalize dependent H.
        induction s2; intros; tac.
        inv Hs2. split; auto.
        by apply H4.
      + wf_montac.
  Qed.
End Instr.

(** Function declarations consist of function id, parameters, variable
declarations, and instructions. *)

Module FDecl.
  Structure t: Type := mk {
    id: ident;
    params: list Var.t;
    locals: list Var.t;
    body: list Instr.t
  }.
  Definition wf (globals:list ident) (f:t): bool :=
    for_all (Instr.wf globals (f.(params) ++ f.(locals))) f.(body).
  Lemma wf_mon globals1 globals2 f
        (Hsubset: list_subset globals1 globals2)
        (H1: wf globals1 f):
    wf globals2 f.
  Proof.
    apply for_all_spec. intros.
    eapply for_all_spec in H1; eauto.
    eapply Instr.wf_mon; eauto.
    reflexivity.
  Qed.
  Definition byid (fid:ident) (f:t): bool :=
    peq fid f.(id).
End FDecl.

(** Programs consist of variable declarations and function
declarations. Function id 1 is assumed to be the "main" function. *)

Module Prog.
  Structure t: Type := mk {
    globals: list (ident * int);
    fdecls: list FDecl.t
  }.
  Definition wf (p:t): bool :=
    for_all (FDecl.wf (map fst p.(globals))) p.(fdecls).
End Prog.

(** Notations for program. *)

Ltac wftac :=
  unfold Prog.wf, FDecl.wf in *; simpl;
  repeat
    (simpl in *; subst; unfold id in *;
     (try done);
     (try basictac_aux);
     (try destac_aux);
     try match goal with
           | [|- _ /\ _] =>
             split
           | [|- context[bin _ ?x (cons ?x _)]] =>
             rewrite bin_left; auto
           | [|- context[bin _ ?x (cons ?y _)]] =>
             rewrite bin_right; auto
         end).

Module SyntaxNotations.
  Export ListNotations.

  Delimit Scope string_scope with string.
  Open Scope string_scope.

  Notation "'%g' var" := (ident_of_string var) (at level 41).
  Notation "'%f' var" := (ident_of_string var) (at level 41).
  Notation "'%i' var" := ((Typ.int, (ident_of_string var)):Var.t) (at level 41).
  Notation "'%p' var" := ((Typ.ptr, (ident_of_string var)):Var.t) (at level 41).

  Notation "e1 'PLUSii' e2" := (Exp.bop Bop.plus Typ.int e1 Typ.int e2) (at level 41).
  Notation "e1 'PLUSip' e2" := (Exp.bop Bop.plus Typ.int e1 Typ.ptr e2) (at level 41).
  Notation "e1 'PLUSpi' e2" := (Exp.bop Bop.plus Typ.ptr e1 Typ.int e2) (at level 41).
  Notation "e1 'MINUSii' e2" := (Exp.bop Bop.minus Typ.int e1 Typ.int e2) (at level 41).
  Notation "e1 'MINUSpi' e2" := (Exp.bop Bop.minus Typ.ptr e1 Typ.int e2) (at level 41).
  Notation "e1 'MINUSpp' e2" := (Exp.bop Bop.minus Typ.ptr e1 Typ.ptr e2) (at level 41).
  Notation "e1 'TIMES' e2" := (Exp.bop Bop.times Typ.int e1 Typ.int e2) (at level 41).
  Notation "e1 'AND' e2" := (Exp.bop Bop.band Typ.int e1 Typ.int e2) (at level 41).
  Notation "e1 'EQii' e2" := (Exp.bop Bop.equal Typ.int e1 Typ.int e2) (at level 41).
  Notation "e1 'EQpp' e2" := (Exp.bop Bop.equal Typ.ptr e1 Typ.ptr e2) (at level 41).

  Notation "'MALLOC' e1" := (RExp.malloc e1 0) (at level 41).
  Notation "'FREE' e1" := (RExp.free e1) (at level 41).
  Notation "'CASTi' e" := (RExp.cast Typ.int e) (at level 41).
  Notation "'CASTp' e" := (RExp.cast Typ.ptr e) (at level 41).
  Notation "'INPUT'" := (RExp.input) (at level 41).
  Notation "'OUTPUT' e1" := (RExp.output e1) (at level 41).

  Notation "'CALL' fid 'WITH' args" := (Instr.call fid args) (at level 42).
  Notation "var '::=' rhs" := (Instr.assign (Some var) rhs) (at level 42).
  Notation "'_::=' rhs" := (Instr.assign None rhs) (at level 42).
  Notation "var '<--' rhs" := (Instr.load (Some var) rhs) (at level 42).
  Notation "'_<--' rhs" := (Instr.load None rhs) (at level 42).
  Notation "rhs '-->' lhs" := (Instr.store lhs rhs) (at level 42).
  Notation "'WHILE' c 'DO' body 'END'" := (Instr.while c body) (at level 43).
  Notation "'IF' c 'THEN' s1 'ELSE' s2 'END'" := (Instr.ifte c s1 s2) (at level 43).

  Notation "'FUNCTION' fid 'WITH' params 'LOCALS' locals 'BEGIN' body 'END'" := (FDecl.mk fid params locals body) (at level 44).

  Notation "'GLOBALS' globals 'FUNCTIONS' fdecls" := (Prog.mk globals fdecls) (at level 45).

  Program Definition program_example :=
    GLOBALS [
      (%g"g", 1)
    ]

    FUNCTIONS [
      FUNCTION %f"foo" WITH [%i"x"; %i"y"]
      LOCALS [%i"z"] BEGIN [
        WHILE 1 DO [
          %i"z" ::= %i"x" EQii %i"y";
          %i"z" --> %g"g"
        ] END
      ] END;

      FUNCTION %f"main" WITH []
      LOCALS [%i"x"; %i"y"; %i"z"] BEGIN [
        WHILE (%i"x" EQii %i"y") AND 1 DO [
          %i"x" ::= 1 PLUSii 1;
          %i"y" ::= 1 AND 1;
          CALL %f"foo" WITH [%i"x"; %i"y"];
          %i"z" <-- %g"g";
          _::= OUTPUT %i"z"
        ] END
      ] END
    ]
  .

  Lemma program_example_wf: Prog.wf program_example.
  Proof. wftac. Qed.
End SyntaxNotations.
