Require Import Arith Bool List Program.


(** [typeof t] is similar to [type of t] but is an expression instead
    of a tactic.  This for example allows [unify (typeof t) T] instead
    of [let T' := type of t in unify T' T].
*)
Definition typeof {T} (t: T) := T.


(** Get the last (bottom) hypothesis. *)
Ltac last_hyp := match goal with H : _ |- _ => H end.



(** * Operating on hypotheses.

    This is a hack to get things like iteration over the hypotheses of
    the current proof context.  Found this in the coq-club archive in
    a post by Eric Jaeger:
    http://permalink.gmane.org/gmane.science.mathematics.logic.coq.club/5128

    The key idea is a tactic that loads all hypotheses into a
    simple list data structure.
*)


(** We store hypotheses as a list of type-value pairs. *)
Inductive _ctx :=
| _ctx_nil: _ctx
| _ctx_cons: forall (T: Type), T -> _ctx -> _ctx
.


(** Reverse. *)
Ltac _ctx_rev src tgt :=
  match src with
    | _ctx_nil => tgt
    | _ctx_cons ?T ?t ?src' => _ctx_rev src' (_ctx_cons T t tgt)
  end.


(** Test membership. *)
Ltac _ctx_mem T t c :=
  match c with
    | _ctx_nil => false
    | _ctx_cons ?T' ?t' ?c' =>
      match t with t' => true | _ => _ctx_mem T t c' end
      (* match T with *)
      (*   | T' => match t with t' => true | _ => _ctx_mem T t c' end *)
      (*   | _ => _ctx_mem T t c' *)
      (* end *)
  end.


(** Fetch the current context. *)
Ltac _ctx_mk acc :=
  match goal with
    | [H: ?T |- _] =>
      match _ctx_mem T H acc with false =>
        _ctx_mk (_ctx_cons T H acc)
      end
    | _ => acc
  end.


Ltac Hdo_ tac hyps :=
  match hyps with
    | _ctx_nil => idtac
    | _ctx_cons _ ?H ?hyps' => tac H; Hdo_ tac hyps'
  end.


(** Iterate over all hypotheses (bottom-up), each time applying [tac]. *)
Tactic Notation "Hdo" tactic(tac) :=
  let hyps := _ctx_rev ltac:(_ctx_mk _ctx_nil) _ctx_nil in
  Hdo_ tac hyps.

(** Iterate over the given hypotheses, each time applying [tac]. *)
Tactic Notation "Hdo" tactic(tac) constr(hyps) :=
  Hdo_ tac hyps.

Notation "$ H1 .. Hn" :=
  (_ctx_cons _ H1 .. (_ctx_cons _ Hn _ctx_nil) ..)
    (at level 0, H1 at level 0, Hn at level 0).



(** * Masking of hypotheses. *)

Ltac mask M H :=
  match type of H with
    | M _ _ => fail 1
    | ?a => first [ change a with (M _ a) in H
                  | replace a with (M _ a) in H; [|solve [auto]] ]
  end.

(* Ltac mask_all M := *)
(*   Hdo (mask M). *)

Ltac unmask M H :=
  match type of H with
    | M ?A ?a => first [ change (M A a) with a in H
                       | replace (M A a) with a in H; [|solve [auto]] ]
  end.

(* Ltac unmask_all M := *)
(*   Hdo (unmask M). *)

(* Ltac remask M M' H := *)
(*   match type of H with *)
(*     | M ?A ?a => first [ change (M A a) with (M' A a) in H *)
(*                        | replace (M A a) with (M' A a) in H; [|solve [auto]] ] *)
(*     | _ => idtac *)
(*   end. *)
Ltac remask M M' H :=
  unmask M H; mask M' H.



(** * Generalized injection. *)

Create HintDb inj.

Ltac injH H :=
  try match (type of H) with
    | ?f ?x = ?f ?y =>
      let X := fresh H in
        assert (X: (x = y)) by (clear -H; auto 2 with nocore inj)
      ; first [clear H | let H' := fresh H in rename H into H']
      ; rename X into H
  end.

Ltac inj :=
  Hdo (fun H => injH H).


(** ** Populating the [inj] database with basic injection lemmas. *)

Lemma some_inj: forall X (x y: X)
    (E: Some x = Some y),
  x = y.
Proof.
  intros; injection E; auto.
Qed.
Hint Resolve some_inj : inj.



(** * Hiding of hypotheses. *)


(** ** Masks. *)

Definition _BURIED_ A (a: A) : A := a.
Arguments _BURIED_ [A] [a] : simpl never.
Global Opaque _BURIED_.
Notation "._."  := _BURIED_.

Definition _HIDDEN_ A (a: A) : A := a.
Arguments _HIDDEN_ [A] [a] : simpl never.
Notation "..."  := _HIDDEN_.

Definition _NEW_ A (a: A) : A := a.
Arguments _NEW_ [A] a : simpl never.


(** ** Hiding some or all. *)

Tactic Notation "bury" :=
  Hdo (fun H => try mask @_BURIED_ H).
Tactic Notation "bury" constr(hyps) :=
  Hdo (fun H => try mask @_BURIED_ H; move H at top) hyps.

Tactic Notation "hide" :=
  Hdo (fun H => try mask @_HIDDEN_ H).
Tactic Notation "hide" constr(hyps) :=
  Hdo (fun H => try mask @_HIDDEN_ H; move H at top) hyps.


(** ** Unhiding some or all. *)

Tactic Notation "unbury" :=
  Hdo (fun H => try remask @_BURIED_ _NEW_ H).
Tactic Notation "unbury" constr(hyps) :=
  Hdo (fun H => try remask @_BURIED_ _NEW_ H; move H at bottom) hyps.

Tactic Notation "unhide" :=
  Hdo (fun H => try unmask @_HIDDEN_ H).
  (* Hdo (fun H => try remask @_HIDDEN_ _NEW_ H). *)
Tactic Notation "unhide" constr(hyps) :=
  Hdo (fun H => try unmask @_HIDDEN_ H; move H at bottom) hyps.
  (* Hdo (fun H => try remask @_HIDDEN_ _NEW_ H; move H at bottom) hyps. *)


(** ** Marked [intros].

    Do [intros] but mark the new hypotheses as such.
*)
Ltac hi :=
  let lasthyp := last_hyp in
  intros;
  let hyps := _ctx_rev ltac:(_ctx_mk _ctx_nil) _ctx_nil in
  let rec hi_ hyps :=
      match hyps with
        | _ctx_nil => idtac
        | _ctx_cons ?T ?H ?hyps' =>
          try match H with
            | lasthyp => fail 1
            | _ => mask _NEW_ H; hi_ hyps'
          end
      end in
  hi_ hyps.
(* TODO generalize hi, allow arbitraty tactic, not just intros, and check for new ones not just at bottom  *)
