Require Import Relations.
Require Import Axioms.
Require Import Coqlib.
Require Import Kanglib.
Require Import paco.
Require Import AST.
Require Import Integers.
Require Import Maps.
Require Import Events.
Require Import Monad.
Require Import WFType.
Require Import vgtac.

(** observation **)
CoInductive observation : Set :=
| obs_term
| obs_pterm
| obs_inftau
| obs_event (evt:Event.t) (obs:observation).

Module Status.
  Inductive t: Type :=
  | error
  | term
  | normal
  .
End Status.

Structure LTS_type : Type := mkLTS_type 
{ Prog: Type
; StaticConfig: Type
; Config: Type 
}.

Structure LTS {ltst: LTS_type} : Type := mkLTS {
  load_prog: forall (prog:ltst.(Prog)) (scfg: ltst.(StaticConfig)) (cfg: ltst.(Config)), Prop;
  get_status: forall (cfg:ltst.(Config)), Status.t;
  step: forall (scfg:ltst.(StaticConfig)) (cfg1:ltst.(Config)) (evt:option Event.t) (cfg2:ltst.(Config)), Prop;

  _step_normal: 
    forall scfg cfg1 evt cfg2 
      (STEP: step scfg cfg1 evt cfg2),
    get_status cfg1 = Status.normal
}.

Section LTS.
  Context {ltst: LTS_type}.
  Variable lts:@LTS ltst.

  Variable scfg: ltst.(StaticConfig).

  Definition taustep cfg1 cfg2 : Prop :=
    lts.(step) scfg cfg1 None cfg2.

  Definition deterministic cfg : Prop :=
    forall evt1 cfg1 evt2 cfg2
      (STEP1: step lts scfg cfg evt1 cfg1)
      (STEP2: step lts scfg cfg evt2 cfg2),
    evt1 = evt2 /\ cfg1 = cfg2.

  Inductive indexed_step
            (cfg1:Indexed.t ltst.(Config))
            (evt:option Event.t)
            (cfg2:Indexed.t ltst.(Config)): Prop :=
  | indexed_step_index
      (Hindex: WF.rel cfg2.(Indexed.index) cfg1.(Indexed.index))
      (Helt: cfg1.(Indexed.elt) = cfg2.(Indexed.elt))
      (Hevt: evt = None)
  | indexed_step_step
      (Hstep: lts.(step) scfg cfg1.(Indexed.elt) evt cfg2.(Indexed.elt))
  .

  Inductive behmatch behave: ltst.(Config) -> observation -> Prop :=
  | behmatch_error
      obs cfg
      (Herror: lts.(get_status) cfg = Status.error):
      behmatch behave cfg obs
  | behmatch_partial
      cfg:
      behmatch behave cfg obs_pterm
  | behmatch_term
      cfg
      (Hterm: lts.(get_status) cfg = Status.term):
      behmatch behave cfg obs_term
  | behmatch_inftau
      cfg1 cfg2
      (Hnormal: lts.(get_status) cfg1 = Status.normal)
      (Hs: taustep cfg1 cfg2)
      (Hs': behave cfg2 obs_inftau : Prop):
      behmatch behave cfg1 obs_inftau
  | behmatch_event
      evt obs cfg1 cfg2
      (Hnormal: lts.(get_status) cfg1 = Status.normal)
      (Hs: step lts scfg cfg1 (Some evt) cfg2)
      (Hs': behave cfg2 obs):
      behmatch behave cfg1 (obs_event evt obs)
  .
  Hint Constructors behmatch.

  Lemma behmatch_mon: monotone2 behmatch.
  Proof.
    repeat intro; destruct IN; eauto.
  Qed.
  Hint Resolve behmatch_mon: paco.

  Inductive _behave behave cfg obs : Prop :=
  | _behave_
      cfg2
      (TAU: rtc taustep cfg cfg2)
      (MAT: behmatch behave cfg2 obs).
  Hint Constructors _behave.

  Lemma behave_mon: monotone2 _behave.
  Proof.
    repeat intro. destruct IN.
    exploit behmatch_mon; eauto.
  Qed.
  Hint Resolve behave_mon: paco.

  Definition behave: _ -> _ -> Prop := paco2 _behave bot2.

  Lemma behave_taus: forall cfg cfg' obs
      (BEH: behave cfg' obs)
      (TAU: rtc taustep cfg cfg'),
    behave cfg obs.
  Proof.
    intros; pfold; punfold BEH; destruct BEH; eauto using rt_trans.
  Qed.

  Lemma behave_tau_deterministic: forall cfg cfg' obs
      (BEH: behave cfg obs)
      (DET: deterministic cfg)
      (TAU: taustep cfg cfg'),
    behave cfg' obs.
  Proof.
    i; punfold BEH; destruct BEH.
    apply clos_rt_rt1n in TAU0.
    induction TAU0; i; pfold.
    - destruct MAT; pclearbot; eauto.
      + by apply _step_normal in TAU; rewrite Herror in TAU.
      + by apply _step_normal in TAU; rewrite Hterm in TAU.
      + exploit DET; [apply Hs | apply TAU | ].
        by i; des; subst; punfold Hs'. 
      + exploit DET; [apply Hs | apply TAU | ].
        by i; des.
    - exploit DET; [apply H | apply TAU | ].
      i; des; subst.
      eauto using clos_rt1n_rt.
  Qed.

End LTS.
Hint Constructors behmatch.
Hint Constructors _behave.
Hint Resolve behmatch_mon: paco.
Hint Resolve behave_mon: paco.

Definition behave_prog {ltst} (lts: @LTS ltst) (prog: ltst.(Prog)) (obs:observation): Prop :=
  exists scfg cfg,
  lts.(load_prog) prog scfg cfg /\
  behave lts scfg cfg obs.

