Require Import Relations.
Require Import Equality.
Require Import Axioms.
Require Import Classical.
Require Import Coqlib.
Require Import Kanglib.
Require Import paco.
Require Import AST.
Require Import WFType.
Require Import Integers.
Require Import Maps.
Require Import Values.
Require Import Memory.
Require Import MemoryC.
Require Import Events.
Require Import Behaviors.
Require Import Monad.
Require Import Syntax.
Require Import Semantics.
Require Import SemanticsC.
Require Import MemoryRelationC.
Require Import StateRelationC.
Require Import SimulationLocalC.
Require Import SimulationProgWeakC.

Import SyntaxNotations.

Definition removed_dead_cast: forall (istr:Instr.t), bool :=
  Instr.rec
    bool
    (fun locals => (* initialize *)
       true)
    (fun fid args => (* call *)
       true)
    (fun var rhs => (* assign *)
       match var, rhs with
         | None, CASTi _ => false
         | _, _ => true
       end)
    (fun var rhs => (* load *)
       true)
    (fun lhs rhs => (* store *)
       true)
    (fun cond body Hbody => (* while *)
       for_all id Hbody)
    (fun cond s1 s2 Hs1 Hs2 => (* ifte *)
       for_all id Hs1 && for_all id Hs2)
.

Definition remove_dead_cast: forall (istr:Instr.t), option Instr.t :=
  Instr.rec
    (option Instr.t)
    (fun locals => (* initialize *)
       Some (Instr.initialize locals))
    (fun fid args => (* call *)
       Some (Instr.call fid args))
    (fun var rhs => (* assign *)
       match var, rhs with
         | None, CASTi _ => None
         | _, _ => Some (Instr.assign var rhs)
       end)
    (fun var rhs => (* load *)
       Some (Instr.load var rhs))
    (fun lhs rhs => (* store *)
       Some (Instr.store lhs rhs))
    (fun cond body Hbody => (* while *)
       Some (Instr.while cond (gather Hbody)))
    (fun cond s1 s2 Hs1 Hs2 => (* ifte *)
       Some (Instr.ifte cond (gather Hs1) (gather Hs2)))
.

Lemma remove_dead_cast_removed_dead_cast
      (i j:Instr.t)
      (H: remove_dead_cast i = Some j):
  removed_dead_cast j.
Proof.
  generalize dependent j.
  induction i using Instr.ind; intros; (try by inv H).
  - inv H.
    destruct var, rhs; inv H1; auto.
    destruct typ; inv H0; auto.
  - inv H. generalize dependent Hbody.
    induction body; intro H; inv H; auto. simpl.
    destruct (remove_dead_cast a) eqn:Ha; auto.
    simpl. exploit H2; [by eauto|]. intro X. rewrite X. auto.
  - inv H. simpl. apply andb_true_iff. split.
    + generalize dependent s1. clear.
      induction s1; intros H; inv H; auto.
      simpl. destruct (remove_dead_cast a) eqn:Ha; auto. 
      simpl. exploit H2; [by eauto|]. intro X. rewrite X. auto.
    + generalize dependent s2. clear.
      induction s2; intros H; inv H; auto.
      simpl. destruct (remove_dead_cast a) eqn:Ha; auto. 
      simpl. exploit H2; [by eauto|]. intro X. rewrite X. auto.
Qed.

Definition remove_dead_cast_func func :=
  (FDecl.mk func.(FDecl.id) func.(FDecl.params) func.(FDecl.locals) (gather (map remove_dead_cast func.(FDecl.body)))).

Lemma remove_dead_cast_func_wf
      globals func (Hwf: FDecl.wf globals func):
  FDecl.wf globals (remove_dead_cast_func func).
Proof.
  unfold FDecl.wf in *. simtac.
  revert Hwf.
  generalize (FDecl.body func).
  generalize (FDecl.params func ++ FDecl.locals func). intro l.
  clear func.
  induction l0; intros; simtac.
  exploit IHl0; eauto. intro X.
  destruct (remove_dead_cast a) eqn:Ha; auto.
  apply andb_true_iff. split; auto. clear -H Ha.
  generalize dependent t.
  generalize dependent H.
  induction a using Instr.ind; intros; inv Ha; simtac.
  - destruct var; simtac. destruct rhs; simtac.
  - generalize dependent body. induction body; intros; simtac.
    inv Hbody. exploit IHbody; [by eauto|by eauto|].
    destruct (remove_dead_cast a) eqn:Ha; simtac.
    exploit H3; [by eauto|by eauto|]. intro X. by rewrite X.
  - split.
    + generalize dependent s1. induction s1; intros; simtac.
      inv Hs1. exploit IHs1; [by eauto|by eauto|].
      destruct (remove_dead_cast a) eqn:Ha; simtac.
      exploit H4; [by eauto|by eauto|]. intro X. by rewrite X.
    + generalize dependent s2. induction s2; intros; simtac.
      inv Hs2. exploit IHs2; [by eauto|by eauto|].
      destruct (remove_dead_cast a) eqn:Ha; simtac.
      exploit H4; [by eauto|by eauto|]. intro X. by rewrite X.
Qed.

Module Aux.
Require Import vgtac.
Lemma mrel_instr
      istr_src istr_tgt istrs_src istrs_tgt
      (st_src:Semantics.State) (st_tgt:State)
      (mrel:MemRelC.t)
      (sst_src:Semantics.StaticState) (sst_tgt:StaticState)
      event st2_tgt
      (Hncall: Instr.is_call istr_src = false)
      (Hinstr: remove_dead_cast istr_src = Some istr_tgt)
      (Hinstrs_src: st_src.(Semantics.context).(Semantics.instrs) = istr_src::istrs_src)
      (Hinstrs_tgt: st_tgt.(context).(instrs) = istr_tgt::istrs_tgt)
      (Histrs: gather (map remove_dead_cast istrs_src) = istrs_tgt)
      (Hmrel_genv: mrel_genv mrel sst_src.(Semantics.genv) sst_tgt.(genv))
      (Hmrel_env: mrel_env mrel st_src.(Semantics.mem) st_tgt.(mem) st_src.(Semantics.context).(Semantics.lenv) st_tgt.(context).(lenv))
      (Hmrel: mrel st_src.(Semantics.mem) st_tgt.(mem))
      (Hwf_src: Semantics.Context_wf sst_src st_src.(Semantics.context))
      (Hwf_tgt: Context_wf sst_tgt st_tgt.(context))
      (Htgt: semC_instr sst_tgt st_tgt (Some (event, st2_tgt))):
  <<Herror: sem_instr sst_src st_src None>> \/
  exists st2_src (mrel2:MemRelC.t),
    <<Hst2_src: sem_instr sst_src st_src (Some (event, st2_src))>> /\
    <<Hinstrs2: gather (map remove_dead_cast st2_src.(Semantics.context).(Semantics.instrs)) = st2_tgt.(context).(instrs)>> /\
    <<Hmrel2: mrel2 st2_src.(Semantics.mem) st2_tgt.(mem)>> /\
    <<Hmrel2_le: MemRelC.le_public mrel mrel2>> /\
    <<Hmrel_env: mrel_env mrel2 st2_src.(Semantics.mem) st2_tgt.(mem) st2_src.(Semantics.context).(Semantics.lenv) st2_tgt.(context).(lenv)>>.
Proof.
  unfold semC_instr in Htgt. destruct istr_src; tac.
  { inv Hwf_src; inv Hinstrs. rtac.
    inv Hwf_tgt; inv Hinstrs. rtac.
    unfold sem_instr in *. rtac.
    exploit mrel_Env_init; eauto. intro. rtac.
    right. eexists. exists mrel. splits; rtac.
    - eapply bindPP_right; eauto.
    - rtac.
    - rtac.
    - reflexivity.
    - rtac.
  }
  { inv Hwf_src. clear Hwf_src. inv Hinstrs. clear Hinstrs.
    unfold sem_instr. rtac.
    destruct var; rtac.
    - inv Hwf_tgt. clear Hwf_tgt. inv Hinstrs. clear Hinstrs. rtac.
      destruct Hr2; rtac.
      exploit mrel_rexp; eauto; rtac.
      intros [X|X]; [left|right]; des.
      { by apply bindPP_left_None. }
      eexists. exists mrel2. splits; eauto.
      + eapply bindPP_right; eauto.
        right. split; eauto. rtac.
        exploit (sem_rexp_type (Semantics.genv sst_src)); eauto.
      + rtac.
      + rtac.
      + rtac. apply mrel_Env_set; rtac.
        eapply (mrel_env_le lenv1 lenv0 mem1); eauto.
        apply MemRelC.le_public_le. eauto.
    - assert (istr_tgt = _::=rhs); rtac.
      { destruct rhs; rtac. }
      inv Hwf_tgt. clear Hwf_tgt. inv Hinstrs. clear Hinstrs. rtac.
      exploit mrel_rexp; eauto; rtac.
      intros [X|X]; [left|right]; des.
      { by apply bindPP_left_None. }
      eexists. exists mrel2. splits; eauto.
      + eapply bindPP_right; eauto.
      + rtac.
      + rtac.
      + rtac. eapply (mrel_env_le lenv1 lenv mem1); eauto.
        apply MemRelC.le_public_le. eauto.
  }
  { inv Hwf_src. clear Hwf_src. inv Hinstrs. clear Hinstrs.
    unfold sem_instr. rtac.
    destruct var; rtac.
    - inv Hwf_tgt. clear Hwf_tgt. inv Hinstrs. clear Hinstrs. rtac.
      destruct Hr2; rtac.
      exploit mrel_exp; eauto; rtac.
      intro. des. inv Hov; clear Hov.
      { left. econstructor; eauto. }
      destruct (Mem.load a mem1) eqn:Hload_src; rtac.
      { exploit MemRelC.load_public; eauto. intro. rtac.
        rewrite Hcond in H0. inv H0. clear H0.
        destruct (Value.has_type t1 (fst t)) eqn:H1_src.
        Focus 2.
        { left. eapply bindPP_right; eauto.
          rewrite Hload_src. rtac.
          rewrite H1_src. by left.
        }
        Unfocus.
        right. eexists. exists mrel. splits; eauto.
        + eapply bindPP_right; eauto. rewrite Hload_src. rtac.
          right. split; eauto.
        + rtac.
        + rtac.
        + reflexivity.
        + rtac. apply mrel_Env_set; rtac.
      }
      { left. unfold sem_instr. rtac.
        eapply bindPP_right; eauto. by rewrite Hload_src.
      }
    - exploit (mrel_exp rhs); eauto; rtac.
      { inv Hwf_tgt. }
      intro. des. inv Hov; clear Hov.
      { left. econstructor; eauto. }
      destruct (Mem.load a mem1) eqn:Hload_src; rtac.
      { exploit MemRelC.load_public; eauto. intro. rtac.
        right. eexists. exists mrel. splits; eauto.
        + eapply bindPP_right; eauto. rewrite Hload_src. rtac.
        + rtac.
        + rtac.
        + reflexivity.
        + rtac.
      }
      { left. unfold sem_instr. rtac.
        eapply bindPP_right; eauto. by rewrite Hload_src.
      }
  }
  { inv Hwf_src. clear Hwf_src. inv Hinstrs. clear Hinstrs.
    inv Hwf_tgt. clear Hwf_tgt. inv Hinstrs. clear Hinstrs. rtac.
    unfold sem_instr. rtac.
    exploit (mrel_exp lhs); eauto; rtac.
    intro. des. inv Hov.
    { left. econstructor; eauto. }
    exploit (mrel_exp rhs); eauto; rtac.
    intro. des. inv Hov0.
    { left. eapply bindPP_right; eauto. econstructor; eauto. }
    exploit (store_public' mrel mem1 mem0 Hmrel a v1); eauto. intro. des.
    { left. unfold sem_instr. rtac.
      repeat (eapply bindPP_right; eauto). by rewrite x0.
    }
    right. eexists. eexists. splits.
    - repeat (eapply bindPP_right; eauto).
      rewrite x0. rtac.
    - rtac.
    - eauto.
    - constructor; eauto; rtac.
    - rtac.
      repeat intro. exploit mrel_env_le; eauto.
  }
  { inv Hwf_src. clear Hwf_src. inv Hinstrs. clear Hinstrs.
    inv Hwf_tgt. clear Hwf_tgt. inv Hinstrs. clear Hinstrs. rtac.
    unfold sem_instr. rtac.
    exploit (mrel_exp cond); eauto; rtac. intro. des. inv Hov. 
    { left. econstructor; eauto. }
    unfold Mem.val2bool, MemC.val2bool in *. rtac.
    destruct (Mem.val2int a mem1) eqn: Hint_tgt; rtac.
    { exploit MemRelC.val2int_public; eauto. intro X. rewrite X in Hcond0. inv Hcond0.
      right. eexists. exists mrel. splits.
      - repeat (eapply bindPP_right; eauto).
        rewrite Hint_tgt. rtac.
      - rtac.
        destruct (negb (Int.eq i Int.zero)); rtac.
        rewrite map_app, gather_app. f_equal.
      - rtac.
      - reflexivity.
      - rtac.
    }
    { left. unfold sem_instr. rtac.
      repeat (eapply bindPP_right; eauto).
      unfold Mem.val2bool.
      by rewrite Hint_tgt.
    }
  }
  { inv Hwf_src. clear Hwf_src. inv Hinstrs. clear Hinstrs.
    inv Hwf_tgt. clear Hwf_tgt. inv Hinstrs. clear Hinstrs. rtac.
    unfold sem_instr. rtac.
    exploit (mrel_exp cond); eauto; rtac. intro. des.
    inv Hov.
    { left. econstructor; eauto. }
    unfold Mem.val2bool, MemC.val2bool in *. rtac.
    destruct (Mem.val2int a mem1) eqn: Hint_tgt; rtac.
    { exploit MemRelC.val2int_public; eauto. intro X. rewrite X in Hcond0. inv Hcond0.
      right. eexists. exists mrel. splits.
      - repeat (eapply bindPP_right; eauto).
        rewrite Hint_tgt. rtac.
      - rtac.
        rewrite map_app, gather_app.
        destruct (negb (Int.eq i Int.zero)); rtac.
      - rtac.
      - reflexivity.
      - rtac.
    }
    { left. unfold sem_instr. rtac.
      repeat (eapply bindPP_right; eauto).
      unfold Mem.val2bool.
      by rewrite Hint_tgt.
    }
  }
Qed.

Lemma mrel_instr_None
      istr_src istr_tgt istrs_src istrs_tgt
      (st_src:Semantics.State) (st_tgt:State)
      (mrel:MemRelC.t)
      (sst_src:Semantics.StaticState) (sst_tgt:StaticState)
      (Hncall: Instr.is_call istr_src = false)
      (Hinstr: remove_dead_cast istr_src = Some istr_tgt)
      (Hinstrs_src: st_src.(Semantics.context).(Semantics.instrs) = istr_src::istrs_src)
      (Hinstrs_tgt: st_tgt.(context).(instrs) = istr_tgt::istrs_tgt)
      (Hmrel_genv: mrel_genv mrel sst_src.(Semantics.genv) sst_tgt.(genv))
      (Hmrel_env: mrel_env mrel st_src.(Semantics.mem) st_tgt.(mem) st_src.(Semantics.context).(Semantics.lenv) st_tgt.(context).(lenv))
      (Hmrel: mrel st_src.(Semantics.mem) st_tgt.(mem))
      (Hwf_src: Semantics.Context_wf sst_src st_src.(Semantics.context))
      (Hwf_tgt: Context_wf sst_tgt st_tgt.(context))
      (Htgt: semC_instr sst_tgt st_tgt None):
  sem_instr sst_src st_src None.
Proof.
  unfold semC_instr in Htgt. destruct istr_src; tac.
  { inv Hwf_src; inv Hinstrs. rtac.
    inv Hwf_tgt; inv Hinstrs. rtac.
    inv Htgt.
  }
  { inv Hwf_src. clear Hwf_src. inv Hinstrs. clear Hinstrs.
    destruct var; rtac.
    - inv Hwf_tgt. clear Hwf_tgt. inv Hinstrs. clear Hinstrs. rtac.
      inv Htgt; rtac.
      + exploit mrel_rexp_None; eauto; rtac. intro X.
        unfold sem_instr. rtac. by apply bindPP_left_None.
      + destruct Hr2; rtac.
        unfold sem_instr. rtac.
        exploit mrel_rexp; eauto; rtac. intros [X|X]; des.
        { by apply bindPP_left_None. }
        eapply bindPP_right; eauto.
        left. split; rtac. intro X. apply H.
        eapply mrel_has_type; eauto.
    - assert (istr_tgt = _::=rhs); rtac.
      { destruct rhs; rtac. }
      inv Hwf_tgt. clear Hwf_tgt. inv Hinstrs. clear Hinstrs. rtac.
      inv Htgt; rtac.
      exploit mrel_rexp_None; eauto; rtac. intro X.
      unfold sem_instr. rtac. by apply bindPP_left_None.
  }
  { inv Hwf_src. clear Hwf_src. inv Hinstrs. clear Hinstrs.
    inv Hwf_tgt. clear Hwf_tgt. inv Hinstrs. clear Hinstrs. rtac.
    unfold sem_instr. rtac.
    destruct var; rtac.
    - inv Htgt; rtac.
      + exploit mrel_exp; eauto; rtac. intro. des.
        inv Hov. econstructor; eauto.
      + exploit mrel_exp; eauto; rtac. intro. des.
        inv Hov.
        { econstructor; eauto. }
        eapply bindPP_right; eauto.
        destruct (Mem.load a mem0) eqn:H1_src; rtac.
        exploit MemRelC.load_public; eauto. intro. rtac.
        rewrite H in Hr2. rtac. destruct Hr2; rtac.
        left. split; rtac. intro X. apply H1.
        exploit mrel_has_type; eauto.
    - inv Htgt; rtac.
      + exploit mrel_exp; eauto; rtac. intro. des.
        inv Hov. econstructor; eauto.
      + exploit mrel_exp; eauto; rtac. intro. des.
        inv Hov.
        { econstructor; eauto. }
        eapply bindPP_right; eauto.
        destruct (Mem.load a mem0) eqn:H1_src; rtac.
        exploit MemRelC.load_public; eauto. intro. rtac.
        rewrite H in Hr2. rtac.
  }
  { inv Hwf_src. clear Hwf_src. inv Hinstrs. clear Hinstrs.
    inv Hwf_tgt. clear Hwf_tgt. inv Hinstrs. clear Hinstrs. tac.
    inv Htgt; clear Htgt.
    { exploit mrel_exp; try apply Hr1; eauto. by rewrite e0.
      intros [ov_src [Hov_src Hov]]. inv Hov.
      econstructor; eauto.
    }
    exploit (mrel_exp lhs); eauto.
    { by rewrite e0. }
    intro. tac. inv H3.
    { econstructor; eauto. }
    inv Hr2; clear Hr2; tac.
    { exploit mrel_exp; try apply Hr1; eauto. by rewrite e0.
      intros [ov_src [Hov_src Hov]]. inv Hov.
      { econstructor; eauto. }
      exploit mrel_exp; try apply Hr0; eauto.
      intros [ov_src' [Hov_src' Hov']]. inv Hov'.
      eapply bindPP_right; simpl; eauto.
      econstructor; eauto.
    }
    exploit (mrel_exp rhs); eauto.
    intro. tac. inv H6.
    { eapply bindPP_right; eauto. simpl.
      econstructor; eauto.
    }
    unfold sem_instr. tac.
    repeat (eapply bindPP_right; eauto).
    destruct (Mem.store a a0 mem0) eqn:Hstore_src; tac.
    exploit (MemRelC.store_public mrel mem0 mem Hmrel a v1); eauto.
    intro. tac.
    rewrite H7 in Hr3. tac.
  }
  { inv Hwf_src. clear Hwf_src. inv Hinstrs. clear Hinstrs.
    inv Hwf_tgt. clear Hwf_tgt. inv Hinstrs. clear Hinstrs. tac.
    inv Htgt; clear Htgt.
    { exploit mrel_exp; try apply Hr1; eauto. by rewrite e0.
      intros [ov_src [Hov_src Hov]]. inv Hov.
      econstructor; eauto.
    }
    exploit mrel_exp; eauto.
    { by rewrite e0. }
    intro. tac. inv H3.
    { econstructor; eauto. }
    unfold sem_instr. tac.
    eapply bindPP_right; eauto.
    unfold Mem.val2bool, MemC.val2bool in *.
    destruct (Mem.val2int a mem0) eqn:Hint_src; tac.
    exploit MemRelC.val2int_public; eauto. intro X. rewrite X in Hr2. tac.
  }
  { inv Hwf_src. clear Hwf_src. inv Hinstrs. clear Hinstrs.
    inv Hwf_tgt. clear Hwf_tgt. inv Hinstrs. clear Hinstrs. tac.
    inv Htgt; clear Htgt.
    { exploit mrel_exp; try apply Hr1; eauto. by rewrite e0.
      intros [ov_src [Hov_src Hov]]. inv Hov.
      econstructor; eauto.
    }
    exploit mrel_exp; eauto.
    { by rewrite e0. }
    intro. tac. inv H6.
    { econstructor; eauto. }
    unfold sem_instr. tac.
    eapply bindPP_right; eauto.
    unfold Mem.val2bool, MemC.val2bool in *.
    destruct (Mem.val2int a mem0) eqn:Hint_src; tac.
    exploit MemRelC.val2int_public; eauto. intro X. rewrite X in Hr2. tac.
  }
Qed.
End Aux.

Lemma remove_dead_cast_sim_local_func_aux
      sst_src sst_tgt (mrel_init:MemRelC.t)
      foo
      (Hfoo_wf_src: FDecl.wf (PTree.keys sst_src.(Semantics.genv)) foo)
      (Hfoo_wf_tgt: FDecl.wf (PTree.keys sst_tgt.(genv)) foo)
      (Hmrel_init_globals: mrel_genv mrel_init sst_src.(Semantics.genv) sst_tgt.(genv)):
  sim_local_func_aux lts ltsC sst_src sst_tgt mrel_init
                     foo (remove_dead_cast_func foo).
Proof.
  repeat intro.
  exploit mrel_bind_params; eauto.
  intro. vgtac.des.
  assert (X: lenv_tgt = env_tgt); simtac.
  { destruct foo. tac.
    rewrite Henv_tgt in Hlenv_tgt. inv Hlenv_tgt. simtac.
  }
  exists WF.elt.
  cut (forall
          lenv_src lenv_tgt (mrel:MemRelC.t) mem_src mem_tgt istrs_src istrs_tgt
          (Hwf_src: Semantics.Context_wf sst_src (Semantics.mkContext lenv_src istrs_src))
          (Hwf_tgt: Context_wf sst_tgt (mkContext lenv_tgt istrs_tgt))
          (Hmrel_le_public: MemRelC.le_public mrel_entry mrel)
          (Hmrel: mrel mem_src mem_tgt)
          (Hlenv: mrel_env mrel mem_src mem_tgt lenv_src lenv_tgt)
          (Hinstrs: gather (map remove_dead_cast istrs_src) = istrs_tgt),
       sim_local lts ltsC sst_src sst_tgt stack_src stack_tgt mrel_entry mrel WF.elt
                 (Some (Semantics.mkState (Semantics.mkContext lenv_src istrs_src) stack_src mem_src))
                 (Some (mkState (mkContext lenv_tgt istrs_tgt) stack_tgt mem_tgt))).
  { intro X. apply X; simtac.
    - eapply Semantics.Context_wf_before_initialize; simtac.
      + eapply Semantics.bind_params_wf; eauto.
      + apply Forall_forall. intros i Hi.
        eapply for_all_spec in Hfoo_wf_src; eauto.
        eapply Instr.wf_mon; eauto; tac.
        intros v Hv. rewrite in_app in *. destruct Hv as [Hv|Hv].
        * left. exploit (Semantics.bind_params_keys2 (FDecl.params foo) argvals_src); eauto.
          rewrite Env.gkspec. by destruct (Env.get v lenv_src).
        * by right.
    - eapply Context_wf_before_initialize; simtac.
      + eapply bind_params_wf; eauto.
      + exploit remove_dead_cast_func_wf; eauto. unfold FDecl.wf.
        intros. apply Forall_forall. intros i Hi.
        eapply for_all_spec in H; eauto. simtac.
        eapply Instr.wf_mon; eauto.
        * reflexivity.
        * apply list_subset_mon.
          symmetry in Hlenv_tgt.
          generalize (bind_params_keys_subset foo.(FDecl.params) argvals_tgt mem_tgt EnvC.empty env_tgt Hlenv_tgt).
          intros Hsubset x Hx. apply Hsubset. rewrite in_app. by left.
    - reflexivity.
  }
  clear lenv_src env_tgt Hlenv_src Henv_tgt Hlenv_tgt Henv.
  clear mem_src mem_tgt Hmrel_entry Hargvals.
  clear argvals_src argvals_tgt.
  pcofix CIH. intros until istrs_src.
  revert mem_tgt. revert mem_src. revert mrel. revert lenv_tgt. revert lenv_src.
  induction istrs_src as [|istr_src istrs_src]; intros; simtac.
  { pfold. eapply sim_local_return; eauto; simtac.
    - destruct stack_src; simtac.
    - destruct stack_tgt; simtac.
    - reflexivity.
  }
  destruct (classic
              (sem_instr sst_src
                         (Semantics.mkState (Semantics.mkContext lenv_src (istr_src::istrs_src)) stack_src mem_src)
                         None)).
  { pfold. eapply sim_local_error.
    apply rt_step. simtac.
    - instantiate (1 := None). simtac.
    - simtac.
  }
  destruct (Instr.is_call istr_src) eqn:Hcall; simtac.
  { destruct istr_src; simtac.
    destruct (sem_bexp_list (Semantics.genv sst_src) lenv_src args) eqn: Hargvals_src; simtac.
    Focus 2.
    { contradict H. unfold sem_instr. simtac.
      destruct ((Semantics.fenv sst_src) ! fid); simtac.
      apply bindPP_left_None; simtac.
    }
    Unfocus.
    inv Hwf_src; simtac. inv Hinstrs; simtac.
    inv Hwf_tgt; simtac. inv Hinstrs; simtac.
    exploit mrel_bexp_list; simtac.
    { instantiate (3 := mrel). eapply mrel_genv_le; eauto.
      inv Hmrel_init. inv Hmrel_le_public. inv Hequiv. auto.
    }
    { eauto. }
    { eauto. }
    { auto. }
    { auto. }
    { eauto. }
    { eauto. }
    intros. vgtac.des. pfold. eapply sim_local_call; eauto; simtac.
    - reflexivity.
    - intros. exists WF.elt. right.
      exploit (CIH lenv_src lenv_tgt mrel3 st3_src.(Semantics.mem) st3_tgt.(mem)); eauto.
      + instantiate (1 := istrs_src).
        constructor; simtac.
      + constructor; simtac.
      + by rewrite Hmrel_le_public.
      + eapply (mrel_env_le lenv_src lenv_tgt mem_src); eauto.
        by inv Hmrel3_le.
      + destruct st3_src, st3_tgt. simtac.
  }
  destruct (remove_dead_cast istr_src) as [istr_tgt|] eqn:Histr_tgt; simtac.
  { pfold. apply sim_local_step; simtac. intros.
    destruct cfg3_tgt as [st3_tgt|]; simtac.
    Focus 2.
    { contradict H.
      eapply Aux.mrel_instr_None; try apply H0; simtac; auto; [|by eauto|by eauto].
      eapply mrel_genv_le; eauto.
      inv Hmrel_init. inv Hmrel_le_public. inv Hequiv. auto.
    }
    Unfocus.
    exploit
      (Aux.mrel_instr
         istr_src istr_tgt istrs_src (gather (map remove_dead_cast istrs_src))
         (Semantics.mkState (Semantics.mkContext lenv_src (istr_src::istrs_src)) stack_src mem_src)
         (mkState (mkContext lenv_tgt (istr_tgt::(gather (map remove_dead_cast istrs_src)))) stack_tgt mem_tgt)
         mrel);
      simtac; eauto.
    { eapply mrel_genv_le; [by eauto|].
      inv Hmrel_le_public. inv Hequiv.
      inv Hmrel_init. auto.
    }
    intro. vgtac.des.
    { by contradict H. }
    eexists. split; [apply rt_refl|].
    exists WF.elt. eexists. split.
    - apply indexed_step_step. simtac.
      instantiate (1 := Some _). simtac. eauto.
    - exists mrel2. vgtac.splits; simtac.
      + by apply MemRelC.le_public_le.
      + right.
        destruct st2_src as [ctx3_src stack3_src mem3_src]. simtac.
        destruct st3_tgt as [ctx3_tgt stack3_tgt mem3_tgt]. simtac.
        exploit sem_instr_not_call_prop; eauto; simtac. intro. simtac.
        exploit sem_instr_wf; eauto; simtac. intro Hctx3_src_wf.
        move Htgt at bottom.
        exploit semC_instr_not_call_prop; try apply Htgt; simtac.
        { destruct istr_src, istr_tgt; simtac.
          destruct var; simtac. destruct rhs; simtac.
        }
        intro. simtac.
        exploit semC_instr_wf; eauto; simtac. intro Hctx3_tgt_wf.
        exploit (CIH ctx3_src.(Semantics.lenv) ctx3_tgt.(lenv) mrel2 mem3_src mem3_tgt); eauto.
        { by destruct ctx3_src. }
        { by destruct ctx3_tgt. }
        { etransitivity; eauto. }
        { destruct ctx3_src, ctx3_tgt. simtac. }
  }
  destruct istr_src; simtac. destruct var; simtac.
  destruct rhs; simtac. destruct typ; simtac.
  cut (exists mem2_src mrel2,
         MemRelC.le_public mrel mrel2 /\
         mrel2 mem2_src mem_tgt /\
         sem_instr
           sst_src (Semantics.mkState (Semantics.mkContext lenv_src (_::=CASTi exp :: istrs_src)) stack_src mem_src)
           (Some (None, (Semantics.mkState (Semantics.mkContext lenv_src istrs_src) stack_src mem2_src)))).
  { intros [mem2_src [mrel2 [Hmrel2_le [Hmrel2 Hstep]]]].
    exploit IHistrs_src; auto.
    { exploit sem_instr_wf; eauto. simtac. }
    { exploit sem_instr_wf; eauto. simtac. }
    { rewrite Hmrel_le_public. eauto. }
    { eauto. }
    { eapply mrel_env_le; try apply Hlenv; eauto.
      by apply MemRelC.le_public_le.
    }
    Ltac tac :=
      try match goal with
            | [|- rtc _ _ _] =>
              eapply rt_trans; [|by eauto];
              apply rt_step; simtac
          end;
      eauto.
    intro X. punfold X. pfold. inv X.
    - eapply sim_local_error; tac.
    - eapply sim_local_return; tac.
      etransitivity; eauto. by apply MemRelC.le_public_le.
    - eapply sim_local_call; tac.
      etransitivity; eauto. by apply MemRelC.le_public_le.
    - eapply sim_local_step; eauto.
      intros. exploit Hstep0; eauto. intro. simtac.
      eexists. split; [by tac|].
      eexists. eexists. split; eauto.
      eexists. rewrite MemRelC.le_public_le; eauto.
  }
  inv Hwf_src. simtac. inv Hinstrs. simtac.
  destruct (Exp.has_type (PTree.keys (Semantics.genv sst_src)) (Env.keys lenv_src) exp) eqn:Htyp; simtac.
  unfold sem_instr. simtac.
  exploit mrel_exp_has_type; try apply Htyp; try apply Hlenv.
  { eapply mrel_genv_le; [by eauto|].
    intros. inv Hmrel_le_public. inv Hequiv. apply Hmbinrel.
    inv Hmrel_init. by apply Hmbinrel0.
  }
  intro Htyp_tgt.
  unfold sem_instr in H. simpl in H.
  eapply not_bindPP_None in H. des.
  eapply not_bindPP_None in H. des.
  destruct (semC_exp_nonempty (genv sst_tgt) lenv_tgt mem_tgt exp) as [v_tgt Hv_tgt].
  exploit mrel_exp; try apply Hv_tgt; try apply Hlenv; eauto.
  { eapply mrel_genv_le; eauto.
    inv Hmrel_init. inv Hmrel_le_public. inv Hequiv. auto.
  }
  { by inv Hwf_tgt. }
  { by rewrite Htyp. }
  intro. vgtac.des.
  inv Hov; [congruence|].
  rename a into v_src. rename b into v_tgt.
  exploit mrel_cast; eauto.
  { reflexivity. }
  instantiate (1 := Typ.int). intro. vgtac.des.
  { destruct v_src; tac.
    - inv Hcast_src.
    - simpl in Hcast_src. destruct (Mem.normalize b o mem_src); inv Hcast_src.
      inv Hr1. inv Hr1. inv Hr2. inv Hr1. inv Hr0.
  }
  exists mem2_src. exists mrel2. vgtac.splits; eauto.
  repeat (eapply bindPP_right; eauto). eauto.
Qed.
