Require Import Equality.
Require Import Axioms.
Require Import Classical.
Require Import Coqlib.
Require Import Kanglib.
Require Import paco.
Require Import AST.
Require Import WFType.
Require Import Integers.
Require Import Maps.
Require Import Monad.
Require Import Values.
Require Import Events.
Require Import Memory.
Require Import Behaviors.
Require Import Syntax.
Require Import vgtac.
Require Import Vlib.
Require Import hpattern.
Require Export SemanticsBase.

Module Env.
  Structure t: Type := mk {
    int: PTree.t Value.t;
    ptr: PTree.t Value.t
  }.
  Definition empty: t := mk (PTree.empty _) (PTree.empty _).

  Definition get (var:Var.t) (e:t): option Value.t :=
    PTree.get
      (snd var)
      match (fst var) with
        | Typ.int => e.(int)
        | Typ.ptr => e.(ptr)
      end.
  Definition set (var:Var.t) (v:Value.t) (e:t): t :=
    match (fst var) with
      | Typ.int => mk (PTree.set (snd var) v e.(int)) e.(ptr)
      | Typ.ptr => mk e.(int) (PTree.set (snd var) v e.(ptr))
    end.
  Lemma gsspec var1 var2 v e:
    Env.get var1 (Env.set var2 v e) =
    if Var.eq var1 var2
    then Some v
    else Env.get var1 e.
  Proof.
    destruct var1, var2; tac.
    destruct (Typ.eq t0 t1); tac.
    - destruct (ident_eq i i0); tac.
      + unfold get, set. tac.
        destruct t1; tac.
      + unfold get, set. tac.
        destruct t1; tac.
    - unfold get, set. tac.
      destruct t0, t1; tac.
  Qed.

  Inductive init: forall (vars:list Var.t) (e1 e2:t), Prop :=
  | init_nil e: init nil e e
  | init_cons_i
      var vars e1 e2
      (Henv: init vars e1 e2):
      init ((Typ.int, var)::vars) e1 (set (Typ.int, var) (Value.int Int.zero) e2)
  | init_cons_p
      var vars e1 e2
      (Henv: init vars e1 e2):
      init ((Typ.ptr, var)::vars) e1 (set (Typ.ptr, var) (Value.ptr 1%positive Int.zero) e2)
  .

  Definition wf (locals:list Var.t) (env:t): bool :=
    for_all
      (fun var =>
         match get var env with
           | None => false
           | Some val => Value.has_type val (fst var)
         end)
      locals.

  Lemma wf_inv_mon locals1 locals2 env
    (Hsubset : list_subset locals2 locals1)
    (Hwf : wf locals1 env):
    wf locals2 env.
  Proof.
    unfold wf in *; rewrite for_all_spec in *; eauto.
  Qed.

  Lemma init_wf
        e1 vars1 (He1: wf vars1 e1)
        e2 vars2 (He2: init vars2 e1 e2):
    wf (vars1++vars2) e2.
  Proof.
    cut (wf (vars2++vars1) e2).
    { intro X. apply for_all_spec. intros.
      eapply for_all_spec in X; eauto.
      apply in_app. apply in_app in Ha.
      destruct Ha; auto.
    }
    generalize dependent e2. induction vars2; intros; auto.
    { inv He2. }
    inv He2; tac.
    - unfold wf. tac. split.
      + rewrite gsspec.
        destruct (Var.eq (Typ.int, var) (Typ.int, var)); tac.
      + exploit IHvars2; [by eauto|]. intro X.
        apply for_all_spec. intros.
        eapply for_all_spec in X; eauto. tac.
        rewrite gsspec.
        destruct (Var.eq a (Typ.int, var)); tac.
    - unfold wf. tac. split.
      + rewrite gsspec.
        destruct (Var.eq (Typ.ptr, var) (Typ.ptr, var)); tac.
      + exploit IHvars2; [by eauto|]. intro X.
        apply for_all_spec. intros.
        eapply for_all_spec in X; eauto. tac.
        rewrite gsspec.
        destruct (Var.eq a (Typ.ptr, var)); tac.
  Qed.

  Lemma init_nonempty vars e1: exists e2, init vars e1 e2.
  Proof.
    induction vars; intros; tac.
    - eexists. econs.
    - destruct a. destruct t0.
      + eexists. econs; eauto.
      + eexists. econs; eauto.
  Qed.

  Definition keys (e:t): list Var.t :=
    (map (fun key => (Typ.int, key)) (PTree.keys e.(int))) ++
    (map (fun key => (Typ.ptr, key)) (PTree.keys e.(ptr))).

  Lemma empty_keys: Env.keys Env.empty = nil.
  Proof. auto. Qed.

  Lemma init_keys1
        locals env1 env2
        (Hinit: Env.init locals env1 env2)
        var val (Hget: Env.get var env2 = Some val):
    (In var locals /\ Value.has_type val (fst var)) \/
    Env.get var env1 = Some val.
  Proof.
    generalize dependent env2.
    induction locals; intros; inv Hinit; auto.
    - tac. rewrite Env.gsspec in Hget.
      destruct (Var.eq var (Typ.int, var0)); tac.
      + left. split; [by left|]. tac.
      + exploit IHlocals; eauto. clear IHlocals. intros [[? ?]|IHlocals].
        * left. split; [|done]. by right.
        * by right.
    - tac. rewrite Env.gsspec in Hget.
      destruct (Var.eq var (Typ.ptr, var0)); tac.
      + left. split; [by left|]. tac.
      + exploit IHlocals; eauto. clear IHlocals. intros [[? ?]|IHlocals].
        * left. split; [|done]. by right.
        * by right.
  Qed.

  Lemma init_keys2
        locals env1 env2
        (Hinit: Env.init locals env1 env2)
        var
        (H: In var locals \/ Env.get var env1):
    Env.get var env2.
  Proof.
    generalize dependent env2.
    generalize dependent H.
    induction locals; intros; inv Hinit; auto.
    { by destruct H. }
    { rewrite Env.gsspec.
      destruct (Var.eq var (Typ.int, var0)); tac.
      destruct H.
      - destruct H; tac.
        apply IHlocals; auto.
      - apply IHlocals; auto.
    }
    { rewrite Env.gsspec.
      destruct (Var.eq var (Typ.ptr, var0)); tac.
      destruct H.
      - destruct H; tac.
        apply IHlocals; auto.
      - apply IHlocals; auto.
    }
  Qed.

  Lemma gkspec var env:
    In var (keys env) <-> get var env <> None.
  Proof.
    destruct var as [typ id].
    unfold keys. rewrite in_app.
    unfold get. tac. destruct typ; constructor; intros.
    - destruct H; apply in_map_iff in H; destruct H as [id' [X Hid']]; inv X.
      by apply PTree.gkspec.
    - left. apply in_map_iff. exists id. split; auto.
      by apply PTree.gkspec.
    - destruct H; apply in_map_iff in H; destruct H as [id' [X Hid']]; inv X.
      by apply PTree.gkspec.
    - right. apply in_map_iff. exists id. split; auto.
      by apply PTree.gkspec.
  Qed.

  Lemma gempty var:
    get var empty = None.
  Proof.
    unfold get, empty.
    destruct var. tac.
    destruct t0; tac; apply PTree.gempty.
  Qed.

  Lemma in_env_keys_iff: forall x env, In x (keys env) <-> 
    exists v, get x env = Some v.
  Proof.
    intros; rewrite gkspec; tac.
    destruct (get x env); split; intro; eauto; tac.
  Qed.

  Lemma in_env_keys: forall x env, In x (keys env) -> 
    exists v, get x env = Some v.
  Proof. intros; rewrite <- in_env_keys_iff; tac. Qed.

  Lemma set_keys: forall x x' v env, In x (keys (set x' v env)) ->
    x = x' \/ In x (keys env).
  Proof.
    setoid_rewrite in_env_keys_iff; setoid_rewrite gsspec; intros; tac.
    destruct (Var.eq x x'); [left | right]; tac; eauto.
  Qed.

  Lemma init_keys_wf
        e1 (He1: wf (keys e1) e1)
        e2 vars (He2: init vars e1 e2):
    wf (keys e2) e2.
  Proof.
    revert e2 He2; induction vars; intros.
    - inversion He2; tac.
    - inversion He2; tac.
      { specialize (IHvars _ Henv).
        unfold wf; apply for_all_spec; intros.
        exploit in_env_keys; eauto; intro; tac.
        rewrite H; rewrite gsspec in H.
        destruct (Var.eq a (Typ.int, var)); tac.
        unfold wf in IHvars; rewrite for_all_spec in IHvars.
        exploit set_keys; eauto; intro X.
        destruct X; tac.
        specialize (IHvars _ H0); rewrite H in IHvars; auto.
      }
      { specialize (IHvars _ Henv).
        unfold wf; apply for_all_spec; intros.
        exploit in_env_keys; eauto; intro; tac.
        rewrite H; rewrite gsspec in H.
        destruct (Var.eq a (Typ.ptr, var)); tac.
        unfold wf in IHvars; rewrite for_all_spec in IHvars.
        exploit set_keys; eauto; intro X.
        destruct X; tac.
        specialize (IHvars _ H0); rewrite H in IHvars; auto.
      }
  Qed.

  Lemma init_keys_subset
        locals env1 env2
        (Hinit: Env.init locals env1 env2):
    list_subset (keys env1 ++ locals) (keys env2).
  Proof.
    repeat intro.
    rewrite in_app_iff in H.
    exploit init_keys2; eauto.
    - destruct H; [right | left]; eauto.
      unfold option_to_bool; exploit in_env_keys; eauto; intro; tac.
      rewrite H0; auto.
    - rewrite gkspec; intro.
      destruct (get a env2); tac.
  Qed.
End Env.

(** StaticState consists of:
 - genv: global environment
 - fenv: function environment
 *)
Structure StaticState: Type := mkStaticState {
  genv: GEnv.t;
  fenv: PTree.t FDecl.t
}.
Definition StaticState_wf (sst:StaticState): bool :=
  PTree_Properties.for_all sst.(fenv) (fun _ => FDecl.wf (PTree.keys sst.(genv))).

(** Context consists of:
 - lenv: local environment
 - instrs: remaining instructions to run
 *)
Structure Context: Type := mkContext {
  lenv: Env.t;
  instrs: list Instr.t
}.
Inductive Context_wf (sst:StaticState) (c:Context): Prop :=
| Context_wf_before_initialize
    (Henv: Env.wf (Env.keys c.(lenv)) c.(lenv))
    locals istrs (Hinstrs: c.(instrs) = (Instr.initialize locals)::istrs)
    (Hinstrs: Forall (Instr.wf (PTree.keys sst.(genv)) (Env.keys c.(lenv) ++ locals)) istrs)
| Context_wf_after_initialize
    (Henv: Env.wf (Env.keys c.(lenv)) c.(lenv))
    (Hinstrs: Forall (Instr.wf (PTree.keys sst.(genv)) (Env.keys c.(lenv))) c.(instrs))
.

(** State consists of:
 - context: the top stack frame
 - stack: remaining stack frames
 - mem: memory
 *)
Structure State: Type := mkState {
  context: Context;
  stack: list Context;
  mem: Mem.t
}.

Ltac tac :=
  repeat
    (simpl in *; unfold SomeP, NoneP, singletonP in *; subst; pclearbot;
     (try done);
     (try basictac_aux);
     (try destac_aux);
     (try monadtac_aux);
     (try maptac_aux);
     (try memtac_aux);
     (try match goal with
            | [st:State |- _] =>
              let lenv := fresh "lenv" in
              let istrs := fresh "istrs" in
              let stack := fresh "stack" in
              let mem := fresh "mem" in
              destruct st as [[lenv istrs] stack mem]
          end)).

Definition unknown_int :=
  liftP (fun v => Value.has_type v Typ.int).

Definition unknown_bool :=
  liftP (fun v => (Value.eq v Int.zero) || (Value.eq v Int.one)).

Lemma unknown_bool_type
      val (H: unknown_bool (Some val)):
  Value.has_type val Typ.int.
Proof.
  unfold unknown_bool in H. simpl in *.
  apply orb_true_iff in H. destruct H.
  - by destruct (Value.eq val Int.zero); [subst|done].
  - by destruct (Value.eq val Int.one); [subst|done].
Qed.

Definition bool2int (b: bool) : int :=
  if b then Int.one else Int.zero.
Coercion bool2int: bool >-> int.

Definition sem_bop (m:Mem.t) (bop:Bop.t) (t1:Typ.t) (v1:Value.t) (t2:Typ.t) (v2:Value.t): option Value.t :=
  match bop, t1, t2 with
  | Bop.plus, Typ.int, Typ.int =>
    match v1, v2 with
    | Value.int i1, Value.int i2 =>
      Some (Value.int (Int.add i1 i2))
    | _, _ =>
      None
    end
  | Bop.minus, Typ.int, Typ.int =>
    match v1, v2 with
    | Value.int i1, Value.int i2 =>
      Some (Value.int (Int.sub i1 i2))
    | _, _ =>
      None
    end
  | Bop.times, Typ.int, Typ.int =>
    match v1, v2 with
    | Value.int i1, Value.int i2 =>
      Some (Value.int (Int.mul i1 i2))
    | _, _ =>
      None
    end
  | Bop.band, Typ.int, Typ.int =>
    match v1, v2 with
    | Value.int i1, Value.int i2 =>
      Some (Value.int (Int.and i1 i2))
    | _, _ =>
      None
    end
  | Bop.equal, Typ.int, Typ.int =>
    match v1, v2 with
    | Value.int i1, Value.int i2 =>
      Some (Value.int (Int.eq i1 i2))
    | _, _ =>
      None
    end

  (* Special semantics for pointers *)
  | Bop.plus, Typ.int, Typ.ptr =>
    match v1, v2 with
    | Value.int i1, Value.ptr b2 o2 =>
      Some (Value.ptr b2 (Int.add i1 o2))
    | _, _ =>
      None
    end
  | Bop.plus, Typ.ptr, Typ.int =>
    match v1, v2 with
    | Value.ptr b1 o1, Value.int i2 =>
      Some (Value.ptr b1 (Int.add o1 i2))
    | _, _ =>
      None
    end
  | Bop.minus, Typ.ptr, Typ.int =>
    match v1, v2 with
    | Value.ptr b1 o1, Value.int i2 =>
      Some (Value.ptr b1 (Int.sub o1 i2))
    | _, _ =>
      None
    end
  | Bop.minus, Typ.ptr, Typ.ptr =>
    match v1, v2 with
    | Value.ptr b1 o1, Value.ptr b2 o2 =>
      if ident_eq b1 b2
      then Some (Value.int (Int.sub o1 o2))
      else None
    | _, _ => None
    end

  | Bop.equal, Typ.ptr, Typ.ptr =>
    match v1, v2 with
      | Value.ptr b1 o1, Value.ptr b2 o2 =>
        if ident_eq b1 b2
        then Some (Value.int (Int.eq o1 o2))
        else
          match Mem.val2ptr v1 m, Mem.val2ptr v2 m with
            | Some (b1, _), Some (b2, _) =>
              if ident_eq b1 b2
              then Some (Value.int (Int.eq o1 o2))
              else Some (Value.int false)
            | _, _ =>
              None
          end
      | _, _ =>
        None
    end
  | _, _, _ => None
  end.

Lemma sem_bop_type (m:Mem.t) (bop:Bop.t)
      t1 t2 typ (Htyp: Bop.has_type bop t1 t2 = Some typ)
      v1 (H1: Value.has_type v1 t1)
      v2 (H2: Value.has_type v2 t2)
      val (Hval: sem_bop m bop t1 v1 t2 v2 = Some val):
  Value.has_type val typ.
Proof.
  Ltac ttac_aux :=
    match goal with
      | [H: Value.int Int.one = Value.int Int.zero |- _] => generalize H; intro X; inv X
      | [H: Value.int Int.zero = Value.int Int.one |- _] => generalize H; intro X; inv X
      | [H: context [Mem.normalize ?a ?b] |- _] => destruct (Mem.normalize a b) in *
      | [H: context [Mem.val2ptr ?a ?b] |- _] => destruct (Mem.val2ptr a b) in *
      | [H: context [Mem.is_valid_ptr ?a ?b ?c] |- _] => destruct (Mem.is_valid_ptr a b c) in *
      | [H: context [Mem.is_nonzero_ptr ?a ?b ?c] |- _] => destruct (Mem.is_nonzero_ptr a b c) in *
      | [H: context [Int.eq ?a ?b] |- _] => destruct (Int.eq a b) in *
      | [H: context [ident_eq ?a ?b] |- _] => destruct (ident_eq a b) in *
      | [H: context [Value.eq ?a ?b] |- _] => destruct (Value.eq a b) in *
      | [H: context [match ?g with | Some _ => _ | None => _ end] |- _] => destruct g in *
      | [H: context [if ?g then _ else _] |- _] => destruct g in *
      | [H: _ || _ = true |- _] => apply orb_true_iff in H; destruct H
    end.
  destruct bop, t1, t2; simpl in *; (try inv H); auto;
  destruct v1, v2;  try (by destruct val; tac); inv Htyp;
  unfold sem_bop in *;
  repeat (try ttac_aux; try destruct p; try destruct p0; subst); inv Hval;
  repeat (try ttac_aux; subst); try constructor;
  simpl in *; congruence.
Qed.

Definition sem_bexp (genv:GEnv.t) (lenv:Env.t) (bexp:Exp.bt)
  : option Value.t :=
  match bexp with
    | Exp.const const =>
      Some (Value.int const)
    | Exp.global id =>
      doT v1 <- PTree.get id genv;
      Some (Value.ptr v1 0)
    | Exp.local var =>
      doT v1 <- Env.get var lenv;
      assertB (Value.has_type v1 (fst var));
      Some v1
  end.

Fixpoint sem_exp (genv:GEnv.t) (lenv:Env.t) (m:Mem.t) (exp:Exp.t): option Value.t -> Prop :=
  match exp with
    | Exp.base bexp => 
      singletonP (sem_bexp genv lenv bexp)
    | Exp.bop bop t1 e1 t2 e2 =>
      doPP v1 <- sem_exp genv lenv m e1;
      assertP (Value.has_type v1 t1);
      doPP v2 <- sem_exp genv lenv m e2;
      assertP (Value.has_type v2 t2);
      singletonP (sem_bop m bop t1 v1 t2 v2)
  end.

Lemma sem_exp_type (genv:GEnv.t) (lenv:Env.t) (m:Mem.t) (exp:Exp.t)
      typ (Htyp: Exp.has_type (PTree.keys genv) (Env.keys lenv) exp = Some typ)
      val (Hval: sem_exp genv lenv m exp (Some val)):
  Value.has_type val typ.
Proof.
  generalize dependent typ.
  generalize dependent val.
  induction exp; (try destruct bexp);
  unfold sem_exp; intros val Hval t Ht;
  repeat (tac; fold sem_exp in *).
  destruct Hr2; tac. destruct Hr2; tac.
  exploit sem_bop_type; eauto.
Qed.

Lemma sem_exp_nonempty genv lenv mem e:
  exists result, sem_exp genv lenv mem e result.
Proof.
  induction e; tac.
  - destruct bexp; eauto.
  - destruct (Value.has_type t t1); [|by eexists; eauto].
    destruct x.
    + destruct (Value.has_type t0 t2) eqn:X.
      * eexists. right. split; eauto. eapply bindPP_right; eauto.
      * exists None. right. split; auto. eapply bindPP_right; eauto.
        left. split; auto. by rewrite X.
    + exists None. right. split; auto. constructor; auto.
Qed.

Fixpoint sem_bexp_list (genv:GEnv.t) (lenv:Env.t) (es:list Exp.bt): option (list Value.t) :=
  match es with
    | nil => Some nil
    | e::es =>
      doT r <- sem_bexp genv lenv e;
      doT rs <- sem_bexp_list genv lenv es;
      Some (r::rs)
  end.

Definition sem_cast (m:Mem.t) (typ:Typ.t) (v1:Value.t): option (Value.t * Mem.t) -> Prop :=
  match typ, v1 with
    | Typ.int, Value.int i => SomeP (v1, m)
    | Typ.int, Value.ptr b o =>
      match Mem.normalize b o m with
        | Some i => SomeP (Value.int i, m)
        | None =>
          doPP val <- liftP (fun _:int => True);
          doPP mem <- liftP (Mem.realize m b (Int.sub val o));
          SomeP (Value.int val, mem)
      end
    | Typ.ptr, Value.int i =>
      match Mem.denormalize i m with
        | Some (b, o) => SomeP (Value.ptr b o, m)
        | None => NoneP
      end
    | Typ.ptr, Value.ptr _ _ => SomeP (v1, m)
  end.

Lemma sem_cast_type (m:Mem.t) (typ:Typ.t) (v1:Value.t)
      val mem' (H: sem_cast m typ v1 (Some (val, mem'))):
  Value.has_type val typ.
Proof.
  destruct typ, v1; simpl in H; try inv H.
  - destruct (Mem.normalize b o m) eqn:Hv1n; tac; try inv H.
  - destruct (Mem.denormalize intval m) eqn:Hv1n; tac; try inv H.
Qed.

Definition sem_rexp (genv:GEnv.t) (lenv:Env.t) (m1:Mem.t) (rexp:RExp.t):
  option (Value.t * option Event.t * Mem.t) -> Prop :=
  match rexp with
    | RExp.exp e1 =>
      doPP v1 <- sem_exp genv lenv m1 e1;
      SomeP (v1, None, m1)
    | RExp.free e1 =>
      doPP v1 <- sem_exp genv lenv m1 e1;
      doP (v2, m2) <- Mem.free v1 m1;
      SomeP (Value.int v2, None, m2)
    | RExp.malloc esize align =>
      doPP vsize <- sem_exp genv lenv m1 esize;
      doP isize <- Mem.val2int vsize m1;
      let '(result, m2) := Mem.alloc isize m1 in
      SomeP (Value.ptr result Int.zero, None, m2)
    | RExp.cast typ e1 =>
      doPP v1 <- sem_exp genv lenv m1 e1;
      doPP (v2, m2) <- sem_cast m1 typ v1;
      SomeP (v2, None, m2)
    | RExp.input =>
      doPP val <- liftP (fun _:int => True);
      SomeP (Value.int val, Some (Event.input val), m1)
    | RExp.output e1 =>
      doPP v1 <- sem_exp genv lenv m1 e1;
      doP i1 <- Mem.val2int v1 m1;
      SomeP (Value.int i1, Some (Event.output i1), m1)
  end.

Lemma sem_rexp_type (genv:GEnv.t) (lenv:Env.t) (m1:Mem.t) (rexp:RExp.t)
      typ (Htyp: RExp.has_type (PTree.keys genv) (Env.keys lenv) rexp = Some typ)
      val evt m2 (Hval: sem_rexp genv lenv m1 rexp (Some (val, evt, m2))):
  Value.has_type val typ.
Proof.
  destruct rexp; tac.
  - exploit sem_exp_type; eauto.
  - exploit sem_cast_type; eauto.
Qed.

Fixpoint bind_params (params:list (Typ.t * ident)) (args:list Value.t) (mem:Mem.t) (lenv1:Env.t): option Env.t :=
  match params, args with
    | nil, nil => Some lenv1
    | param::params, arg::args =>
      assertB (Value.has_type arg (fst param));
      bind_params params args mem (Env.set param arg lenv1)
    | _, _ => None
  end.

Definition is_terminated (st:State): bool :=
  is_empty st.(context).(instrs) &&
  is_empty st.(stack).

Definition sem_instr (sst:StaticState) (st1:State):
  option (option Event.t * State) -> Prop :=
  match st1.(context).(instrs) with
    | nil =>
      match st1.(stack) with
        | nil => emptyP
        | newctx::newstack =>
          SomeP (None, mkState newctx newstack st1.(mem))
      end
    | istr::istrs =>
      match istr with
        | Instr.initialize locals =>
          doPP nlenv <- liftP (Env.init locals st1.(context).(lenv));
          SomeP
            (None,
             mkState (mkContext nlenv istrs) st1.(stack) st1.(mem))
        | Instr.call fid args =>
          doP func <- PTree.get fid sst.(fenv);
          doPP tvargs <- singletonP (sem_bexp_list sst.(genv) st1.(context).(lenv) args);
          doP nlenv <- bind_params func.(FDecl.params) tvargs st1.(mem) Env.empty;
          SomeP
            (None,
             mkState
               (mkContext nlenv ((Instr.initialize func.(FDecl.locals)):: func.(FDecl.body)))
               ((mkContext st1.(context).(lenv) istrs)::st1.(stack))
               st1.(mem))
        | Instr.assign (Some var) rhs =>
          doPP (vrhs, event, m1) <- sem_rexp sst.(genv) st1.(context).(lenv) st1.(mem) rhs;
          assertP (Value.has_type vrhs (fst var));
          SomeP
            (event,
             mkState
               (mkContext
                  (Env.set var vrhs st1.(context).(lenv))
                  istrs)
               st1.(stack)
               m1)
        | Instr.assign None rhs =>
          doPP (vrhs, event, m1) <- sem_rexp sst.(genv) st1.(context).(lenv) st1.(mem) rhs;
          SomeP
            (event,
             mkState (mkContext st1.(context).(lenv) istrs) st1.(stack) m1)
        | Instr.load (Some var) rhs =>
          doPP vrhs <- sem_exp sst.(genv) st1.(context).(lenv) st1.(mem) rhs;
          doP v <- Mem.load vrhs st1.(mem);
          assertP (Value.has_type v (fst var));
          SomeP
            (None,
             mkState
               (mkContext
                  (Env.set var v st1.(context).(lenv))
                  istrs)
               st1.(stack)
               st1.(mem))
        | Instr.load None rhs =>
          doPP vrhs <- sem_exp sst.(genv) st1.(context).(lenv) st1.(mem) rhs;
          doP v <- Mem.load vrhs st1.(mem);
          SomeP
            (None,
             mkState (mkContext st1.(context).(lenv) istrs) st1.(stack) st1.(mem))
        | Instr.store lhs rhs =>
          doPP vlhs <- sem_exp sst.(genv) st1.(context).(lenv) st1.(mem) lhs;
          doPP vrhs <- sem_exp sst.(genv) st1.(context).(lenv) st1.(mem) rhs;
          doP m2 <- Mem.store vlhs vrhs st1.(mem);
          SomeP
            (None,
             mkState
               (mkContext st1.(context).(lenv) istrs)
               st1.(stack)
               m2)
        | Instr.while econd body =>
          doPP vcond <- sem_exp sst.(genv) st1.(context).(lenv) st1.(mem) econd;
          doP bcond <- Mem.val2bool vcond st1.(mem);
          SomeP
            (None,
             mkState
               (mkContext
                  st1.(context).(lenv)
                  (if bcond
                   then body++st1.(context).(instrs)
                   else istrs))
               st1.(stack)
               st1.(mem))
        | Instr.ifte econd s1 s2 =>
          doPP vcond <- sem_exp sst.(genv) st1.(context).(lenv) st1.(mem) econd;
          doP bcond <- Mem.val2bool vcond st1.(mem);
          SomeP
            (None,
             mkState
               (mkContext
                  st1.(context).(lenv)
                  ((if bcond then s1 else s2)++istrs))
               st1.(stack)
               st1.(mem))
      end
  end.

Lemma sem_instr_not_call_prop
      istr istrs st1 st2 evt sst
      (Histrs: st1.(context).(instrs) = istr::istrs)
      (Hncall: Instr.is_call istr = false)
      (Hsem: sem_instr sst st1 (Some (evt, st2))):
  st1.(stack) = st2.(stack).
Proof.
  tac.
  unfold sem_instr in Hsem. destruct istr; tac.
  - destruct var; tac. destruct Hr2; tac.
  - destruct var; tac. destruct Hr2; tac.
Qed.

Lemma Env_set_Env_wf
      lenv var val (Hval: Value.has_type val (fst var))
      (Hlenv: Env.wf (Env.keys lenv) lenv):
  Env.wf (Env.keys (Env.set var val lenv)) (Env.set var val lenv).
Proof.
  apply for_all_spec. intros v Hv.
  apply Env.gkspec in Hv. rewrite Env.gsspec in *.
  destruct (Var.eq v var); tac.
  eapply for_all_spec in Hlenv; eauto.
  apply Env.gkspec. auto.
Qed.

Lemma Env_set_Instr_wf
      globals lenv var val i (Hval: Value.has_type val (fst var))
      (Hi: Instr.wf globals (Env.keys lenv) i):
  Instr.wf globals (Env.keys (Env.set var val lenv)) i.
Proof.
  eapply Instr.wf_mon; eauto.
  - reflexivity.
  - intros v Hv. apply Env.gkspec. rewrite Env.gsspec.
    destruct (Var.eq v var); tac.
    apply Env.gkspec. auto.
Qed.

Lemma Forall_app {T:Type} (P: T -> Prop) (l1 l2:list T):
  Forall P (l1++l2) <-> Forall P l1 /\ Forall P l2.
Proof.
  revert l2. induction l1; intros; tac; constructor; intros; tac.
  - inv H. apply IHl1 in H3. destruct H3. split; auto.
  - constructor.
    + inv H.
    + apply IHl1. split; tac. inv H.
Qed.

Lemma sem_instr_wf
      sst st1 (Hst1: Context_wf sst st1.(context))
      evt st2 (H: sem_instr sst st1 (Some (evt, st2)))
      (Hstack: st1.(stack) = st2.(stack)):
  Context_wf sst st2.(context).
Proof.
  unfold sem_instr in H. tac.
  destruct istrs0; tac.
  { destruct stack0; tac. clear -H3.
    apply (eq_apply (@length _)) in H3. tac. omega.
  }
  destruct t;
    repeat
      (tac;
       try match goal with
             | [H: Context_wf _ _ = true |- _] =>
               apply andb_true_iff in H; destruct H
             | [|- Context_wf _ _ = true] =>
               apply andb_true_iff; split
             | [|- _ && _ = true] =>
               apply andb_true_iff; split
           end).
  - inv Hst1; inv Hinstrs. tac.
    constructor; tac.
    + eapply Env.init_keys_wf; eauto.
    + eapply Forall_impl; eauto. tac. intros i Hi.
      eapply Instr.wf_mon; eauto.
      * reflexivity.
      * by apply Env.init_keys_subset.
  - clear -H3. exfalso. induction stack0; inv H3.
  - inv Hst1. tac. destruct var; tac.
    + destruct Hr2; tac. constructor; tac.
      * by apply Env_set_Env_wf.
      * inv Hinstrs. eapply Forall_impl; eauto. tac. intros istr Histr.
        by apply Env_set_Instr_wf.
    + constructor; tac. inv Hinstrs.
  - inv Hst1. tac. destruct var; tac.
    + destruct Hr2; tac. constructor; tac.
      * by apply Env_set_Env_wf.
      * inv Hinstrs. eapply Forall_impl; eauto. tac. intros istr Histr.
        by apply Env_set_Instr_wf.
    + constructor; tac. inv Hinstrs.
  - inv Hst1. tac. constructor; tac. inv Hinstrs.
  - inv Hst1. tac. constructor; tac. inv Hinstrs.
    destruct b; tac. apply Forall_app. split; tac.
    inv Hinstrs. apply andb_true_iff in H3. destruct H3.
    revert H0. clear. induction body; intros; tac.
    constructor; tac. by apply IHbody.
  - inv Hst1. tac. constructor; tac. inv Hinstrs.
    inv Hinstrs. apply andb_true_iff in H1. tac.
    destruct b; tac.
    + apply Forall_app. split; tac.
      revert H0. clear. induction s1; intros; tac.
      constructor; tac. by apply IHs1.
    + apply Forall_app. split; tac.
      revert H1. clear. induction s2; intros; tac.
      constructor; tac. by apply IHs2.
Qed.

Definition Exp_has_type_instrs
           (globals: list ident) (locals:list Var.t) (istrs:list Instr.t) (exp:Exp.t): option Typ.t :=
  Exp.has_type
    globals
    (match istrs with
       | (Instr.initialize vars)::_ => locals ++ vars
       | _ => locals
     end)
    exp.

Lemma Env_set_keys env1 var val:
  list_subset (Env.keys env1) (Env.keys (Env.set var val env1)).
Proof.
  repeat intro. rewrite Env.gkspec in *. intro X. apply H.
  rewrite Env.gsspec in X. destruct (Var.eq a var); tac.
Qed.

Lemma sem_instr_exp_has_type
      sst st1 (Hst1: Context_wf sst st1.(context))
      exp typ (Htyp: Exp_has_type_instrs (PTree.keys sst.(genv)) (Env.keys st1.(context).(lenv)) st1.(context).(instrs) exp = Some typ)
      evt st2 (H: sem_instr sst st1 (Some (evt, st2)))
      (Hstack: st1.(stack) = st2.(stack)):
  Exp_has_type_instrs (PTree.keys sst.(genv)) (Env.keys st2.(context).(lenv)) st2.(context).(instrs) exp = Some typ.
Proof.
  unfold sem_instr in H. tac.
  destruct istrs0; tac.
  { destruct stack0; tac. clear -H3.
    apply (eq_apply (@length _)) in H3. tac. omega.
  }
  destruct t;
    repeat
      (tac;
       try match goal with
             | [H: Context_wf _ _ = true |- _] =>
               apply andb_true_iff in H; destruct H
             | [|- Context_wf _ _ = true] =>
               apply andb_true_iff; split
             | [|- _ && _ = true] =>
               apply andb_true_iff; split
           end).
  - inv Hst1; inv Hinstrs; tac.
    unfold Exp_has_type_instrs in *; tac.
    eapply Env.init_keys_subset in Hr1; eauto.
    eapply Exp.has_type_mon in Htyp; [|reflexivity|by eauto].
    destruct istrs0; tac. destruct t; tac.
    inv Hinstrs0.
  - apply (eq_apply (@length _)) in H3. tac. omega.
  - inv Hst1; inv Hinstrs; tac.
    unfold Exp_has_type_instrs in *; tac.
    destruct var; tac.
    + destruct Hr2; tac.
      destruct istrs; tac;
      try by (eapply Exp.has_type_mon; eauto; try reflexivity; try apply Env_set_keys).
      destruct t0; tac;
      try by (eapply Exp.has_type_mon; eauto; try reflexivity; try apply Env_set_keys).
      inv H3.
    + destruct istrs; tac. destruct t1; tac.
      inv H3.
  - inv Hst1; inv Hinstrs; tac.
    unfold Exp_has_type_instrs in *; tac.
    destruct var; tac.
    + destruct Hr2; tac.
      destruct istrs; tac;
      try by (eapply Exp.has_type_mon; eauto; try reflexivity; try apply Env_set_keys).
      destruct t1; tac;
      try by (eapply Exp.has_type_mon; eauto; try reflexivity; try apply Env_set_keys).
      inv H3.
    + destruct istrs; tac. destruct t0; tac.
      inv H3.
  - inv Hst1; inv Hinstrs; tac.
    unfold Exp_has_type_instrs in *; tac.
    destruct istrs; tac. destruct t; tac.
    inv H2.
  - inv Hst1; inv Hinstrs; tac.
    unfold Exp_has_type_instrs in *; tac.
    inv Hinstrs. inv H3. tac.
    destruct b.
    + destruct body; tac.
      destruct t; tac.
    + destruct istrs0; tac. destruct t; tac.
      inv H4.
  - inv Hst1; inv Hinstrs; tac.
    unfold Exp_has_type_instrs in *; tac.
    inv Hinstrs. inv H4. tac.
    destruct b.
    + destruct s1; tac.
      * destruct istrs0; tac. destruct t; tac.
        inv H5.
      * destruct t; tac.
    + destruct s2; tac.
      * destruct istrs0; tac. destruct t; tac.
        inv H5.
      * destruct t; tac.
Qed.

Lemma bind_params_keys1
      params args mem env1 env2
      (Hbind: bind_params params args mem env1 = Some env2)
      var val (Hval: Env.get var env2 = Some val):
  (In var params /\ Value.has_type val (fst var)) \/
  Env.get var env1 = Some val.
Proof.
  generalize dependent env2.
  generalize dependent env1.
  generalize dependent args.
  induction params; destruct args as [|arg args]; intros; tac.
  { by right. }
  exploit IHparams; eauto. intros [[? ?]|X]; tac.
  - left. split; tac. by right.
  - rewrite Env.gsspec in X.
    destruct (Var.eq var (t, i)); tac.
    + left. split; tac. by left.
    + by right.
Qed.

Lemma bind_params_keys2
      params args mem env1 env2
      (Hbind: bind_params params args mem env1 = Some env2)
      var
      (H: In var params \/ Env.get var env1):
  Env.get var env2.
Proof.
  generalize dependent env2.
  generalize dependent env1.
  generalize dependent args.
  induction params; destruct args as [|arg args]; intros; tac.
  { by destruct H. }
  destruct H.
  - exploit IHparams; eauto.
    destruct H; tac.
    + right. rewrite Env.gsspec.
      destruct (Var.eq (t, i) (t, i)); tac.
    + by left.
  - exploit (IHparams args (Env.set (t0, i0) arg env1)); eauto.
    right. rewrite Env.gsspec.
    destruct (Var.eq (t, i) (t0, i0)); tac.
Qed.

Corollary bind_params_keys_subset
  params args mem env1 env2
  (Hbind: bind_params params args mem env1 = Some env2):
  list_subset (params ++ Env.keys env1) (Env.keys env2).
Proof.
  repeat intro.
  rewrite in_app_iff in H.
  exploit bind_params_keys2; eauto.
  - destruct H; [left | right]; eauto.
    unfold option_to_bool; exploit Env.in_env_keys; eauto; intro; tac.
    rewrite H0; auto.
  - rewrite Env.gkspec; intro.
    destruct (Env.get a env2); tac.
Qed.

Lemma bind_params_wf: forall vars mem vals env 
  (Hbind : bind_params vars vals mem Env.empty = Some env),
  Env.wf (Env.keys env) env.
Proof.
  intros; unfold Env.wf; rewrite for_all_spec; intros.
  rewrite Env.in_env_keys_iff in Ha; tac.
  exploit bind_params_keys1; eauto.
  rewrite H; rewrite Env.gempty; intro param; destruct param; tac.
Qed.

Lemma sem_instr_wf_call
      sst (Hsst: StaticState_wf sst)
      st1 evt st2 (H: sem_instr sst st1 (Some (evt, st2)))
      fid args istrs (Hinstrs: st1.(context).(instrs) = (Instr.call fid args)::istrs):
  Context_wf sst st2.(context).
Proof.
  unfold sem_instr in H. tac. eapply Context_wf_before_initialize; tac.
  - eapply bind_params_wf; eauto.
  - apply Forall_forall. intros i Hi.
    eapply PTree_Properties.for_all_correct in Hsst; [|by eauto].
    eapply for_all_spec in Hsst; eauto.
    eapply Instr.wf_mon; eauto; tac.
    intros v Hv. rewrite in_app in *. destruct Hv as [Hv|Hv].
    + left. exploit bind_params_keys2; eauto.
      rewrite Env.gkspec. by destruct (Env.get v lenv0).
    + by right.
Qed.

Inductive init_globals: forall (globals:list (ident * int)) (genv1:GEnv.t) (m1:Mem.t) (result:GEnv.t * Mem.t), Prop :=
| init_globals_nil
    genv m:
    init_globals nil genv m (genv, m)
| init_globals_cons
    gvar isize globals genv1 genv3 gb m1 m2 m3
    (H12: Mem.alloc isize m1 = (gb, m2))
    (H23: init_globals globals (PTree.set gvar gb genv1) m2 (genv3, m3)):
    init_globals ((gvar,isize)::globals) genv1 m1 (genv3, m3)
.

Lemma init_globals_keys
      globals genv1 mem1 genv2 mem2
      (H: init_globals globals genv1 mem1 (genv2, mem2))
      gvar:
  In gvar (PTree.keys genv2) <-> In gvar (map fst globals) \/ In gvar (PTree.keys genv1).
Proof.
  generalize dependent mem2.
  generalize dependent mem1.
  generalize dependent genv2.
  generalize dependent genv1.
  induction globals; intros; inv H; tac.
  { constructor; intro.
    - by right.
    - by destruct H0.
  }
  exploit IHglobals; eauto. clear IHglobals. intros IHglobals.
  constructor; intro X.
  - apply IHglobals in X. destruct X as [Hin|Hin].
    + left. right. auto.
    + unfold PTree.keys in Hin.
      idtac.
      apply PTree.gkspec in Hin. tac.
      destruct (peq gvar gvar0); tac.
      { left. left. auto. }
      right. by apply PTree.gkspec.
  - apply IHglobals. destruct X as [[Hin|Hin]|Hin].
    + right. apply PTree.gkspec. tac.
    + by left.
    + right. apply PTree.gkspec. tac.
      destruct (peq gvar gvar0); tac.
      by apply PTree.gkspec.
Qed.

Import SyntaxNotations.

Definition load_prog (prog:Prog.t) (sst:StaticState): forall (ost:option State), Prop :=
  doPP (genv, m1) <- liftP (init_globals prog.(Prog.globals) (PTree.empty _) Mem.init);
  let fenv :=
      fold_left
        (fun acc fdecl =>
           PTree.set fdecl.(FDecl.id) fdecl acc)
        prog.(Prog.fdecls)
               (PTree.empty _)
  in
  assumeP (sst = mkStaticState genv fenv);
  SomeP (mkState (mkContext Env.empty [CALL (%f"main") WITH []]) nil m1).

Lemma FDecl_find funcs fid:
  find (FDecl.byid fid) (rev funcs) =
  (fold_left
     (fun (acc : PTree.t FDecl.t) (fdecl : FDecl.t) =>
        PTree.set (FDecl.id fdecl) fdecl acc) funcs
     (PTree.empty FDecl.t)) ! fid.
Proof.
  rewrite <- fold_left_rev_right.
  induction (rev funcs); clear funcs; intros; tac.
  { by rewrite PTree.gempty. }
  unfold FDecl.byid. destruct a. tac.
  destruct (peq fid id); tac.
Qed.

Lemma load_prog_wf prog (Hprog: Prog.wf prog)
      sst st (Hload: load_prog prog sst (Some st)):
  StaticState_wf sst /\ Context_wf sst st.(context).
Proof.
  unfold Prog.wf in Hprog.
  inv Hload. clear Hload.
  destruct v1 as [genv1 m1]. simpl in *. destruct Hr2. subst.
  inv H0. clear H0. simpl in *.
  split.
  - apply PTree_Properties.for_all_correct. simpl.
    rewrite <- fold_left_rev_right.
    rewrite <- for_all_rev in Hprog.
    revert Hprog genv1 m1 Hr1.
    generalize dependent (rev (Prog.fdecls prog)).
    induction l; simpl; intros.
    + by rewrite PTree.gempty in H.
    + tac. destruct (peq x (FDecl.id a)); tac.
      * eapply (FDecl.wf_mon (map fst (Prog.globals prog)) _); eauto.
        repeat intro. exploit init_globals_keys; eauto.
        intros [_ X]. apply X. by left.
      * eapply IHl; eauto.
  - intros. constructor; tac.
    apply Forall_forall. intros i Hi. inv Hi.
Qed.

Definition sem_get_status (mem_good: Mem.t -> bool) ost :=
  match ost with
    | None => Status.error
    | Some st =>
      if is_terminated st
      then Status.term
      else (if mem_good st.(mem) then Status.normal else Status.error)
  end.

Definition sem_step 
    (sem: StaticState -> State -> option (option Event.t * State) -> Prop)
    sst ost1 evt ost2 :=
  match ost1 with
  | Some st1 =>
    match ost2 with
      | Some st2 => sem sst st1 (Some (evt, st2))
      | None => sem sst st1 None /\ evt = None
    end
  | _ => False
  end.

Definition ltst_lang := mkLTS_type Prog.t StaticState (option State).

Program Definition lts : LTS :=
  mkLTS ltst_lang
    load_prog
    (sem_get_status (fun _ => true))
    (sem_step sem_instr)
    _.
Next Obligation.
  destruct cfg1 as [st1|]; tac.
  destruct istrs as [|istr istrs]; tac.
  destruct stack0; tac.
  destruct cfg2; tac.
Qed.
