Require Import RelationClasses List.
Require Import Relations Equivalence EquivDec.
Require Import Coqlib.

Set Implicit Arguments.

Notation rtc := (clos_refl_trans _). (* reflexive transitive closure *)
Notation tc := (clos_trans _). (* transitive closure *)
Hint Immediate rt_step rt_refl t_step.

Lemma eq_apply {A B:Type} (f:A -> B) a1 a2 (H:a1 = a2):
  f a1 = f a2.
Proof. by subst. Qed.

Lemma if_Some {A B:Prop} {T:Type} (p:sumbool A B) (a b:T):
  (if p then Some a else Some b) = Some (if p then a else b).
Proof. by destruct p. Qed.

Fixpoint gather {T:Type} (l:list (option T)): list T :=
  match l with
    | nil => nil
    | None::l => gather l
    | (Some a)::l => a::(gather l)
  end.

Lemma gather_app {T:Type} (l1 l2: list (option T)):
  gather (l1 ++ l2) = gather l1 ++ gather l2.
Proof.
  revert l2. induction l1; intros; simpl; auto.
  destruct a; auto.
    by rewrite IHl1.
Qed.

Fixpoint concat {T:Type} (l:list (list T)): list T :=
  match l with
    | nil => nil
    | l::ls => l ++ (concat ls)
  end.

Fixpoint for_all {A : Type} (pred:A -> bool) (l:list A): bool :=
  match l with
    | nil => true
    | a::l => (pred a) && (for_all pred l)
  end.

Lemma for_all_spec (A:Type) (pred:A -> bool) (l:list A):
  for_all pred l <-> (forall a (Ha: In a l), pred a).
Proof.
  generalize dependent l. induction l; intros.
  - constructor; intros; auto.
    inv Ha.
  - constructor; intros.
    + simpl in H. apply andb_true_iff in H. destruct H.
      inv Ha; auto.
      by apply IHl.
    + simpl. apply andb_true_iff. split.
      * exploit H; eauto. by left.
      * apply IHl. intros. apply H. by right.
Qed.

Lemma for_all_app {A:Type} (pred:A -> bool) l1 l2:
  for_all pred (l1 ++ l2) = for_all pred l1 && for_all pred l2.
Proof.
  generalize dependent l1. induction l1; simpl; auto.
  rewrite IHl1. destruct (pred a); auto.
Qed.

Lemma for_all_rev {A:Type} (pred:A -> bool) l:
  for_all pred (rev l) = for_all pred l.
Proof.
  generalize dependent l. induction l; simpl; auto.
  rewrite for_all_app, IHl. simpl.
  destruct (for_all pred l), (pred a); auto.
Qed.

Lemma for_all_map_fun: forall {A} f (l : list A), for_all id (map f l) = for_all f l.
Proof.
  induction l; simpl; auto.
  rewrite IHl; auto.
Qed.

Lemma find_app {A:Type} pred (l1 l2:list A):
  find pred (l1++l2) =
  match find pred l1 with
    | Some x => Some x
    | None => find pred l2
  end.
Proof.
  induction l1; simpl; intros; auto.
  destruct (pred a) eqn:Ha; auto.
Qed.

Lemma find_spec (A:Type) (pred:A -> bool) (l:list A) a
      (H: find pred l = Some a):
  pred a /\
  exists l1 l2, l = l1 ++ a :: l2 /\
                find pred l1 = None.
Proof.
  generalize dependent l. induction l; intros.
  - inv H.
  - simpl in H. destruct (pred a0) eqn:Hpred.
    + inv H. split; [done|]. exists nil. exists l. done.
    + exploit IHl; eauto. clear IHl.
      intros [Ha [l1 [l2 [? H12]]]]. subst.
      split; [done|].
      exists (a0::l1). exists l2. split; [done|].
      simpl. by rewrite Hpred.
Qed.

Inductive ForallT {A : Type} (P : A -> Type) : list A -> Type :=
| ForallT_nil: ForallT P nil
| ForallT_cons: forall (x : A) (l : list A),
                  P x -> ForallT P l -> ForallT P (x :: l)
.
Hint Constructors ForallT.

Lemma ForallT_eq {T:Type}
      (l1 l2:list T)
      (H1: ForallT (fun x => forall y, {x = y} + {x <> y}) l1):
  {l1 = l2} + {l1 <> l2}.
Proof.
  generalize dependent l2.
  generalize dependent H1.
  induction l1; intros H1 [|e1 l2]; inv H1;
  (try by left); (try by right).
  destruct (X e1);
    [subst|by right; intro Y; inv Y].
  destruct (IHl1 X0 l2);
    [subst|by right; intro Y; inv Y].
    by left.
Qed.

Fixpoint ForallT_gather {T U:Type} (l:list T) (H: ForallT (fun _:T => option U) l): list U :=
  match H with
    | ForallT_nil => nil
    | ForallT_cons _ _ (Some u) tl => u::(ForallT_gather tl)
    | ForallT_cons _ _  None tl => ForallT_gather tl
  end.

Definition ForallT_pred {T:Type} (l:list T) (H: ForallT (fun _:T => bool) l): bool.
Proof.
  generalize dependent H.
  induction l; intros H.
  - apply true.
  - inv H.
    apply (H1 && (IHl X)).
Defined.

Definition is_empty {X:Type} (l:list X): bool :=
  match l with
    | nil => true
    | _ => false
  end.

Lemma strong_nat_ind:
  forall P : nat -> Prop,
    (forall n : nat, (forall k : nat, (k < n -> P k)%nat) -> P n) ->
    forall n : nat, P n.
Proof.
  intros P H.
  assert (forall n, (forall k, k < n -> P k)%nat).
  - induction n; intros k Hk.
    + inversion Hk.
    + apply Lt.le_lt_or_eq in Hk. destruct Hk as [Hk|Hk].
      * apply Lt.lt_S_n in Hk. apply IHn. auto.
      * inversion_clear Hk. apply H. apply IHn.
  - intros. eapply H0. eauto.
Qed.

Lemma iff_forall {A:Type} (P Q:A -> Prop)
      (H: forall a, P a <-> Q a):
  (forall a, P a) <-> (forall a, Q a).
Proof.
    by constructor; repeat intro; apply H.
Qed.

Lemma list_forall2_refl: forall {A} (l : list A), list_forall2 (fun x y => x = y) l l.
Proof. induction l; constructor; auto. Qed.

Lemma find_rev_norepet {A:Type} (l:list (positive * A))
      (H: list_norepet (List.map fst l))
      i:
  (find (fun id : positive * A => peq i (fst id)) (rev l)) =
  (find (fun id : positive * A => peq i (fst id)) l).
Proof.
  generalize dependent l.
  induction l; intros H; simpl; [done|].
  destruct a. simpl in *. inv H.
  rewrite <- IHl; [|done]. clear IHl.
  rewrite in_rev, <- map_rev in H2.
  apply list_norepet_rev in H3. rewrite <- map_rev in H3.
  generalize dependent H3.
  generalize dependent H2.
  induction (rev l); simpl; intros Hnin Hnor; auto.
  destruct a0. simpl in *. inv Hnor.
  destruct (peq i p0); [subst|]; simpl in *.
  - destruct (peq p0 p); [subst|done]; simpl.
    contradict Hnin. by left.
  - apply IHl0; auto.
Qed.

Definition list_subset {A:Type}
           (l1 l2:list A): Prop :=
  forall a, In a l1 -> In a l2.

Global Program Instance list_subset_Preorder {A} : PreOrder (@list_subset A).
Next Obligation.
  by repeat intro.
Qed.
Next Obligation.
  repeat intro. apply H0. apply H. done.
Qed.

Lemma list_subset_mon: forall {A} (l1 l2 l3 : list A), 
  list_subset l1 l2 -> list_subset (l1 ++ l3) (l2 ++ l3).
Proof.
  unfold list_subset; setoid_rewrite in_app_iff; intros; auto.
  destruct H0; eauto.
Qed.

Definition pair_eq
           {A B:Type}
           (HA: forall (x y:A), {x = y} + {x <> y})
           (HB: forall (x y:B), {x = y} + {x <> y})
           (x y:A * B): {x = y} + {x <> y}.
Proof.
  destruct x, y.
  destruct (HA a a0); [subst|by right; intro X; inv X].
  destruct (HB b b0); [subst|by right; intro X; inv X].
    by left.
Defined.

Definition bin {A:Type}
           (dec: forall x y : A, {x = y} + {x <> y})
           (a : A) (l : list A): bool :=
  in_dec dec a l.

Lemma bin_true: forall {A} eq_dec (x : A) l, bin eq_dec x l = true -> In x l.
Proof. 
  unfold bin; simpl; intros.
  destruct (in_dec eq_dec x l); simpl; auto.
  inv H.
Qed.
Hint Resolve bin_true.

Lemma bin_left {A:Type}
      (dec: forall x y : A, {x = y} + {x <> y})
      (a : A) (l : list A):
  bin dec a (a::l).
Proof.
  unfold bin. simpl. by destruct (dec a a).
Qed.

Lemma bin_right {A:Type}
      (dec: forall x y : A, {x = y} + {x <> y})
      (a b : A) (l : list A):
  bin dec a l -> bin dec a (b::l).
Proof.
  unfold bin. simpl. destruct (dec b a); auto.
  destruct (in_dec dec a l); auto.
Qed.

Lemma bin_subset {A:Type}
      (g:A) l1 l2
      (Hl12: forall g, In g l1 -> In g l2)
      (eq: forall a b:A, {a = b} + {a <> b})
      (Hl1: bin eq g l1):
  bin eq g l2.
Proof.
  unfold bin in *.
  destruct (in_dec eq g l1); simpl in *; try done.
  destruct (in_dec eq g l2); simpl in *; try done.
  by apply Hl12 in i.
Qed.

Ltac basictac_aux :=
  match goal with
    | [H: False |- _] => inv H
    | [H: is_true false |- _] => inv H
    | [H: (is_true true /\ _) \/ (~ is_true true /\ _) |- _] =>
      destruct H as [[_ H]|[H _]]; [|done]
    | [|- (is_true true /\ _) \/ (~ is_true true /\ _)] =>
      left; split; [done|]
    | [H: (~ is_true true /\ _) \/ (is_true true /\ _) |- _] =>
      destruct H as [[H _]|[_ H]]; [done|]
    | [|- (~ is_true true /\ _) \/ (is_true true /\ _)] =>
      right; split; [done|]

    | [H: bin ?eq ?v ?l = true |- _] =>
      unfold bin in H; destruct (in_dec eq v l); [|done]
    | [H: is_true (bin ?eq ?v ?l) |- _] =>
      unfold bin in H; destruct (in_dec eq v l); [|done]
    | [H: In ?a ?b |- is_true (bin ?eq ?a ?b)] =>
      unfold bin; by destruct (In_dec eq a b)
    | [H: In ?a ?b |- bin ?eq ?a ?b = true] =>
      unfold bin; by destruct (In_dec eq a b)

    | [H: (proj_sumbool (option_eq ?eq ?a ?b)) = true |- _] =>
      destruct (option_eq eq a b); [|done]
    | [H: is_true (proj_sumbool (option_eq ?eq ?a ?b)) |- _] =>
      destruct (option_eq eq a b); [|done]
    | [H: (proj_sumbool (option_eq ?eq ?a ?b)) = true |- _] =>
      destruct (option_eq eq a b); [|done]
    | [|- is_true (proj_sumbool (option_eq ?eq ?a ?a))] =>
      by destruct (option_eq eq a a)
    | [|- (proj_sumbool (option_eq ?eq ?a ?a)) = true] =>
      by destruct (option_eq eq a a)

    | [H: ?a = ?a |- _] => clear H
    | [H: is_true true |- _] => clear H
    | [|- is_true true /\ _] =>
      split; [done|]
    | [|- _ /\ is_true true] =>
      split; [|done]
    | [|- true /\ _ = true] =>
      split; [done|]
    | [|- _ /\ true = true] =>
      split; [|done]

    | [H: Some _ = Some _ |- _] => inversion H; (try subst); clear H
    | [H: Some _ = None |- _] => inv H
    | [H: None = Some _ |- _] => inv H
    | [H: None = None |- _] => inv H
    | [H: (_,_) = (_,_) |- _] => inv H

    | [H: ?a <> ?b |- context[option_eq ?d ?a ?b]] =>
      destruct (option_eq d a b); [done|]
    | [H: option_map _ ?cond = Some _ |- _] =>
      let Hcond := fresh "Hcond" in
      destruct cond eqn:Hcond; [inv H|done]
    | [H: Some _ = option_map _ ?cond |- _] =>
      let Hcond := fresh "Hcond" in
      destruct cond eqn:Hcond; [inv H|done]
    | [H: option_map _ ?cond = None |- _] =>
      let Hcond := fresh "Hcond" in
      destruct cond eqn:Hcond; [done|inv H]
    | [H: None = option_map _ ?cond |- _] =>
      let Hcond := fresh "Hcond" in
      destruct cond eqn:Hcond; [done|inv H]

    | [H: (match ?cond with Some _ => _ | None => None end) = Some _ |- _] =>
      let Hcond := fresh "Hcond" in
      destruct cond eqn:Hcond; [inv H|done]
    | [H: (match ?cond with | Some _ => None | None => _ end) = Some _ |- _] =>
      let Hcond := fresh "Hcond" in
      destruct cond eqn:Hcond; [done|]
    | [H: Some = (match ?cond with Some _ => _ | None => None end) |- _] =>
      let Hcond := fresh "Hcond" in
      destruct cond eqn:Hcond; [inv H|done]
    | [H: Some = (match ?cond with | Some _ => None | None => _ end) |- _] =>
      let Hcond := fresh "Hcond" in
      destruct cond eqn:Hcond; [done|]

    | [H: (if ?cond then None else _) = Some _ |- _] =>
      let Hcond := fresh "Hcond" in
      destruct cond eqn:Hcond; [done|]
    | [H: (if ?cond then _ else None) = Some _ |- _] =>
      let Hcond := fresh "Hcond" in
      destruct cond eqn:Hcond; [|done]
    | [H: Some _ = (if ?cond then None else _) |- _] =>
      let Hcond := fresh "Hcond" in
      destruct cond eqn:Hcond; [done|]
    | [H: Some _ = (if ?cond then _ else None) |- _] =>
      let Hcond := fresh "Hcond" in
      destruct cond eqn:Hcond; [|done]
    | [H: negb _ = true |- _] =>
      rewrite negb_true_iff in H
    | [|- negb _ = true] =>
      apply negb_true_iff
    | [H: is_true (_ && _) |- _] =>
      apply andb_true_iff in H; destruct H
    | [H: _ && _ = true |- _] =>
      apply andb_true_iff in H; destruct H
    | [|-  is_true (_ && _)] =>
      apply andb_true_iff
    | [|- _ && _ = true] =>
      apply andb_true_iff
    | [H: is_true (_ || _) |- _] =>
      apply orb_true_iff in H; destruct H
    | [H: _ || _ = true |- _] =>
      apply orb_true_iff in H; destruct H
    | [H: _ \/ _ |- _] =>
      apply orb_true_iff in H
  end.

Ltac destac_aux :=
  match goal with
    | [x: (_ * _)%type |- _] => destruct x
    | [H: (_ /\ _) |- _] => destruct H
    | [H: _ && _ = true |- _] => apply andb_true_iff in H; destruct H
    | [H: exists _, _ |- _] => destruct H
    | [H: Some _ = (let (_, _) := ?cond in _) |- _] =>
      let Hcond := fresh "Hcond" in
      destruct cond eqn:Hcond
    | [H: (let (_, _) := ?cond in _) = Some _ |- _] =>
      let Hcond := fresh "Hcond" in
      destruct cond eqn:Hcond
  end.

Notation "p /0\ q" :=
  (p /\ q)
  (at level 50, no associativity, only parsing).

Notation "p /1\ q" :=
  (fun x0 => p x0 /\ q x0)
  (at level 50, no associativity).

Notation "p /2\ q" :=
  (fun x0 x1 => p x0 x1 /\ q x0 x1)
  (at level 50, no associativity).

Notation "p /3\ q" :=
  (fun x0 x1 x2 => p x0 x1 x2 /\ q x0 x1 x2)
  (at level 50, no associativity).

Notation "p /4\ q" :=
  (fun x0 x1 x2 x3 => p x0 x1 x2 x3 /\ q x0 x1 x2 x3)
  (at level 50, no associativity).

Notation "p /5\ q" :=
  (fun x0 x1 x2 x3 x4 => p x0 x1 x2 x3 x4 /\ q x0 x1 x2 x3 x4)
  (at level 50, no associativity).

Notation "p /6\ q" :=
  (fun x0 x1 x2 x3 x4 x5 => p x0 x1 x2 x3 x4 x5 /\ q x0 x1 x2 x3 x4 x5)
  (at level 50, no associativity).

Notation "p /7\ q" :=
  (fun x0 x1 x2 x3 x4 x5 x6 => p x0 x1 x2 x3 x4 x5 x6 /\ q x0 x1 x2 x3 x4 x5 x6)
  (at level 50, no associativity).

Notation "p /8\ q" :=
  (fun x0 x1 x2 x3 x4 x5 x6 x7 => p x0 x1 x2 x3 x4 x5 x6 x7 /\ q x0 x1 x2 x3 x4 x5 x6 x7)
  (at level 50, no associativity).

Notation "p /9\ q" :=
  (fun x0 x1 x2 x3 x4 x5 x6 x7 x8 => p x0 x1 x2 x3 x4 x5 x6 x7 x8 /\ q x0 x1 x2 x3 x4 x5 x6 x7 x8)
  (at level 50, no associativity).

Notation "p /10\ q" :=
  (fun x0 x1 x2 x3 x4 x5 x6 x7 x8 x9 => p x0 x1 x2 x3 x4 x5 x6 x7 x8 x9 /\ q x0 x1 x2 x3 x4 x5 x6 x7 x8 x9)
  (at level 50, no associativity).

Notation "p /11\ q" :=
  (fun x0 x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 => p x0 x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 /\ q x0 x1 x2 x3 x4 x5 x6 x7 x8 x9 x10)
  (at level 50, no associativity).

Notation "p /12\ q" :=
  (fun x0 x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 x11 => p x0 x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 x11 /\ q x0 x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 x11)
  (at level 50, no associativity).

Notation "p /13\ q" :=
  (fun x0 x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 x11 x12 => p x0 x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 x11 x12 /\ q x0 x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 x11 x12)
  (at level 50, no associativity).

Notation "p /14\ q" :=
  (fun x0 x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 x11 x12 x13 => p x0 x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 x11 x12 x13 /\ q x0 x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 x11 x12 x13)
  (at level 50, no associativity).

Notation "p /15\ q" :=
  (fun x0 x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 x11 x12 x13 x14 => p x0 x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 x11 x12 x13 x14 /\ q x0 x1 x2 x3 x4 x5 x6 x7 x8 x9 x10 x11 x12 x13 x14)
  (at level 50, no associativity).
