Require Import Coqlib.

Definition emptyP {T:Type}: option T -> Prop :=
    fun _ => False.
Definition singletonP {T:Type} (m:option T): option T -> Prop :=
  fun m' => m = m'.
Definition NoneP {T:Type}: option T -> Prop := singletonP None.
Definition SomeP {T:Type} (x:T): option T -> Prop := singletonP (Some x).
Definition liftP {T:Type} (r:T -> Prop): option T -> Prop :=
  fun my =>
    match my with
      | Some y => r y
      | _ => False
    end.

Definition bind {T U:Type} (x:option T) (f: T -> option U): option U :=
  match x with
    | None => None
    | Some v => f v
  end.
Definition bindP {T U:Type} (mt:option T) (p:T -> option U -> Prop): option U -> Prop :=
  match mt with
    | None => NoneP
    | Some v => p v
  end.
Inductive bindPP {T U:Type} (r1:option T -> Prop) (r2:T -> option U -> Prop) (ty:option U): Prop :=
| bindPP_left_None
    (Hr1: r1 None)
    (Hty: ty = None):
    bindPP r1 r2 ty
| bindPP_right
    v1 (Hr1: r1 (Some v1))
    (Hr2: r2 v1 ty):
    bindPP r1 r2 ty
.

Definition or_else {T: Type} (a b : option T) : option T:=
  match a with
  | Some x => Some x
  | None => b
  end.

Lemma or_else_next: forall {A} (a b : option A), a = None -> or_else a b = b.
Proof. by destruct a. Qed.

Lemma or_else_not_None: forall {A} a (b : A), or_else a (Some b) <> None.
Proof. by destruct a. Qed.

Notation "x >>= f" := (bind x f) (at level 41).
Notation "'doT' x <- a ; b" := (a >>= (fun x => b)) (at level 41).
Notation "'doT' ( x , y ) <- a ; b" := (a >>= (fun xy => let x := fst xy in let y := snd xy in b)) (at level 41).
Notation "'doT' ( x , y , z ) <- a ; b" := (a >>= (fun xyz => let x := fst (fst xyz) in let y := snd (fst xyz) in let z := snd xyz in b)) (at level 41).
Notation "'doT' ( x , y , z , w ) <- a ; b" := (a >>= (fun xyzw => let x := fst (fst (fst xyzw)) in let y := snd (fst (fst xyzw)) in let z := snd (fst xyzw) in let w := snd xyzw in b)) (at level 41).
Notation "x '>>>=' p" := (bindP x p) (at level 41).
Notation "'assumeP' p ; b" := (fun x => p /\ b x) (at level 41).
Notation "'assertP' p ; b" := (fun x => (~p /\ x = None) \/ (p /\ b x)) (at level 41).
Notation "'assertB' p ; b" := (if p then b else None) (at level 41).
Notation "'doP' x <- a ; b" := (a >>>= (fun x => b)) (at level 41).
Notation "'doP' ( x , y ) <- a ; b" := (a >>>= (fun xy => let x := fst xy in let y := snd xy in b)) (at level 41).
Notation "'doP' ( x , y , z ) <- a ; b" := (a >>>= (fun xyz => let x := fst (fst xyz) in let y := snd (fst xyz) in let z := snd xyz in b)) (at level 41).
Notation "'doP' ( x , y , z , w ) <- a ; b" := (a >>>= (fun xyzw => let x := fst (fst (fst xyzw)) in let y := snd (fst (fst xyzw)) in let z := snd (fst xyzw) in let w := snd xyzw in b)) (at level 41).
Notation "r1 '>>>>=' r2" := (bindPP r1 r2) (at level 41).
Notation "'doPP' x <- a ; b" := (a >>>>= (fun x => b)) (at level 41).
Notation "'doPP' ( x , y ) <- a ; b" := (a >>>>= (fun xy => let x := fst xy in let y := snd xy in b)) (at level 41).
Notation "'doPP' ( x , y , z ) <- a ; b" := (a >>>>= (fun xyz => let x := fst (fst xyz) in let y := snd (fst xyz) in let z := snd xyz in b)) (at level 41).
Notation "'doPP' ( x , y , z , w ) <- a ; b" := (a >>>>= (fun xyzw => let x := fst (fst (fst xyzw)) in let y := snd (fst (fst xyzw)) in let z := snd (fst xyzw) in let w := snd xyzw in b)) (at level 41).

Lemma bindPP_exists {T U:Type}
      (v:T) (e:_ -> Prop) (Hv: e (Some v))
      (f:_ -> option U -> Prop) (Hf: exists r, f v r):
  exists r, (doPP w <- e; f w) r.
Proof.
  destruct Hf as [r Hr].
  exists r.
  eapply bindPP_right; eauto.
Qed.
Lemma bindPP_exists' {T U:Type}
      (v:T) (e:_ -> Prop) (Hv: e (Some v))
      (f:_ -> option U -> Prop) (Hf: exists r, f v r) H:
  exists r, (((doPP w <- e; f w) r) \/ H r).
Proof.
  exploit @bindPP_exists; eauto. intros [r Hr].
  exists r. left. auto.
Qed.

Lemma not_bindPP_None {T U:Type}
      (r1 : option T -> Prop)
      (r2 : T -> option U -> Prop)
      (H: ~ bindPP r1 r2 None):
  ~ (r1 None) /\
  forall t, r1 (Some t) -> ~ r2 t None.
Proof.
  split.
  - intro X. apply H. by constructor.
  - repeat intro. apply H. eapply bindPP_right; eauto.
Qed.

Lemma bindPP_True: forall {U V} f r, (doPP x <- liftP (fun _ : V => True); f x) r <-> 
  exists v : V, (fun (x : V) (y : option U) => f x y) v r.
Proof.
  split; intros.
  - inv H; inv Hr1. eexists. eauto.
  - inv H. eapply bindPP_right; eauto. constructor.
Qed.

Ltac monadtac_aux :=
  match goal with
    | [H: SomeP _ _ |- _] => inv H
    | [H: NoneP _ _ |- _] => inv H
    | [H: singletonP _ _ |- _] => inv H
    | [|- exists x, singletonP _ x] => by eexists
    | [|- exists x, singletonP _ x \/ _] => eexists; by left

    | [H: assertB ?cond; _ = Some _ |- _] =>
      let Hcond := fresh "Hcond" in
      destruct cond eqn:Hcond; [|done]
    | [|- assertB ?cond; _ <> None] =>
      let Hcond := fresh "Hcond" in
      destruct cond eqn:Hcond; [|done]
    | [H: bind ?c _ = Some _ |- _] =>
      let Hcond := fresh "Hcond" in
      destruct c eqn:Hcond; [|done]

    | [H: (?p:option _ -> Prop) ?v |- exists r, (doPP w <- ?p; _) r] =>
      destruct v;
      [eapply bindPP_exists; eauto|eexists; eapply bindPP_left_None; eauto]
    | [H: (?p:option _ -> Prop) ?v |- exists r, ((doPP w <- ?p; _) r) \/ _] =>
      destruct v;
      [eapply bindPP_exists'; eauto|eexists; left; eapply bindPP_left_None; eauto]

    | [H: bindPP _ _ (Some _) |- _] => inv H
    | [H: bindP ?c _ (Some _) |- _] =>
      let Hcond := fresh "Hcond" in
      destruct c eqn:Hcond; [simpl in H|done]

    | [H: ~ bindPP ?r1 ?r2 None |- _] =>
      apply not_bindPP_None in H; destruct H
  end.

Ltac monadtac_agr_aux :=
  match goal with
    | [H: context[bind ?x _] |- _] =>
      let Hx := fresh "Hx" in
      destruct x eqn:Hx
    | [|- context[bind ?x _]] =>
      let Hx := fresh "Hx" in
      destruct x eqn:Hx

    | [H: context[bindP ?x _ _] |- _] =>
      let Hx := fresh "Hx" in
      destruct x eqn:Hx
    | [|- context[bindP ?x _ _]] =>
      let Hx := fresh "Hx" in
      destruct x eqn:Hx

    | [H: context[bindPP ?x _ _] |- _] =>
      inv H

    | [|- bindPP _ _ (Some _)] =>
      eapply bindPP_right; eauto
  end.

Definition option_to_bool {A:Type} (o:option A): bool :=
  if o then true else false.
Coercion option_to_bool: option >-> bool.
